/*
 * hgcommons 7
 * Hammurapi Group Common Library 
 * Copyright (C) 2003  Hammurapi Group
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * URL: http://www.hammurapi.biz/hammurapi-biz/ef/xmenu/hammurapi-group/products/products/hgcommons/index.html
 * e-Mail: support@hammurapi.biz
 */
package biz.hammurapi.ant;

import java.lang.reflect.InvocationTargetException;
import java.util.Collection;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import org.apache.tools.ant.AntClassLoader;
import org.apache.tools.ant.BuildException;
import org.apache.tools.ant.Task;
import org.apache.tools.ant.types.Path;

import biz.hammurapi.config.ConfigurationException;
import biz.hammurapi.config.Parameterizable;

/**
 * Object entry. Base class for configurable objects. 
 * @ant.element name="objectentry"
 * @author Pavel Vlasov	
 * @version $Revision: 1.5 $
 */
public class ObjectEntry extends Task {
	private String className;
	private String value;
	private List parameters=new LinkedList();
	
	/**
	 * Configuration parameter. Object entry class must implement biz.hammurapi.config.Parameterizable
	 * @ant.non-required
	 * @param parameter
	 * @throws BuildException
	 */
	public void addConfiguredParameter(Param parameter) throws BuildException {
		if (parameter.getName()==null) {
			throw new BuildException("Unnamed parameter");
		}
		parameters.add(parameter);
	}		

	/**
	 * Either value or class name is required. If both class name and value are
	 * specified then that class should have a constructor with one String parameter.
	 * @ant.non-required
	 * @param className
	 */
	public void setClassName(String className) {
		this.className = className;
	}

	/**
	 * Not required if class name is set.
	 * @ant.non-required
	 * @param value
	 */
	public void setValue(String value) {
		this.value = value;
	}

	/**
	 * @ant.ignore
	 * @return
	 */
	public Collection getParameters() {
		return parameters;
	}
	
	private Object theObject;
	
	protected void validateClass(Class clazz) throws BuildException {
		
	}
	
	protected void validateInstance(Object instance) throws BuildException {
		
	}
	
	public Object getObject(ClassLoader masterClassLoader) throws BuildException {
		if (theObject==null) {
			if (className==null) {
				if (parameters.isEmpty()) {
					if (value==null) {
						return null;
					}
					theObject=value;
				} else {
					throw new BuildException("Nested parameters supported only if classname attribute is set");
				}
			} else {
				try {
					ClassLoader classLoader=masterClassLoader==null ? getProject().getCoreLoader() : masterClassLoader;
					if (classPath!=null) {
						if (masterClassLoader==null) {
							classLoader=new AntClassLoader(getProject(), classPath);
						} else {
							classLoader=new AntClassLoader(masterClassLoader, getProject(), classPath, false);
						}
					}
					
					if (classLoader==null) {
						classLoader=getClass().getClassLoader();
					}
					
					Class theClass=classLoader.loadClass(className);
					validateClass(theClass);
					if (!(parameters.isEmpty() || Parameterizable.class.isAssignableFrom(theClass))) {
						throw new BuildException(className+" does not support parameters");
					}
					
					if (value==null) {
						theObject=theClass.newInstance();
					} else {
						theObject=theClass.getConstructor(new Class[] {String.class}).newInstance(new Object[] {value});
					}
					
					if (!parameters.isEmpty()) {
						// Double check
						if (theObject instanceof Parameterizable) {
							Iterator it=parameters.iterator();
							while (it.hasNext()) {
								Param param=(Param) it.next();
								try {
									if (!((Parameterizable) theObject).setParameter(param.getName(), param.getObject(masterClassLoader))) {
									    throw new BuildException(theObject.getClass().getName()+" does not support parameter "+param.getName());
									}
								} catch (ConfigurationException e) {
									throw new BuildException("Could not set parameter "+param.getName()+" for object entry "+theObject.getClass().getName(), e);
								}
							}
						} else {						
							throw new BuildException(className+" does not support parameters");
						}
					}
					validateInstance(theObject);
				} catch (ClassNotFoundException e) {
					throw new BuildException("Class not found: "+className, e);
				} catch (InstantiationException e) {
					throw new BuildException("Can not instantiate: "+className, e);
				} catch (IllegalAccessException e) {
					throw new BuildException("Can not instantiate: "+className, e);
				} catch (InvocationTargetException e) {
					throw new BuildException("Can not instantiate: "+className, e);
				} catch (NoSuchMethodException e) {
					throw new BuildException("Constructor "+className+"(String) not found", e);
				}	
			}
		}
		return theObject;
	}
	/**
	 * @return
	 */
	protected String getClassName() {
		return className;
	}

	/**
	 * @return
	 */
	protected String getValue() {
		return value;
	}
	
	/**
	 * Classpath for loading classes
	 * @ant:non-required 
	 */
	private Path classPath;
	
	public void setClassPath(Path classPath) {
		if (this.classPath == null) {
			this.classPath = classPath;
		} else {
			this.classPath.append(classPath);
		}
	}

	/**
	 * Maybe creates a nested classpath element.
	 * @ant:non-required
	 */
	public Path createClasspath() {
		if (classPath == null) {
			classPath = new Path(getProject());
		}
		return classPath.createPath();
	}	
	
	public void addText(String text) {
		this.value=text;
	}
}
