/*
  * hgee 2
 * Hammurapi Group Enterprise Extensions 
 * Copyright (C) 2006  Hammurapi Group
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * URL: http://www.hammurapi.biz/products/enterpriseextensions/index.html
 * e-Mail: support@hammurapi.biz 
 */
package biz.hammurapi.cache;

import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;
import java.util.Timer;
import java.util.TimerTask;

import biz.hammurapi.CarryOverException;
import biz.hammurapi.RuntimeException;
import biz.hammurapi.cache.AbstractProducer;
import biz.hammurapi.cache.Cache;
import biz.hammurapi.cache.Entry;
import biz.hammurapi.cache.Producer;
import biz.hammurapi.cache.sql.CacheEntry;
import biz.hammurapi.cache.sql.FileCacheEngine;
import biz.hammurapi.config.ConfigurationException;
import biz.hammurapi.sql.DataIterator;
import biz.hammurapi.sql.SQLProcessor;
import biz.hammurapi.sql.SQLRuntimeException;
import biz.hammurapi.sql.Transaction;
import biz.hammurapi.sql.hypersonic.HypersonicStandaloneDataSource;
import biz.hammurapi.util.Acceptor;


/**
 * Caches objects in files.
 * @author Pavel Vlasov
 * @version $Revision: 1.7 $
 */
public class FileCache extends AbstractProducer implements Cache {

	private static final int CLEANUP_INTERVAL = 10*60*1000;
	private static final int MAX_FILES_PER_DIRECTORY = 10000;
	private static final int ACTIVE_DIRS = 5;
	private HypersonicStandaloneDataSource ds;
	private FileCacheEngine engine; 
	private File dataDir;
	private long size;
	private long maxSize;
	private Producer producer;
		
	private class DirectoryInfo {
		File directory;
		int fileCount;
				
		DirectoryInfo() {
			synchronized (FileCache.this) {
				do {
					directory=new File(dataDir, Long.toString(System.currentTimeMillis(), Character.MAX_RADIX));
				} while (directory.exists());
				
				directory.mkdir();
				fileCount=directory.listFiles().length;
			}
		}
		
		DirectoryInfo(File dir) {
			directory=dir;
			fileCount=directory.listFiles().length;
		}
		
		File nextFile() throws IOException {
			fileCount++;
			return File.createTempFile("cache_", ".tmp", directory);
		}
	}
	
	private List directories=new ArrayList();
	
	private File nextFile() throws IOException {
		Iterator it=directories.iterator();
		while (it.hasNext()) {
			DirectoryInfo di=(DirectoryInfo) it.next();
			if (di.fileCount>MAX_FILES_PER_DIRECTORY) {
				it.remove();
			}
		}
		
		while (directories.size()<ACTIVE_DIRS) {
			directories.add(new DirectoryInfo());
		}
		
		return ((DirectoryInfo) directories.get(((int) (Math.random() * Integer.MAX_VALUE)) % ACTIVE_DIRS)).nextFile();
	}

	/**
	 * 
	 * @param producer
	 * @param dir
	 * @param maxSize - Maximum cache size. Number &lt;=0 means no limit
	 * @throws IOException
	 */
	public FileCache(Producer producer, File dir, long maxSize) throws IOException {
		super();
		this.maxSize=maxSize>0 ? maxSize : Long.MAX_VALUE;
		this.producer=producer;
		if (producer!=null) {
			producer.addCache(this);
		}
		
		dataDir=new File(dir, "data");
		if (!dataDir.exists()) {
			dataDir.mkdir();
		}
		if (!dataDir.exists()) {
			throw new IOException("Cannot create directory "+dataDir.getAbsolutePath());
		}
		if (!dataDir.isDirectory()) {
			throw new IOException("Not a directory: "+dataDir.getAbsolutePath());			
		}
		
		File[] dirs=dataDir.listFiles();		
		for (int i=0; i<dirs.length; i++) {
			directories.add(new DirectoryInfo(dirs[i]));
			File[] entries=dirs[i].listFiles();		
			for (int j=0; j<entries.length; i++) {				
				size+=entries[i].length();
			}
		}
		
		try {
			ds=new HypersonicStandaloneDataSource(
					new File(dir, "entries").getAbsolutePath(), 
					new Transaction() {

						public boolean execute(SQLProcessor processor) throws SQLException {
							try {
								processor.executeScript(new InputStreamReader(getClass().getResourceAsStream("FileCache.sql")));
							} catch (IOException e) {
								throw new CarryOverException(e);
							}
							return true;
						}
					});
			
			engine = new FileCacheEngine(new SQLProcessor(ds, null));
		} catch (ClassNotFoundException e) {
			throw new IOException("Caused by "+e);
		} catch (SQLException e) {
			throw new IOException("Caused by "+e);
		} catch (CarryOverException e) {
			throw (IOException) e.getCause();
		}
	}
	
	private boolean shutDown=false;
	
	private void checkShutdown() {
		if (shutDown) {
			throw new IllegalStateException("Shut down");
		}
	}
	
	/**
	 * Shuts down entries database and janitor thread.
	 */
	public void stop() {
		shutDown=true;
		ds.shutdown();
		janitorTask.cancel();
		if (isOwnTimer) {
			timer.cancel();
		}
	}

	synchronized public void put(Object key, Object value, long time, long expirationTime) {
		checkShutdown();
		if (key instanceof String && value instanceof Serializable) {
			try {
				remove(key);
				
				// TODO - modify to avoid running into OS limitations of files per dir.
				File out=nextFile();
				ObjectOutputStream oos=new ObjectOutputStream(new FileOutputStream(out));
				try {
					oos.writeObject(value);
				} finally {
					oos.close();
				}
				
				String fileName = out.getParentFile().getName()+File.separator+out.getName();
				//System.out.println(key + " -> " + fileName);
				engine.insertCacheEntry((String) key, fileName, time, expirationTime, 0);
				size+=out.length();
				
				if (size>maxSize) {
					Iterator it=engine.getCacheEntryLastAccessOrdered().iterator();
					while (size>maxSize && it.hasNext()) {
						CacheEntry entry=(CacheEntry) it.next();
						File file = new File(dataDir, entry.getValueFile());
						long length=file.length();
						if (file.delete()) {
							size-=length;
						}
						engine.deleteCacheEntry(entry.getEntryKey());					
					}
					
					((DataIterator) it).close();
				}
			} catch (IOException e) {
				e.printStackTrace();
				// ignore
			} catch (SQLException e) {
				e.printStackTrace();
				// ignore
			}
		} else {
			System.err.println("WARN: Cannot serialize "+value.getClass());
		}
	}
	
	private TimerTask janitorTask = new TimerTask() {
			
			public void run() {
				try {
					long now = System.currentTimeMillis();
					Iterator it=engine.getCacheEntryExpiresLE(now).iterator();
					synchronized (FileCache.this) {
						while (it.hasNext()) {				
							File file = new File(dataDir, ((CacheEntry) it.next()).getValueFile());
							long length=file.length();
							if (file.delete()) {
								size-=length;
							}
						}
					}						
					engine.deleteCacheEntryExpiresLE(now);						
				} catch (SQLException e) {
					e.printStackTrace();
					// ignore
				} catch (SQLRuntimeException e) {
					e.printStackTrace();
					// ignore
				}
			}
		};

	synchronized public void clear() {
		checkShutdown();
		try {
			engine.deleteCacheEntry();
		} catch (SQLException e) {
			 e.printStackTrace();
			// ignore
		}
		File[] entries=dataDir.listFiles();
		for (int i=0; i<entries.length; i++) {
			long length=entries[i].length();
			if (entries[i].delete()) {
				size-=length;
			}
		}
	}

	synchronized public void remove(Object key) {
		checkShutdown();
		if (key instanceof String) {
			try {
				CacheEntry cacheEntry = engine.getCacheEntry((String) key);
				if (cacheEntry!=null) {
					File file = new File(dataDir, cacheEntry.getValueFile());
					long length=file.length();
					if (file.delete()) {
						size-=length;
					}
					engine.deleteCacheEntry((String) key);
				}
			} catch (SQLException e) {
				e.printStackTrace();
			}			
		}
		onRemove(key);
	}

	public Entry get(Object key) {
		checkShutdown();
		Entry ret=key instanceof String ? _get((String) key) : null;
		if (ret==null && producer!=null) {
			ret=producer.get(key);
			if (ret!=null) {
				put(key, ret.get(), ret.getTime(), ret.getExpirationTime());
			}
		}
		//System.out.println("*** "+key+" Restored ---> "+ret);
		return ret;
	}
		
	synchronized private Entry _get(String key) {
		try {			
			CacheEntry entry=engine.getCacheEntry(key);
			if (entry==null) {
				return null; 
			}
			long now=System.currentTimeMillis();
			if (entry.getEntryExpires()<=0 || now<=entry.getEntryExpires()) {
				engine.setLastAccess(now, key);
				File valueFile=new File(dataDir, entry.getValueFile());
				//System.out.println("Restoring "+key+" <- "+entry.getValueFile()+" ("+valueFile.length()+" bytes)");
				ObjectInputStream ois=new ObjectInputStream(new FileInputStream(valueFile));
				try {
					final Object value=ois.readObject();
					final long expires=entry.getEntryExpires();
					final long time=entry.getEntryTime();
					return new Entry() {

						public long getExpirationTime() {
							return expires;
						}

						public long getTime() {
							return time;
						}
						
						public Object get() {
							return value;
						}				
					};
				} finally {
					ois.close();
				}
			}
			return null;
		} catch (IOException e) {
			e.printStackTrace();
			remove(key);
			return null;
		} catch (ClassNotFoundException e) {
			e.printStackTrace();
			remove(key);
			return null;
		} catch (SQLException e) {
			e.printStackTrace();
			return null;			
		}
	}

	public static void main(String[] args) throws Exception {
		FileCache fileCache=new FileCache(new Producer() {

			public Entry get(Object key) {
				try {
					int size=(int) (Math.random()*20000);
					ByteArrayOutputStream bos=new ByteArrayOutputStream();
					ObjectOutputStream oos=new ObjectOutputStream(bos);
					
					while (bos.size()<size) {
						oos.writeObject(key+"_at_"+bos.size());										
					}
					oos.close();
					bos.close();
					
					final byte[] bytes = bos.toByteArray();
					return new Entry() {
	
						public long getExpirationTime() {
							return 0;
						}
	
						public long getTime() {
							return 0;
						}
						
						public Object get() {
							return bytes;
						}
						
					};
				} catch (IOException e) {
					e.printStackTrace();
					return null;
				}
			}

			public void addCache(Cache cache) {
			    // No action
			}

			public Set keySet() {
				return null;
			}
			
		},
		new File("C:\\_fileCacheTest"),
		100000);
		
		for (int i=0; i<300; i++) {
			fileCache.get("Key_"+i);
		}
		
		fileCache.stop();
	}

	public void remove(Acceptor acceptor) {
		checkShutdown();
		Iterator it=engine.getCacheEntry().iterator();
		try {
			CacheEntry ce=(CacheEntry) it.next();
			if (acceptor.accept(ce.getEntryKey())) {
				try {
					File file = new File(dataDir, engine.getCacheEntry(ce.getEntryKey()).getValueFile());
					long length=file.length();
					if (file.delete()) {
						size-=length;
					}
					engine.deleteCacheEntry(ce.getEntryKey());
				} catch (SQLException e) {
					e.printStackTrace();
				}			
			}
			onRemove(ce.getEntryKey());
		} finally {
			try {
				((DataIterator) it).close();
			} catch (SQLException e) {
				throw new RuntimeException(e);
			}
		}
	}
	
	public Set keySet() {
		HashSet ret = new HashSet();
		Iterator it=engine.getCacheEntry().iterator();
		while (it.hasNext()) {
			ret.add(((CacheEntry) it.next()).getEntryKey());
		}
		
		if (producer!=null) {
			Set pkeys=producer.keySet();
			if (pkeys!=null) {
				ret.addAll(pkeys);
			}
		}
		
		return ret;
	}

	public boolean isActive() {
		return !shutDown;
	}
	
	private Timer timer;
	private boolean isOwnTimer;

	public void start() throws ConfigurationException {
		if (timer==null) {
			timer=new Timer();
			isOwnTimer=true;
		}
		
		timer.schedule(janitorTask, CLEANUP_INTERVAL, CLEANUP_INTERVAL);
	}

	public void setOwner(Object owner) {
		// Nothing
		
	}
	
}
