/*
  * hgcommons 7
 * Hammurapi Group Common Library 
 * Copyright (C) 2003  Hammurapi Group
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * URL: http://www.hammurapi.biz/hammurapi-biz/ef/xmenu/hammurapi-group/products/products/hgcommons/index.html
 * e-Mail: support@hammurapi.biz 
 */
package biz.hammurapi.cache;

import java.lang.ref.ReferenceQueue;
import java.lang.ref.SoftReference;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Map;
import java.util.Set;
import java.util.Timer;
import java.util.TimerTask;
import java.util.WeakHashMap;

import biz.hammurapi.config.Component;
import biz.hammurapi.config.ConfigurationException;
import biz.hammurapi.metrics.MeasurementCategory;
import biz.hammurapi.util.Acceptor;


/**
 * Memory sensitive cache. Uses soft references to cache objects.
 * @author Pavel Vlasov
 * @version $Revision: 1.7 $
 */
public class MemoryCache extends AbstractProducer implements Cache, Component {
	protected Map cache=new HashMap();
	protected Map reverseCache=new WeakHashMap();
	
	private ReferenceQueue queue=new ReferenceQueue();
	protected Producer producer;
	private Cache fallBack;
	private MeasurementCategory measurementCategory;
		
	private void addMeasurement(String name, double value, long time) {
		if (measurementCategory!=null) {
			measurementCategory.addMeasurement(name, value, time);
		}
	}
	
	private void addMeasurement(String name, double value) {
		if (measurementCategory!=null) {
			measurementCategory.addMeasurement(name, value, 0);
		}
	}
	
	private class CacheReference extends SoftReference {

		private Object key;
		private long expires;
		private long time;

		/**
		 * @param referent
		 * @param q
		 * @param expires
		 */
		public CacheReference(Object key, Object value, long time, long expires) {
			super(value, queue);
			this.key=key;
			this.expires=expires;
			this.time=time;
		}

		/**
		 * @param string
		 * @param i
		 * @param l
		 */
		public void addMeasurement(String name, double value) {
			MemoryCache.this.addMeasurement(name, value);
		}		
	}
	
	/**
	 * Used as entry if fallBack cache is used.
	 * @author Pavel Vlasov
	 * @version $Revision: 1.7 $
	 */
	private class FallBackEntry {
		Object value;
		private long expires;
		private long time;
		private Object key;
		
		FallBackEntry(Object key, Object value, long time, long expires) {
			this.key=key;
			this.value=value;
			this.expires=expires;
			this.time=time;
		}
		
		protected void finalize() throws Throwable {
			if (fallBack.isActive() && (expires<=0 || System.currentTimeMillis()<expires)) {
				fallBack.put(key, value, time, expires);
			}
			
			super.finalize();
		}		
	}
	
	private static class ReferenceThread extends Thread {
		private ReferenceQueue queue;
		private Map cache;

		private ReferenceThread(ReferenceQueue queue, Map cache) {
			this.queue=queue;
			this.cache=cache;
			setDaemon(true);
			setName("Removes cleared entries");
			start();
		}
		
		public void run() {
			try {
				while (true) {
					CacheReference cacheReference = (CacheReference) queue.remove();
					Object key=cacheReference.key;
					synchronized (cache) {
						cache.remove(key);						
					}
					cacheReference.addMeasurement("garbage collected", 1);
				}
			} catch (InterruptedException e) {
				return;
			}
		}
	}
	
	private Thread referenceThread=new ReferenceThread(queue, cache);
	
	private static class JanitorTask extends TimerTask {
		private Map cache;
		private MeasurementCategory measurementCategory;
		private Map reverseCache;
		
		private void addMeasurement(String name, double value, long time) {
			if (measurementCategory!=null) {
				measurementCategory.addMeasurement(name, value, time);
			}
		}
		
		private JanitorTask(Map cache, Map reverseCache, MeasurementCategory measurementCategory) {
			this.cache=cache;
			this.reverseCache=reverseCache;
			this.measurementCategory=measurementCategory;
		}
			
		public void run() {
			long now=System.currentTimeMillis();
			int expired=0;
			synchronized (cache) {
				Iterator it=new LinkedList(cache.keySet()).iterator();
				while (it.hasNext()) {
					Object key=it.next();
					CacheReference cacheReference = (CacheReference) cache.get(key);
					long expires = (cacheReference).expires;
					if (expires>0 && expires<now) {
						cache.remove(key);								
						reverseCache.remove(cacheReference.get());
						expired++;
					}
				}
				addMeasurement("size", cache.size(), now);
			}
			
			if (expired>0) {
				addMeasurement("expired", expired, now);
			}
		}			
	}
	
	private TimerTask janitorTask;
	
	/**
	 * Default cache cleanup interval.
	 */
	public static final long CLEANUP_INTERVAL=60000;

	/**
	 * Constructs cache with default cleanup interval (1 minute).
	 * @param producer Producer
	 * @param fallBack Fallback cache
	 * @param measurementCategory Measurement category to report cache statistics
	 */
	public MemoryCache(Producer producer, Cache fallBack, MeasurementCategory measurementCategory) {
		this(producer, fallBack, measurementCategory, null, CLEANUP_INTERVAL);
	}
	
	/**
	 * Constructs cache with default cleanup interval (1 minute).
	 * @param producer Producer
	 * @param fallBack Fallback cache
	 * @param measurementCategory Measurement category to report cache statistics
	 * @param cleanupInterval Interval between removals of expired entries.
	 * @param timer Timer which will invoke cleanup tasks, if it is null then a new timer is created internally. 
	 * Use this parameter to share timers between multiple caches in order to reduce number of threads in the 
	 * application.
	 */
	public MemoryCache(Producer producer, Cache fallBack, MeasurementCategory measurementCategory, Timer timer, long cleanupInterval) {
		super();
		
		this.producer=producer;
		if (producer!=null) {
			producer.addCache(this);
		}
		
		this.fallBack=fallBack;
		if (fallBack!=null) {
			addCache(fallBack);
		}
		
		this.measurementCategory=measurementCategory;		
		this.timer=timer;
		
		this.cleanupInterval=cleanupInterval;
	}
	
	private boolean isOwnTimer;
	private Timer timer;
	private long cleanupInterval;
	
	public void put(Object key, Object value, long time, long expirationTime) {
		checkShutdown();
		synchronized (cache) {
			cache.put(key, new CacheReference(key, toValue(key, value, time, expirationTime), time, expirationTime));
			reverseCache.put(value, key);
			addMeasurement("put", 1, time);
		}
	}

	/**
	 * @param key
	 * @param value
	 * @param expirationTime
	 * @return
	 */
	private Object toValue(Object key, Object value, long time, long expirationTime) {
		return fallBack==null ? value : new FallBackEntry(key, value, time, expirationTime);
	}
	
	private Object fromValue(Object object) {
		if (object instanceof FallBackEntry) {
			return ((FallBackEntry) object).value;
		}
		return object;
	}

	public Entry get(Object key) {
		checkShutdown();
		synchronized (cache) {
			CacheReference cr=(CacheReference) cache.get(key);
			long now = System.currentTimeMillis();
			addMeasurement("get", 1, now);
			boolean doFallBack=true;
			if (cr!=null) {
				if ((cr.expires>0 && cr.expires<now) || cr.get()==null) {
					cache.remove(key);	
					reverseCache.remove(fromValue(cr.get()));
					doFallBack=false;
				} else {
					final Object o=fromValue(cr.get());
					final long et=cr.expires;
					final long time=cr.time;
					return new Entry() {

						public long getExpirationTime() {
							return et;
						}

						public long getTime() {
							return time;
						}

						public Object get() {
							return o;
						}
						
					};
				}
			} 
			
			if (doFallBack && fallBack!=null) {
				Entry entry = fallBack.get(key);
				if (entry!=null) {
					cache.put(key, new CacheReference(key, toValue(key, entry.get(), entry.getTime(), entry.getExpirationTime()), entry.getTime(), entry.getExpirationTime()));
					reverseCache.put(entry.get(), key);
					return entry;
				}
			}
			
			if (key instanceof ProducingKey) {
				Entry entry = ((ProducingKey) key).get();
				cache.put(key, new CacheReference(key, toValue(key, entry.get(), entry.getTime(), entry.getExpirationTime()), entry.getTime(), entry.getExpirationTime()));
				reverseCache.put(entry.get(),key);
				addMeasurement("produce", 1, now);
				return entry;
			}
			
			if (producer!=null) {
				Entry entry = producer.get(key);
				cache.put(key, new CacheReference(key, toValue(key, entry.get(), entry.getTime(), entry.getExpirationTime()), entry.getTime(), entry.getExpirationTime()));
				reverseCache.put(entry.get(),key);
				addMeasurement("produce", 1, now);
				return entry;
			}
			
			return null;
		}
	}
	
	public void clear() {
		synchronized (cache) {
			cache.clear();
			reverseCache.clear();
			
			if (fallBack!=null) {
				fallBack.clear();
			}
		}
	}
	
	public void remove(Object key) {
		synchronized (cache) {
			CacheReference cr = (CacheReference) cache.remove(key);
			if (cr!=null) {
				reverseCache.remove(fromValue(cr.get()));
			}
		}
		addMeasurement("remove", 1);
		onRemove(key);
	}

	public void remove(Acceptor acceptor) {
		synchronized (cache) {
			Iterator it=cache.keySet().iterator();
			int removed=0;
			while (it.hasNext()) {
				Object key=it.next();
				if (acceptor.accept(key)) {
					Entry entry=(Entry) cache.get(key);
					if (entry!=null) {
						reverseCache.remove(fromValue(entry.get()));
					}
					it.remove();
					removed++;
					onRemove(key);
				}
			}
			if (removed>0) {
				addMeasurement("remove", removed);
			}
		}
	}
	
	private boolean shutDown=false;
	
	private void checkShutdown() {
		if (shutDown) {
			throw new IllegalStateException("Shut down");
		}
	}

	public void stop() {
		if (!shutDown) {			
			cache.clear();
			reverseCache.clear();
			referenceThread.interrupt();
			if (janitorTask!=null) {
				janitorTask.cancel();
			}
			if (isOwnTimer && timer!=null) {
				timer.cancel();
			}
			shutDown=true;
		}
	}
	
	protected void finalize() throws Throwable {
		if (isActive()) {
			stop();
		}
		super.finalize();
	}

	public Set keySet() {
		HashSet ret = new HashSet();
		synchronized (cache) {
			ret.addAll(cache.keySet());
		}
		
		if (producer!=null) {
			Set pkeys=producer.keySet();
			if (pkeys!=null) {
				ret.addAll(pkeys);
			}
		}
		
		if (fallBack!=null) {
			Set fkeys=fallBack.keySet();
			if (fkeys!=null) {
				ret.addAll(fkeys);
			}
		}
		
		return ret;
	}

	public boolean isActive() {
		return !shutDown;
	}

	/**
	 * Creates timer if neccessary, creates cleanup task and schedules it. 
	 */
	public void start() throws ConfigurationException {
		if (timer==null) {
			timer=new Timer(true);
			isOwnTimer=true;
		}
		
		janitorTask=new JanitorTask(cache, reverseCache, measurementCategory);
		timer.schedule(janitorTask, cleanupInterval, cleanupInterval);
	}

	public void setOwner(Object owner) {
		// TODO Nothing		
	}
}


