/*
  * hgcommons 7
 * Hammurapi Group Common Library 
 * Copyright (C) 2003  Hammurapi Group
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * URL: http://www.hammurapi.biz/hammurapi-biz/ef/xmenu/hammurapi-group/products/products/hgcommons/index.html
 * e-Mail: support@hammurapi.biz 
 */
package biz.hammurapi.cache;

import java.lang.ref.SoftReference;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

import biz.hammurapi.config.ConfigurationException;
import biz.hammurapi.util.Acceptor;


/**
 * Memory sensitive cache. Uses soft references to cache objects.
 * Doesn't create helper threads - housekeepeng happens on every 'cleanupInterval' get() invocation.
 * @author Pavel Vlasov
 * @version $Revision: 1.3 $
 */
public class PassiveMemoryCache extends AbstractProducer implements Cache {
	private Map cache=new HashMap();
	private Producer producer;
	private Cache fallBack;
	private int cleanupInterval;
    private int cleanupCounter;
	
	private class CacheReference extends SoftReference {
		private long expires;
		private long time;

		/**
		 * @param referent
		 * @param q
		 * @param expires
		 */
		public CacheReference(Object value, long time, long expires) {
			super(value);
			this.expires=expires;
			this.time=time;
		}		
	}
	
	/**
	 * Used as entry if fallBack cache is used.
	 * @author Pavel Vlasov
	 * @version $Revision: 1.3 $
	 */
	private class FallBackEntry {
		Object value;
		private long expires;
		private long time;
		private Object key;
		
		FallBackEntry(Object key, Object value, long time, long expires) {
			this.key=key;
			this.value=value;
			this.expires=expires;
			this.time=time;
		}
		
		protected void finalize() throws Throwable {
			if (fallBack.isActive() && (expires<=0 || System.currentTimeMillis()<expires)) {
				fallBack.put(key, value, time, expires);
			}
			
			super.finalize();
		}		
	}

	private void cleanup() {
		long now=System.currentTimeMillis();
		synchronized (cache) {
			Iterator it=cache.entrySet().iterator();
			while (it.hasNext()) {
				Map.Entry entry=(Map.Entry) it.next();
				CacheReference cacheReference = (CacheReference) entry.getValue();
				if (cacheReference.get()==null || (cacheReference.expires>0 && cacheReference.expires<now)) {
					it.remove();
				}
			}
		}
		cleanupCounter=0;
	}

    private void checkShutdown() {
    	if (shutDown) {
    		throw new IllegalStateException("Shut down");
    	}
    }
	
	/**
	 * 
	 */
	public PassiveMemoryCache(Producer producer, Cache fallBack, int cleanupInterval) {
		super();
		
		this.producer=producer;
		if (producer!=null) {
			producer.addCache(this);
		}
		
		this.fallBack=fallBack;
		if (fallBack!=null) {
			addCache(fallBack);
		}
		
		this.cleanupInterval=cleanupInterval;
	}
	
	public void put(Object key, Object value, long time, long expirationTime) {
		checkShutdown();
		synchronized (cache) {
			cache.put(key, new CacheReference(toValue(key, value, time, expirationTime), time, expirationTime));
		}
	}

	/**
	 * @param key
	 * @param value
	 * @param expirationTime
	 * @return
	 */
	private Object toValue(Object key, Object value, long time, long expirationTime) {
		return fallBack==null ? value : new FallBackEntry(key, value, time, expirationTime);
	}
	
	private Object fromValue(Object object) {
		if (object instanceof FallBackEntry) {
			return ((FallBackEntry) object).value;
		}
		return object;
	}
	
	

	public Entry get(Object key) {	    
		checkShutdown();
		if (cleanupInterval>0 && cleanupCounter++ > cleanupInterval) {
		    cleanup();
		}
		
		synchronized (cache) {
			CacheReference cr=(CacheReference) cache.get(key);
			long now = System.currentTimeMillis();
			boolean doFallBack=true;
			if (cr!=null) {
				if ((cr.expires>0 && cr.expires<now) || cr.get()==null) {
					cache.remove(key);		
					doFallBack=false;
				} else {
					final Object o=fromValue(cr.get());
					final long et=cr.expires;
					final long time=cr.time;
					return new Entry() {

						public long getExpirationTime() {
							return et;
						}

						public long getTime() {
							return time;
						}

						public Object get() {
							return o;
						}
						
					};
				}
			} 
			
			if (doFallBack && fallBack!=null) {
				Entry entry = fallBack.get(key);
				if (entry!=null) {
					cache.put(key, new CacheReference(toValue(key, entry.get(), entry.getTime(), entry.getExpirationTime()), entry.getTime(), entry.getExpirationTime()));
					return entry;
				}
			}
			
			if (key instanceof ProducingKey) {
				Entry entry = ((ProducingKey) key).get();
				cache.put(key, new CacheReference(toValue(key, entry.get(), entry.getTime(), entry.getExpirationTime()), entry.getTime(), entry.getExpirationTime()));
				return entry;
			}
			
			if (producer!=null) {
				Entry entry = producer.get(key);
				cache.put(key, new CacheReference(toValue(key, entry.get(), entry.getTime(), entry.getExpirationTime()), entry.getTime(), entry.getExpirationTime()));
				return entry;
			}
			
			return null;
		}
	}
	
	public void clear() {
		synchronized (cache) {
			cache.clear();
			
			if (fallBack!=null) {
				fallBack.clear();
			}
		}
	}
	
	public void remove(Object key) {
		synchronized (cache) {
			cache.remove(key);
		}
		onRemove(key);
	}

	public void remove(Acceptor acceptor) {
		synchronized (cache) {
			Iterator it=cache.keySet().iterator();
			while (it.hasNext()) {
				Object key=it.next();
				if (acceptor.accept(key)) {
					it.remove();
					onRemove(key);
				}
			}
		}
	}
	
	private boolean shutDown=false;
	
	public void stop() {
		if (!shutDown) {			
			cache.clear();
			shutDown=true;
		}
	}
	
	protected void finalize() throws Throwable {
		if (isActive()) {
			stop();
		}
		super.finalize();
	}

	public Set keySet() {
		HashSet ret = new HashSet();
		synchronized (cache) {
			ret.addAll(cache.keySet());
		}
		
		if (producer!=null) {
			Set pkeys=producer.keySet();
			if (pkeys!=null) {
				ret.addAll(pkeys);
			}
		}
		
		if (fallBack!=null) {
			Set fkeys=fallBack.keySet();
			if (fkeys!=null) {
				ret.addAll(fkeys);
			}
		}
		
		return ret;
	}

	/* (non-Javadoc)
	 * @see biz.hammurapi.cache.Cache#isActive()
	 */
	public boolean isActive() {
		return !shutDown;
	}

	public void start() throws ConfigurationException {
		// TODO Nothing
		
	}

	public void setOwner(Object owner) {
		// TODO Nothing
		
	}
}
