/*
  * hgcommons 7
 * Hammurapi Group Common Library 
 * Copyright (C) 2003  Hammurapi Group
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * URL: http://www.hammurapi.biz/hammurapi-biz/ef/xmenu/hammurapi-group/products/products/hgcommons/index.html
 * e-Mail: support@hammurapi.biz
 */
package biz.hammurapi.config;

import java.util.Hashtable;

import javax.naming.InitialContext;
import javax.naming.NamingException;

import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import biz.hammurapi.convert.CompositeConverter;
import biz.hammurapi.xml.dom.DOMUtils;
import biz.hammurapi.xml.dom.DOMUtils;

/** 
 * Binds object from jndi to pv. 
 * Configuration:<BR/> 
 * Attribute: <code>jndi-name</code><BR> 
 * Nested elements: <code>environment-property</code> with attribute <code>name</code>. 
 * Example:<PRE>&lt;topicConnectionFactory type="biz.hammurapi.config.JndiWrapper" jndi-name="TopicConnectionFactory"&gt; 

 &nbsp;&nbsp;&nbsp;&nbsp;&lt;environment-property name="java.naming.factory.initial"&gt;<I>factory class</I>&lt;/environment-property&gt;

 &nbsp;&nbsp;&nbsp;&nbsp;&lt;environment-property name="java.naming.provider.url"&gt;<I>provider urle</I>&lt;/environment-property&gt;

 &lt;/topicConnectionFactory&gt; 
 </PRE> 
 * Environment properties are passed to the constructor of initial JNDI context. 
 * If <code>jndi-name</code> attribute is not set then this class returns InitialContext as its master. 
 * 
 * If jndi-name is not set or master object is instance of javax.naming.Context then this component acts as a JNDI bridge.
 * @author Pavel Vlasov 
 * @revision $Revision$ 
 */
public class JndiWrapper extends ComponentBase implements DomConfigurable, Wrapper {
	
	private Hashtable ctxProps = new Hashtable();

	private Object master;

	private String name;

	public JndiWrapper() {
		super();
	}

	/** 
	 * Looks up master in JNDI. 
	 */
	public void start() throws ConfigurationException {
		try {
			InitialContext jndiContext = new InitialContext(ctxProps);
			master = name == null ? jndiContext : jndiContext.lookup(name);
		} catch (NamingException e) {
			throw new ConfigurationException(e);
		}
	}

	public void stop() throws ConfigurationException {
		// Nothing 
	}

	public void configure(Node configNode, Context context)	throws ConfigurationException {
		try {
			NodeList nl = DOMUtils.selectNodeList(configNode, "environment-property");
			for (int i=0, l=nl.getLength(); i<l; ++i) {
				Element e = (Element) nl.item(i);
				Object value = DOMUtils.getNonBlankElementText(e);
				if (e.hasAttribute("type")) {
					value = CompositeConverter.getDefaultConverter().convert(value, Class.forName(e.getAttribute("type")), false);
				}
				ctxProps.put(e.getAttribute("name"), value);
			}
			
			Element ce = (Element) configNode;
			if (ce.hasAttribute("jndi-name")) {
				name = ce.getAttribute("jndi-name");
			}
		} catch (Exception e) {
			throw new ConfigurationException(e);
		}
	}

	public Object getMaster() {
		return master;
	}

	protected Object getChild(String name) {
		if (master instanceof javax.naming.Context) {
			try {
				return ((javax.naming.Context) master).lookup(name);
			} catch (NamingException e) {
				throw new RuntimeConfigurationException("Lookup failed: "+e, e);
			}
		}
		return super.getChild(name);
	}
}
