package biz.hammurapi.config;

public abstract class ServiceBase extends ComponentBase implements Service {
	
	private String status;
	
	private ThreadLocal prevServiceTL = new ThreadLocal();
	
	ServiceBase prevService;
	ServiceBase nextService;
	
	/**
	 * Services shall implement initialization sequence in this method.
	 * @throws ConfigurationException
	 */
	protected abstract void startInternal() throws ConfigurationException;

	/**
	 * This implementation can be invoked multiple times. ServiceBase class maintains internal started flag to ensure
	 * that startInternal() method is invoked only once. The naming bus invokes this method every time before returning
	 * service from get() method.
	 */
	public final synchronized void start() throws ConfigurationException {
		if (!"Started".equals(status)) {
			if ("Starting".equals(status)) {
				throw new ConfigurationException("Circular dependency: start() method reentered by the same thread");
			}
			startInternal();
			
			prevService = (ServiceBase) prevServiceTL.get();
			prevServiceTL.set(this);
			if (prevService!=null) {
				prevService.nextService=this;
			}
		}
	}
	
	protected abstract void stopInternal() throws ConfigurationException;
	
	/**
	 *  Invokes stopInternal if all dependent services has been stopped.
	 *  Otherwise goes into "stoppable" state. When stops, also stops all
	 *  stoppable previous services.
	 */
	public final synchronized void stop() throws ConfigurationException {
		if (nextService==null || nextService.status==null) {
			stopInternal();
			status = null;
			if (prevService!=null && "Stoppable".equals(prevService.status)) {
				prevService.stop();
			}
		} else {
			status = "Stoppable";
		}
	}

}
