/*
  * hgcommons 7
 * Hammurapi Group Common Library 
 * Copyright (C) 2003  Hammurapi Group
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * URL: http://www.hammurapi.biz/hammurapi-biz/ef/xmenu/hammurapi-group/products/products/hgcommons/index.html
 * e-Mail: support@hammurapi.biz
 */
package biz.hammurapi.config;

import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.net.MalformedURLException;
import java.net.URL;
import java.text.MessageFormat;

import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.FactoryConfigurationError;
import javax.xml.parsers.ParserConfigurationException;

import org.w3c.dom.Document;
import org.xml.sax.SAXException;

/**
 * Helper class to read XML configuration from file, url or classloader
 * resource.
 * 
 * @author Pavel Vlasov
 * @version $Revision: 1.2 $
 */
public class XmlSource implements Parameterizable {

    private String prefix;

    private Class masterClass;

    private String defaultResourceExtension;

    /**
     * Constructor
     * 
     * @param prefix
     *            Prefix for configuration parameters. Parameter specifying
     *            configruration file would be &lt;prefix&gt;-file, url -
     *            &lt;prefix&gt;-url, resource - &lt;prefix&gt;-resource
     * @param masterClass
     *            Master class' classloader will be used to load resources and
     *            default resource would be masterClass name with
     *            defaultResourceExtension. If master class is null then this
     *            class classloader will be used and no default resource will be
     *            loaded.
     * @param defaultResourceExtension
     *            Extension for default resource. If it is null then default
     *            resource will not be loaded.
     */
    public XmlSource(String prefix, Class masterClass,
            String defaultResourceExtension) {
        this.prefix = prefix;
        this.masterClass = masterClass;
        this.defaultResourceExtension = defaultResourceExtension;
    }

    private static final String CONFIG_ERROR_MESSAGE = "'{0}-file', "
            + "'{0}-resource', and '{0}-url' parameters are mutually exclusive";

    private String resource;

    private String url;

    private String file;

    private Document document;

    /**
     * @return parsed XML or null if config not found.
     * @throws ConfigurationException
     */
    public Document getConfigDocument() throws ConfigurationException {
        if (document == null) {
            try {
                InputStream configStream = getStream();

                if (configStream!=null) { 
	                document = DocumentBuilderFactory.newInstance()
	                        .newDocumentBuilder().parse(configStream);
	                configStream.close();
                }
            } catch (IOException e) {
                throw new ConfigurationException("Cannot load document, "
                        + e.getMessage(), e);
            } catch (SAXException e) {
                throw new ConfigurationException("Cannot load document, "
                        + e.getMessage(), e);
            } catch (ParserConfigurationException e) {
                throw new ConfigurationException("Cannot load document, "
                        + e.getMessage(), e);
            } catch (FactoryConfigurationError e) {
                throw new ConfigurationException("Cannot load document, "
                        + e.getMessage(), e);
            }
        }

        return document;
    }

    /**
     * @return Stream
     * @throws FileNotFoundException
     * @throws IOException
     * @throws MalformedURLException
     * @throws ConfigurationException
     */
    public InputStream getStream() throws ConfigurationException {
        try {
	        Class clazz = masterClass==null ? getClass() : masterClass;
	        if (file != null) {
	            return new FileInputStream(file);
	        } else if (url != null) {
	            return new URL(url).openStream();
	        } else if (resource != null) {
	            
	            InputStream configStream = clazz.getClassLoader().getResourceAsStream(resource);
	            if (configStream == null) {
	                throw new ConfigurationException("Resource not found: " + resource);
	            }
	            
				return configStream;
	        } else {
	            if (defaultResourceExtension==null) {
	                return null;
	            }
	            
				String className = clazz.getName();
				int idx = className.lastIndexOf('.');
				String rName = (idx == -1 ? className : className
				        .substring(idx + 1))
				        + defaultResourceExtension;
				return clazz.getResourceAsStream(rName);
	        }
        } catch (IOException e) {
            throw new ConfigurationException("Cannot open stream, " + e.getMessage(), e);
        }
    }

    public boolean setParameter(String name, Object value)
            throws ConfigurationException {
        if ((prefix+"-file").equals(name)) {
            if (resource != null) {
                throw new ConfigurationException(MessageFormat.format(CONFIG_ERROR_MESSAGE, new Object[] {prefix}));
            }
            if (url != null) {
                throw new ConfigurationException(MessageFormat.format(CONFIG_ERROR_MESSAGE, new Object[] {prefix}));
            }
            file = value.toString();
        } else if ((prefix+"-resource").equals(name)) {
            if (file != null) {
                throw new ConfigurationException(MessageFormat.format(CONFIG_ERROR_MESSAGE, new Object[] {prefix}));
            }
            if (url != null) {
                throw new ConfigurationException(MessageFormat.format(CONFIG_ERROR_MESSAGE, new Object[] {prefix}));
            }
            resource = value.toString();
        } else if ((prefix+"-url").equals(name)) {
            if (file != null) {
                throw new ConfigurationException(MessageFormat.format(CONFIG_ERROR_MESSAGE, new Object[] {prefix}));
            }
            if (resource != null) {
                throw new ConfigurationException(MessageFormat.format(CONFIG_ERROR_MESSAGE, new Object[] {prefix}));
            }
            url = value.toString();
        } else {
            return false;
        }
        return true;
    }
}