/*
 * hgcommons 7
 * Hammurapi Group Common Library 
 * Copyright (C) 2003  Hammurapi Group
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * URL: http://www.hammurapi.biz/hammurapi-biz/ef/xmenu/hammurapi-group/products/products/hgcommons/index.html
 * e-Mail: support@hammurapi.biz
 */
package biz.hammurapi.convert;

import java.io.File;
import java.io.FileReader;
import java.io.Reader;
import java.lang.reflect.Constructor;
import java.lang.reflect.Method;
import java.math.BigDecimal;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.Iterator;
import java.util.List;

import biz.hammurapi.util.ClassHierarchyVisitable;
import biz.hammurapi.util.Visitor;
import biz.hammurapi.xml.dom.CompositeDomSerializer;
import biz.hammurapi.xml.dom.DomSerializable;


/**
 * @author Pavel Vlasov
 * @version $Revision: 1.7 $
 */
public class CompositeConverter {

    /**
     * Creates new composite converter populated by default with some generic converters.
     */
    public CompositeConverter() {
		_addConverter(Number.class, byte.class, "byteValue", null);
		_addConverter(Number.class, double.class, "doubleValue", null);
		_addConverter(Number.class, float.class, "floatValue", null);
		_addConverter(Number.class, int.class, "intValue", null);
		_addConverter(Number.class, long.class, "longValue", null);
		_addConverter(Number.class, short.class, "shortValue", null);
		
		_addConverter(Number.class, Byte.class, "byteValue", null);
		_addConverter(Number.class, Double.class, "doubleValue", null);
		_addConverter(Number.class, Float.class, "floatValue", null);
		_addConverter(Number.class, Integer.class, "intValue", null);
		_addConverter(Number.class, Long.class, "longValue", null);
		_addConverter(Number.class, Short.class, "shortValue", null);
		
		discoverConverter(String.class, Integer.class);
		discoverConverter(String.class, Long.class);
		discoverConverter(String.class, Double.class);
		discoverConverter(String.class, Float.class);
		discoverConverter(String.class, Byte.class);
		discoverConverter(String.class, Short.class);
		discoverConverter(String.class, BigDecimal.class);
		
		try {
			_addConverter(File.class, Reader.class, null, FileReader.class.getConstructor(new Class[] {File.class}));
		} catch (SecurityException e) {
			throw new ConversionException(e);
		} catch (NoSuchMethodException e) {
			throw new ConversionException(e);
		}
		
		addConverter(String.class, boolean.class, new Converter() {

			public Object convert(Object source) {
				if (source==null) {
					return Boolean.FALSE;
				}
				
				String str = ((String) source).trim();
				if (str.length()==0 || "no".equalsIgnoreCase(str) || "false".equalsIgnoreCase(str) || "0".equals(source)) {
					return Boolean.FALSE;
				} else if ("yes".equalsIgnoreCase(str) || "true".equalsIgnoreCase(str) || "1".equals(source)) {
					return Boolean.TRUE;
				} 
				
				throw new ConversionException("Cannot convert string '"+source+"' to boolean");
			}
			
		});
		
		addConverter(String.class, byte.class, new Converter() {

			public Object convert(Object source) {
				return new Byte((String) source);
			}
			
		});
		
		addConverter(String.class, double.class, new Converter() {

			public Object convert(Object source) {
				return new Double((String) source);
			}
			
		});
		
		addConverter(String.class, float.class, new Converter() {

			public Object convert(Object source) {
				return new Float((String) source);
			}
			
		});
		

		addConverter(String.class, int.class, new Converter() {

			public Object convert(Object source) {
				return new Integer((String) source);
			}
			
		});
		

		addConverter(String.class, long.class, new Converter() {

			public Object convert(Object source) {
				return new Long((String) source);
			}
			
		});
		

		addConverter(String.class, short.class, new Converter() {

			public Object convert(Object source) {
				return new Short((String) source);
			}
			
		});
		
		addConverter(Object.class, DomSerializable.class, new Converter() {

			public Object convert(Object source) {				
				return CompositeDomSerializer.getThreadInstance().toDomSerializable(source);
			}
			
		});
		
    }
    
    private boolean immutable;

    /**
     * @return true if no additional converters can be added to this converter.
     */
	public boolean isImmutable() {
		return immutable;
	}
	
	/**
	 * Makes converter immutable.
	 */
	public void setImmutable() {
		this.immutable = true;
	}
	
	private static CompositeConverter defaultConverter;
	
	static {
		defaultConverter=new CompositeConverter();
		defaultConverter.setImmutable();
	}
	
    public static CompositeConverter getDefaultConverter() {
       return defaultConverter; 
    }
    
	// TODO - chained conversions
	
	private static class ReflectionConverter implements Converter {
		Method accessor;
		Constructor constructor;
		
		/**
		 * @param accessor
		 * @param constructor
		 */
		ReflectionConverter(Method accessor, Constructor constructor) {
			super();
			this.accessor = accessor;
			this.constructor = constructor;
		}
		
		public Object convert(Object source) {
			try {
				Object param = accessor==null ? source : accessor.invoke(source, null);
				return constructor==null ? param : constructor.newInstance(new Object[] {param});
			} catch (Exception e) {
				throw new ConversionException("Cannot convert "+source.getClass().getName()+" to " + constructor.getDeclaringClass(), e);
			}
		}
	}
	
	private static class ConverterEntry {
		Class source;
		Class target;
        private Converter converter;
		
		/**
		 * @param accessor
		 * @param constructor
		 */
		ConverterEntry(Class source, Class target, Converter converter) {
			super();
			this.source = source;
			this.target = target;
			this.converter = converter;
		}
		
		boolean isCompatible(Class source, Class target) {
			return this.source.isAssignableFrom(source) && target.isAssignableFrom(this.target);
		}
		
		boolean isCompatible(ConverterEntry otherEntry) {
			return isCompatible(otherEntry.source, otherEntry.target);
		}		
	}
	
	private List converters=new ArrayList();
	
	/**
	 * Adds a converter which uses method of source object and constructor of target object to
	 * perform conversion. 
	 * @param source Source object.
	 * @param target Target class.
	 * @param accessor Method name to invoke on source to obtain intermediate object. Can be null.
	 * @param constructor Target constructor. Can be null.
	 */
	public Converter addConverter(Class source, Class target, String accessor, Constructor constructor) {
		if (immutable) {
			throw new ConversionException("Converter is immutable");
		}
		return _addConverter(source, target, accessor, constructor);
	}
	
	private Converter _addConverter(Class source, Class target, String accessor, Constructor constructor) {
		ReflectionConverter ret;
		try {
			ret = new ReflectionConverter(accessor==null ? null : source.getMethod(accessor, null), constructor);
		} catch (SecurityException e) {
			throw new ConversionException(e);
		} catch (NoSuchMethodException e) {
			throw new ConversionException(e);
		}
		_addConverter(source, target, ret);
		return ret;
	}
	
	/**
	 * Adds a converter from source object to target class.
	 * @param source
	 * @param target
	 * @param converter
	 */
	public void addConverter(Class source, Class target, Converter converter) {
		if (immutable) {
			throw new ConversionException("Converter is immutable");
		}
		_addConverter(source, target, converter);
	}
		
	private void _addConverter(final Class source, Class target, final Converter converter) {
		synchronized (converters) {
			new ClassHierarchyVisitable(target).accept(
					new Visitor() {

						public boolean visit(Object target) {
							Iterator it = converters.iterator();
							while (it.hasNext()) {
								ConverterEntry ce = (ConverterEntry) it.next();
								if (ce.source.equals(source) && ce.target.equals(target)) {
									return true; // Converter already exists.
								}
							}
							converters.add(new ConverterEntry(source, (Class) target, converter));							
							return true;
						}
						
					});
			
			Collections.sort(
					converters, 
					new Comparator() {

						public int compare(Object o1, Object o2) {
							if (o1==o2) {
								return 0;
							}
							
							if (o1 instanceof ConverterEntry && o2 instanceof ConverterEntry) {
							    ConverterEntry c1=(ConverterEntry) o1;
							    ConverterEntry c2=(ConverterEntry) o2;
								
								// More specific converters go first.
								if (c1.isCompatible(c2)) {
									return c2.isCompatible(c1) ? 0 : 1;
								}
								
								if (c2.isCompatible(c1)) {
									return -1;
								}
								
								return 0;
							}
														
							return o1.hashCode()-o2.hashCode();
						}
						
					});
			
		}
	}
	
	private Converter findConverter(Class source, Class target) {
		synchronized (converters) {
			Iterator it=converters.iterator();
			while (it.hasNext()) {
			    ConverterEntry ret=(ConverterEntry) it.next();
				if (ret.isCompatible(source, target)) {
					return ret.converter;
				}
			}
			return null;
		}
	}
	
    /**
     * Converts source object to target class instance
     * @param source Source object
     * @param target Target class. If target class is String then toString() is always used.
     * @param lenient When true null is returned if conversion cannot be performed, 
     * otherwise ConversionException is thrown
     * @return Instance of target class.
     * @throws ConversionException If lenient=false and conversion cannot be performed.
     */
	public Object convert(Object source, Class target, boolean lenient) {
		if (source==null) {
			return null;
		} else if (target.isInstance(source)) {
			return source;
		} else if (String.class.equals(target)) {
			return source.toString();
		} else {
			try {
				Converter converter=findConverter(source.getClass(), target);
				
				if (converter==null) {
					converter = discoverConverter(source.getClass(), target);
				}
				
				if (converter==null) {
					throw new ConversionException("No appropriate converter from "+source.getClass()+" to "+target);
				}
				
				return converter.convert(source);
			} catch (ConversionException e) {
				if (lenient) {
					return null;
				}
				
				throw e;
			}
		}
	}

	/**
	 * @param class1
	 * @param target
	 * @return Discovered converter or null.
	 */
	private Converter discoverConverter(Class source, Class target) {
		Converter ret=null;
		Constructor constructor=null;
		Constructor stringConstructor=null;
		for (int i=0, cc=target.getConstructors().length; i<cc; i++) {
			Constructor candidate = target.getConstructors()[i];
			if (candidate.getParameterTypes().length==1) {
				Converter c=_addConverter(candidate.getParameterTypes()[0], candidate.getDeclaringClass(), null, candidate);
				
				if (candidate.getParameterTypes()[0].isInstance(source)	&& (constructor==null || constructor.getParameterTypes()[0].isAssignableFrom(candidate.getParameterTypes()[0]))) {
					constructor=candidate;
					ret=c;
				}	
				
				if (stringConstructor==null	&& java.lang.String.class.equals(candidate.getParameterTypes()[0])) {
					stringConstructor=candidate;
				}								
			}			
		}
		
		if (ret!=null) {
			return ret;
		}
		
		if (stringConstructor!=null) {
			return _addConverter(Object.class, target, "toString", stringConstructor);
		}
		
		if (target.isInterface()) {
			ret = DuckConverterFactory.getConverter(source, target);
			if (ret==null) {
				ret=ContextConverterFactory.getConverter(source, target);
			}
		}
		
		return ret;
	}

    
}
