package biz.hammurapi.convert;

import java.lang.reflect.InvocationHandler;
import java.lang.reflect.Method;
import java.lang.reflect.Proxy;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Map;

import biz.hammurapi.config.Context;
import biz.hammurapi.util.ClassHierarchyVisitable;
import biz.hammurapi.util.Visitor;


/**
 * Creates converters which use "duck" typing.
 * @author Pavel
 *
 */
public class DuckConverterFactory {
	
	/**
	 * Returns source object unchanged
	 */
	private static Converter ZERO_CONVERTER = new Converter() {

		public Object convert(Object source) {			
			return source;
		}
		
	};
	
	/**
	 * Contains [sourceClass, targetClass] -> ProxyConverter(targetMethod -> sourceMethod) mapping.
	 */
	private static Map converterMap = new HashMap();
	
	private static class ProxyConverter implements Converter {

		/**
		 * Maps target methods to source methods. Unmapped methods
		 * are invoked directly (e.g. equals() or if method in both source and target belongs
		 * to the same interface (partial overlap)).
		 */
		private Map methodMap;
		
		private Class[] targetInterfaces;

		private ClassLoader classLoader;
		
		public ProxyConverter(Class targetInterface, Map methodMap, ClassLoader classLoader) {
			this.methodMap = methodMap;
			this.targetInterfaces = new Class[] {targetInterface};
			this.classLoader = classLoader;
		}
		
		public Object convert(final Object source) {
			
			InvocationHandler ih = new InvocationHandler() {

				public Object invoke(Object proxy, Method method, Object[] args) throws Throwable {
					Method sourceMethod = (Method) (methodMap==null ? null : methodMap.get(method));
					return sourceMethod==null ? method.invoke(source, args) : sourceMethod.invoke(source, args);
				}
				
			};
			
			return Proxy.newProxyInstance(classLoader, targetInterfaces, ih);
		}
		
	}

	/**
	 * @param sourceClass
	 * @param targetInterface
	 * @return Converter which can "duck-type" instance of source class to target interface or null if conversion is not possible.
	 * Methods are mapped as follows: return types shall be compatible, arguments shall be compatible, exception declarations are ignored.
	 */
	public static Converter getConverter(Class sourceClass, Class targetInterface) {
		if (targetInterface.isAssignableFrom(sourceClass)) {
			return ZERO_CONVERTER;
		}
		
		Collection key=new ArrayList();
		key.add(sourceClass);
		key.add(targetInterface);
		synchronized (converterMap) {
			Object value = converterMap.get(key);
			if (Boolean.FALSE.equals(value)) {
				return null;
			}
			
			if (value==null) {
				Method[] targetMethods = targetInterface.getMethods();
				Method[] sourceMethods = sourceClass.getMethods();
				Map methodMap = new HashMap();
				
				Z:
				for (int i=0, l=targetMethods.length; i<l; ++i) {
					if (Object.class.equals(targetMethods[i].getDeclaringClass())) { 
						continue;
					}					
					
					Method targetMethod = targetMethods[i];
					int candidateIndex=-1;
					Method candidateMethod = null;
					
					Y:
					for (int j=0, m=sourceMethods.length; j<m; ++j) {
						Method sourceMethod = sourceMethods[j];
						if (sourceMethod!=null) {
							if (targetMethod.equals(sourceMethod)) { // No mapping neccesary
								sourceMethods[j]=null; // Method is taken
								continue Z;
							}
							
							if (targetMethod.getName().equals(sourceMethod.getName()) && targetMethod.getParameterTypes().length == sourceMethod.getParameterTypes().length) {
								// Check for compatibility
								
								// Return type shall "widen"
								if (!targetMethod.getReturnType().isAssignableFrom(sourceMethod.getReturnType())) {
									continue;
								}
								
								Class[] targetParameterTypes = targetMethod.getParameterTypes();
								Class[] sourceParameterTypes = sourceMethod.getParameterTypes();
								for (int k=0, n=targetParameterTypes.length; k<n; ++k) {
									if (!sourceParameterTypes[k].isAssignableFrom(targetParameterTypes[k])) {
										continue Y;
									}
								}
								
								if (candidateMethod!=null) {
									int oldAffinity = classAffinity(candidateMethod.getReturnType(), targetMethod.getReturnType());
									int newAffinity = classAffinity(sourceMethod.getReturnType(), targetMethod.getReturnType());
									if (oldAffinity<newAffinity) {
										continue;
									}
									
									Class[] candidateParameterTypes = candidateMethod.getParameterTypes();
									for (int k=0, n=sourceParameterTypes.length; k<n; ++k) {
										oldAffinity = classAffinity(targetParameterTypes[k], candidateParameterTypes[k]);
										newAffinity = classAffinity(targetParameterTypes[k], sourceParameterTypes[k]);
										if (oldAffinity<newAffinity) {
											continue Y;
										}
									}
									
									sourceMethods[candidateIndex] = candidateMethod; // return method back
								}
								
								candidateMethod=sourceMethod;
								candidateIndex=j;
								sourceMethods[j]=null; // Method is taken
								// Compare with existing candidate
								
							}
						}
					}	
					
					if (candidateMethod==null) {
						converterMap.put(key, Boolean.FALSE); // To indicate that we tried and failed
						return null;
					}
					
					methodMap.put(targetMethod, candidateMethod);
				}
				
				ClassLoader cl = getChildClassLoader(sourceClass.getClassLoader(), targetInterface.getClassLoader());
				if (cl==null) {
					converterMap.put(key, Boolean.FALSE); // To indicate that we tried and failed
					return null;					
				}
				
				value = new ProxyConverter(targetInterface, methodMap.isEmpty() ? null : methodMap, cl);
				converterMap.put(key, value);
			}
			return (Converter) value;
		}
	}

	/**
	 * Calculates how close is subclass to superclass in class hierarchy.
	 * @param subClass
	 * @param superClass
	 * @return affinity, or Integer.MAX_VALUE if classes don't belong to the same class hierarchy.
	 */
	public static int classAffinity(Class subClass, final Class superClass) {
		if (superClass.isAssignableFrom(subClass)) {
			final int[] caffinity={0};
			new ClassHierarchyVisitable(subClass).accept(new Visitor() {
	
				public boolean visit(Object target) {
					if (target.equals(superClass)) {
						return false;
					} 
					
					caffinity[0]++;
					return true;
				}
				
			});
			return caffinity[0];
		}
		
		return Integer.MAX_VALUE;		
	}
	
	/**
	 * @param cl1
	 * @param cl2
	 * @return Child classloader or null if classloaders are not related
	 */
	private static ClassLoader getChildClassLoader(ClassLoader cl1, ClassLoader cl2) {
		if (cl1==null) {
			return cl2;
		}
		if (cl2==null) {
			return cl1;
		}
		for (ClassLoader cl = cl1; cl!=null && cl!=cl.getParent(); cl=cl.getParent()) {
			if (cl2.equals(cl)) {
				return cl1;
			}
		}
		for (ClassLoader cl = cl2; cl!=null && cl!=cl.getParent(); cl=cl.getParent()) {
			if (cl1.equals(cl)) {
				return cl2;
			}
		}
		return null;
	}
	
	public static void main(String[] args) {
		Converter converter = getConverter(Map.class, Context.class);
		System.out.println(converter);
		
		Map testMap = new HashMap();
		testMap.put("odin", "dva");
		
		System.out.println(((Context) converter.convert(testMap)).get("odin"));
		
		Converter c2 = getConverter(Map.class, String.class);
		System.out.println(c2);
		
		Context ctx = (Context) CompositeConverter.getDefaultConverter().convert(testMap, Context.class, false);
		System.out.println(ctx.get("odin"));
	}
		
}
