package biz.hammurapi.jms;

import javax.jms.Connection;
import javax.jms.ConnectionFactory;
import javax.jms.Destination;
import javax.jms.JMSException;
import javax.jms.Message;
import javax.jms.MessageConsumer;
import javax.jms.MessageProducer;
import javax.jms.Session;
import javax.jms.TemporaryQueue;
import javax.naming.Context;
import javax.naming.NamingException;

import biz.hammurapi.config.ConfigurationException;
import biz.hammurapi.config.RuntimeConfigurationException;

/**
 * Base class for JMS adapters which operate on two queue pairs.
 * 
 * @author Pavel Vlasov
 */
public abstract class Adapter extends MessageProcessor {

	private static final String BACK_END_INITIAL_CONTEXT_PREFIX = "backEndInitialContext/";

	protected Context backEndInitialContext;

	private String backEndConnectionFactoryName;

	private String backEndDestinationName;

	private String backEndReplyDestinationName;

	private String replyDestinationName;

	/**
	 * Initial context to look-up back-end destinations and connection factory.
	 * If back-end initial context is not set then back-end JMS objects are
	 * looked up in the "front-end" initial context.
	 * 
	 * @param initialContext
	 */
	public void setBackEndInitialContext(Context initialContext) {
		this.initialContext = initialContext;
	}

	/**
	 * Back-end request destination (queue or topic) name.
	 * 
	 * @param destinationName
	 */
	public void setBackEndDestination(String destinationName) {
		this.backEndDestinationName = destinationName;
	}

	/**
	 * Back-end reply destination (queue or topic) name.
	 * 
	 * @param destinationName
	 */
	public void setBackEndReplyDestination(String destinationName) {
		this.backEndReplyDestinationName = destinationName;
	}

	/**
	 * Front-end reply destination (queue or topic) name.
	 * 
	 * @param destinationName
	 */
	public void setReplyDestination(String destinationName) {
		this.replyDestinationName = destinationName;
	}

	/**
	 * Back-end JMS connection name. It it is not set then the "front-end"
	 * connection factory is used.
	 * 
	 * @param connectionName
	 */
	public void setBackEndConnectionFactory(String connectionFactoryName) {
		this.backEndConnectionFactoryName = connectionFactoryName;
	}

	protected Connection backEndConnection;

	private Destination backEndDestination;

	private Destination backEndReplyDestination;

	private Destination replyDestination;

	protected boolean isBackEndTransacted;

	protected int backEndAcknowledgeMode = Session.AUTO_ACKNOWLEDGE;
	
	protected long replyTimeToLive;
	
	protected long backEndRequestTimeToLive;
	
	protected long backEndTimeout;
	
	/**
	 * Timeout for backend reply. 
	 * @param backEndTimeout
	 */
	public void setBackEndTimeout(long backEndTimeout) {
		this.backEndTimeout = backEndTimeout;
	}
	
	/**
	 * Time to live for reply.
	 * @param replyTimeToLive
	 */
	public void setReplyTimeToLive(long replyTimeToLive) {
		this.replyTimeToLive = replyTimeToLive;
	}
	
	/**
	 * Time to live for back-end request
	 * @param backEndRequestTimeToLive
	 */
	public void setBackEndRequestTimeToLive(long backEndRequestTimeToLive) {
		this.backEndRequestTimeToLive = backEndRequestTimeToLive;
	}

	/**
	 * Back-end acknoledge mode. Valid values: AUTO (default), CLIENT, DUPS_OK.
	 * 
	 * @param acknowledgeModeName
	 */
	public void setBackEndAcknowledgeMode(String acknowledgeModeName) {
		try {
			this.backEndAcknowledgeMode = Session.class.getField(acknowledgeModeName + "_ACKNOWLEDGE").getInt(null);
		} catch (Exception e) {
			throw new IllegalArgumentException("Invalid back-end acknowledge mode '" + acknowledgeModeName + "': " + e);
		}
	}

	/**
	 * Explicitly sets back-end connection. This method is useful when several
	 * components share one connection.
	 * 
	 * @param connection
	 */
	public void setBackEndConnection(Connection connection) {
		this.backEndConnection = connection;
	}

	/**
	 * Transactional attribute of back-end JMS Session.
	 * 
	 * @param isTransacted
	 */
	public void setBackEndTransacted(boolean isTransacted) {
		this.isBackEndTransacted = isTransacted;
	}

	private String backEndUser;

	private String backEndPwd;

	protected ConnectionFactory backEndConnectionFactory;

	/**
	 * Back-end JMS Connection user name. Optional.
	 * 
	 * @param user
	 */
	public void setBackEndUser(String user) {
		this.backEndUser = user;
	}

	/**
	 * Back-end JMS Connection password. Optional.
	 * 
	 * @param pwd
	 */
	public void setPassword(String pwd) {
		this.backEndPwd = pwd;
	}

	public void start() throws ConfigurationException {
		super.start();
		try {
			logger.info(this, "Starting ...");

			if (backEndConnectionFactoryName == null) {
				backEndConnectionFactory = connectionFactory;
				backEndConnection = connection;
			} else {
				backEndConnectionFactory = (ConnectionFactory) initialContext.lookup(backEndConnectionFactoryName);
				if (backEndUser == null) {
					backEndConnection = backEndConnectionFactory.createConnection();
				} else {
					backEndConnection = backEndConnectionFactory.createConnection(backEndUser, backEndPwd);
				}
				backEndConnection.start();
			}

			if (backEndInitialContext == null) {
				backEndInitialContext = initialContext;
			}

			backEndDestination = (Destination) backEndInitialContext.lookup(backEndDestinationName);
			backEndReplyDestination = (Destination) backEndInitialContext.lookup(backEndReplyDestinationName);

			if (replyDestinationName!=null) {
				replyDestination = (Destination) initialContext.lookup(replyDestinationName);
			}

			logger.info(this, "Started");
		} catch (Exception e) {
			if (e instanceof ConfigurationException) {
				throw (ConfigurationException) e;
			}

			throw new ConfigurationException("Could not start message processor: " + e, e);
		}
	}

	/**
	 * Processes request message by passing to processRequest() and
	 * processResponse() methods.
	 * 
	 * @param request
	 *            Request message
	 * @param session
	 *            Session if message is processed in the message listener thread
	 *            (worker is null or cannot process jobs), null otherwise.
	 */
	protected void processMessage(Message request, Session session) {
		try {
			Session frontEndSession = session == null ? borrowSession() : session;
			Destination frontEndReplyDestination = request.getJMSReplyTo() == null ? replyDestination : request.getJMSReplyTo();
			Message frontEndReply = null;
			try {
				Session backEndSession = connection == backEndConnection ? session : borrowBackEndSession();
				try {
					long start = System.currentTimeMillis();
					Message backEndRequest = processRequest(request, frontEndSession, backEndSession);
					long afterProcessRequest = System.currentTimeMillis();
					addMeasurement("process-request", afterProcessRequest-start, afterProcessRequest);
					if (backEndRequest!=null) {					
						Destination brd = backEndReplyDestination;
						if (brd==null) {
							brd = backEndSession.createTemporaryQueue();
						}
						backEndRequest.setJMSReplyTo(brd);
			            MessageProducer backEndProducer = backEndSession.createProducer(backEndDestination); 
			
			            String messageSelector;
			
			            try {
			            	backEndProducer.send(backEndRequest); 
			                messageSelector = messageSelector(backEndRequest);
			            } finally {
			            	backEndProducer.close();
			            }
			
			            Message backEndReply;
						MessageConsumer consumer;
						if (brd instanceof TemporaryQueue) {
				            logger.debug(this, "Waiting on "+brd);            
							consumer = session.createConsumer(brd);
						} else {
				            logger.debug(this, "Waiting for "+messageSelector+" on "+brd);	            
							consumer = session.createConsumer(brd, messageSelector);
						}
							
						try {
							backEndReply = backEndTimeout == 0 ? consumer.receive() : consumer.receive(backEndTimeout);
			            } finally {
			                consumer.close();
							if (brd instanceof TemporaryQueue) {
								((TemporaryQueue) brd).delete();
							}
			            }
			            long afterBackEndReply = System.currentTimeMillis();
			            addMeasurement("back-end", afterBackEndReply-afterProcessRequest, afterBackEndReply);
			
						if (frontEndReplyDestination==null) {
							logger.warn(this, "Unable to send reply to front-end - reply destination is null");
						} else {
							frontEndReply = processReply(backEndReply, request, frontEndSession, backEndSession);
							long afterProcessReply = System.currentTimeMillis();
							addMeasurement("process-reply", afterProcessReply-afterBackEndReply, afterProcessReply);							
						}
					}
				} catch (Exception e) {
					frontEndReply = handleException(e, frontEndSession);
				} finally {
					if (backEndSession!=frontEndSession) {
						releaseBackEndSession(backEndSession);
					}
				}
				
				if (frontEndReply!=null) {
					if (frontEndReply.getJMSCorrelationID()==null) {
						frontEndReply.setJMSCorrelationID(request.getJMSCorrelationID()==null ? request.getJMSMessageID() : request.getJMSCorrelationID());
					}
					MessageProducer replyProducer = frontEndSession.createProducer(frontEndReplyDestination);
					try {
						if (replyTimeToLive!=0) {
							replyProducer.setTimeToLive(replyTimeToLive);	
						}
						replyProducer.send(frontEndReply);
					} finally {
						replyProducer.close();
					}
				}
			} finally {
				if (session == null) {
					releaseSession(frontEndSession);
				}
			}
		} catch (Exception e) {
			logger.error(e, e.toString());
		}
	}

	/**
	 * Processes front-end request message and builds back-end request message.
	 * 
	 * @param request
	 *            Request from front-end
	 * @param frontEndSession
	 *            Front-end JMS session
	 * @param backEndSession
	 *            Back-end JMS session. Equals to the front-end session if back
	 *            end connection is not set and the front-end connection is used
	 *            to connect to both front and back end.
	 * @return Message to be sent to the back-end request queue. Null if no message shall be sent (e.g. in the case of exception).
	 */
	protected abstract Message processRequest(Message frontEndRequest, Session frontEndSession, Session backEndSession);

	/**
	 * Processes back-end reply and forms reply to front-end. This method is
	 * responsible for setting correlation ID.
	 * 
	 * @param backEndReply
	 *            Reply message from back-end. Null if request timed out.
	 * @param frontEndRequest
	 *            Original request from front end.
	 * @param frontEndSession
	 *            Front-end JMS session.
	 * @param backEndSession
	 *            Back-end JMS session. Equals to the front-end session if back
	 *            end connection is not set and the front-end connection is used
	 *            to connect to both front and back end.
	 * @return Message to be sent to the front-end reply queue. Null if no message shall be sent (e.g. in the case of exception).
	 */
	protected abstract Message processReply(Message backEndReply, Message frontEndRequest, Session frontEndSession, Session backEndSession);
	
	/**
	 * Method to report exceptions to fron-end (excluding JMS exceptions in front-end JMS objects). This implementation returns null.
	 * @param exception Exception to report
	 * @param frontEndSession Front-end JMS session
	 * @return Message to send to front-end to reply queue. Null if no message shall be sent.
	 */
	protected Message handleException(Exception exception, Session frontEndSession) {
		return null;
	}

	/**
	 * Extracts message selector for reply back-end message from the back-end
	 * request message. This implementation returns
	 * <code>"JMSCorrelationID='" +backEndRequest.getJMSMessageID()+"'"</code>
	 * 
	 * @param backEndRequest
	 * @return Message selector to retrieve back-end reply.
	 * @throws JMSException
	 */
	protected String messageSelector(Message backEndRequest) throws JMSException {
		return "JMSCorrelationID='" + backEndRequest.getJMSMessageID() + "'";
	}

	/**
	 * Stops worker (thread pool), if any, and connection.
	 */
	public void stop() throws ConfigurationException {
		try {
			logger.info(this, "Stopping ...");

			if (backEndConnection != null && backEndConnection != connection) {
				backEndConnection.close();
			}

			logger.info(this, "Stopped");
		} catch (Exception e) {
			if (e instanceof ConfigurationException) {
				throw (ConfigurationException) e;
			}

			throw new ConfigurationException("Could not stop message processor");
		}
	}

	/**
	 * This implementation simply creates a new session. Subclasses can override
	 * this method to implement session pooling.
	 * 
	 * @return
	 * @throws JMSException
	 */
	protected Session borrowBackEndSession() throws JMSException {
		return backEndConnection.createSession(isBackEndTransacted,	backEndAcknowledgeMode);
	}

	/**
	 * This implementation simply closes the session. Subclasses can override
	 * this method to implement session pooling.
	 * 
	 * @throws JMSException
	 */
	protected void releaseBackEndSession(Session session) throws JMSException {
		session.close();
	}

	/**
	 * Provides access to back-end destination, reply destinations, initial
	 * context, and connection in addition to objects accessible through
	 * superclass getChild() method. Bridges to back-end initial context. For
	 * names in a form backEndInitialContext/&lt;name&gt; the name is looked up
	 * in initial context.
	 */
	protected Object getChild(String name) {
		if ("replyDestination".equals(name)) {
			return replyDestination;
		}

		if ("backEndReplyDestination".equals(name)) {
			return backEndReplyDestination;
		}

		if ("backEndDestination".equals(name)) {
			return backEndDestination;
		}

		if ("backEndInitialContext".equals(name)) {
			return backEndInitialContext;
		}

		if (name != null && name.startsWith(BACK_END_INITIAL_CONTEXT_PREFIX)) {
			String jndiName = name.substring(BACK_END_INITIAL_CONTEXT_PREFIX.length());
			try {
				return backEndInitialContext.lookup(jndiName);
			} catch (NamingException e) {
				throw new RuntimeConfigurationException("Lookup in back-end inital context failed for JNDI name '" + jndiName + "':" + e, e);
			}
		}

		if ("backEndConnection".equals(name)) {
			return backEndConnection;
		}

		if ("backEndConnectionFactory".equals(name)) {
			return backEndConnectionFactory;
		}

		return super.getChild(name);
	}

}
