/*
 * hgee 2
 * Hammurapi Group Enterprise Extensions 
 * Copyright (C) 2006  Hammurapi Group
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * URL: http://www.hammurapi.biz/products/enterpriseextensions/index.html
 * e-Mail: support@hammurapi.biz
 */
package biz.hammurapi.jms;

import javax.jms.Connection;
import javax.jms.ConnectionFactory;
import javax.jms.JMSException;

import org.w3c.dom.Element;
import org.w3c.dom.Node;

import biz.hammurapi.config.ConfigurationException;
import biz.hammurapi.config.Context;
import biz.hammurapi.config.JndiWrapper;
import biz.hammurapi.xml.dom.AbstractDomObject;


/**
 * This class reads configuration from XML, looks up JMS connection factory in JNDI and obtains connection from the factory.
 * Configuration:<BR/>
 * Attribute: <code>jndi-name</code><BR>
 * Nested elements: <code>environment-property</code> with attribute <code>name</code>, <code>user</code> and <code>password</code> elements for authentication.
 * Example:<PRE>&lt;queueConnection type="biz.hammurapi.jms.ConnectionWrapper" jndi-name="ConnectionFactory"&gt; 
	&nbsp;&nbsp;&nbsp;&nbsp;&lt;environment-property name="java.naming.factory.initial"&gt;<I>factory class</I>&lt;/environment-property&gt;
	&nbsp;&nbsp;&nbsp;&nbsp;&lt;environment-property name="java.naming.provider.url"&gt;<I>provider url</I>&lt;/environment-property&gt;
	&nbsp;&nbsp;&nbsp;&nbsp;&lt;user&gt;usr&lt;/user&gt;
	&nbsp;&nbsp;&nbsp;&nbsp;&lt;password&gt;pwd&lt;/password&gt;
	&lt;/queueConnection&gt;
	</PRE>
 * Environment properties are passed to the constructor of initial JNDI context.
 * @author Pavel Vlasov
 * @revision $Revision$
 */
public class ConnectionWrapper extends JndiWrapper {

	private Connection connection;

	/**
	 * Returns queue connection.
	 */
	public Object getMaster() {
		return connection;
	}

	/**
	 * Obstains connection factory from JNDI, gets connection and starts it.
	 */
	public void start() throws ConfigurationException {		
		super.start();
		try {
			ConnectionFactory connectionFactory = (ConnectionFactory) super.getMaster();
				
			if (user==null) {
				connection = connectionFactory.createConnection(); 
			} else {
				connection = connectionFactory.createConnection(user, password);
			}
				
			connection.start();
		} catch (JMSException e) {
			throw new ConfigurationException("Could not obtain or start connection: "+e, e);
		}
	}

	/**
	 * Stops connection.
	 */
	public void stop() throws ConfigurationException {
		if (connection!=null) {
			try {
				connection.stop();
				connection=null;
			} catch (JMSException e) {
				throw new ConfigurationException("Could not stop connection: "+e, e);
			}
		}
	}

	public void setOwner(Object owner) {
		// Nothing
	}
	
	private String user;
	private String password;

	public void configure(Node configNode, Context context) throws ConfigurationException {
		super.configure(configNode, context);
		
		try {
			user = AbstractDomObject.getElementText((Element) configNode, "user");
			if (user!=null) {
				password=AbstractDomObject.getElementText((Element) configNode, "password");
			}
		} catch (Exception e) {
			throw new ConfigurationException("Could not read user name and password: "+e, e);
		}		
	}

}
