/*
 * hgee 2
 * Hammurapi Group Enterprise Extensions 
 * Copyright (C) 2006  Hammurapi Group
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * URL: http://www.hammurapi.biz/products/enterpriseextensions/index.html
 * e-Mail: support@hammurapi.biz
 */
package biz.hammurapi.jms;

import java.io.IOException;
import java.io.Serializable;

import javax.jms.JMSException;
import javax.jms.Message;
import javax.jms.ObjectMessage;
import javax.jms.Queue;
import javax.jms.QueueConnection;
import javax.jms.QueueConnectionFactory;
import javax.jms.QueueSender;
import javax.jms.QueueSession;
import javax.jms.Session;
import javax.jms.TextMessage;
import javax.xml.parsers.FactoryConfigurationError;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.TransformerException;

import org.w3c.dom.Element;
import org.w3c.dom.Node;

import biz.hammurapi.config.ComponentBase;
import biz.hammurapi.config.ConfigurationException;
import biz.hammurapi.config.Context;
import biz.hammurapi.config.DomConfigurable;
import biz.hammurapi.util.Worker;
import biz.hammurapi.xml.dom.DOMUtils;
import biz.hammurapi.xml.dom.DomSerializable;


/**
 * Posts serializable and dom-serializable jobs to
 * JMS queue.
 * @author Pavel Vlasov
 * @revision $Revision$
 */
public class JmsWorker extends ComponentBase implements Worker, DomConfigurable {

	private String queueName;
	private String connectionFactoryName;
	private String user;
	private String password;
	
	private Queue queue;
	private QueueConnection connection;
	
	/**
	 * Jobs can implement this interface to set
	 * JMS message headers 
	 * @author Pavel Vlasov
	 * @revision $Revision$
	 */
	public interface Selectable {
		void setMessageHeaders(Message message);
	}

	public JmsWorker() {
		super();
	}

	public boolean post(Runnable job) {
		try {
	        QueueSession session = connection.createQueueSession(false, Session.AUTO_ACKNOWLEDGE);
	        try {
	        	if (job instanceof JmsSerializable) {
	        		sendMessage(session, ((JmsSerializable) job).toMessage(session));
	        		return true;
	        	} else if (job instanceof DomSerializable) {
			        TextMessage message = session.createTextMessage();
			        if (job instanceof Selectable) {
			        	((Selectable) job).setMessageHeaders(message);
			        }
			        message.setStringProperty("type", job.getClass().getName());
			        message.setText(DOMUtils.toXmlString(job, "job"));
			        
			        sendMessage(session, message);
			        return true;
	        	} else if (job instanceof Serializable) {
			        ObjectMessage message = session.createObjectMessage();
			        if (job instanceof Selectable) {
			        	((Selectable) job).setMessageHeaders(message);
			        }
			        message.setStringProperty("type", job.getClass().getName());
			        message.setObject((Serializable) job);
			
			        sendMessage(session, message);
			        return true;
	        	} 
	    		return false;
			} finally {
	        	session.close();
	        }
		} catch (JMSException e) {
			throw new RuntimeException("Could not post job to queue: "+e, e);
        } catch (IOException e) {
			throw new RuntimeException("Could not serialize job to XML: "+e, e);
		} catch (TransformerException e) {
			throw new RuntimeException("Could not serialize job to XML: "+e, e);
		} catch (ParserConfigurationException e) {
			throw new RuntimeException("Could not serialize job to XML: "+e, e);
		} catch (FactoryConfigurationError e) {
			throw new RuntimeException("Could not serialize job to XML: "+e, e);
		}
	}

	/**
	 * @param session
	 * @param message
	 * @throws JMSException
	 */
	private void sendMessage(QueueSession session, Message message) throws JMSException {
		QueueSender sender = session.createSender(queue);
		try {
			sender.send(message);
		} finally {
			sender.close();
		}
	}

	public void start() throws ConfigurationException {
		try {
			QueueConnectionFactory connectionFactory = (QueueConnectionFactory) get(connectionFactoryName);
			queue = (Queue) get(queueName);

			if (user == null) {
				connection = connectionFactory.createQueueConnection();
			} else {
				connection = connectionFactory.createQueueConnection(user, password);
			}
		} catch (JMSException e) {
			throw new ConfigurationException(e);
		}
	}

	public void stop() throws ConfigurationException {
		if (connection!=null) {
			try {
				connection.close();
			} catch (JMSException e) {
				throw new ConfigurationException(e);
			}
		}
	}

	public void configure(Node configNode, Context context) throws ConfigurationException {
		Element ce = (Element) configNode;
		queueName=ce.getAttribute("queue");
		connectionFactoryName=ce.getAttribute("connection-factory");
		if (ce.hasAttribute("user")) {
			user=ce.getAttribute("user");
			password=ce.getAttribute("password");
		}
	}
}
