package biz.hammurapi.jms;

import javax.jms.Connection;
import javax.jms.ConnectionFactory;
import javax.jms.Destination;
import javax.jms.ExceptionListener;
import javax.jms.JMSException;
import javax.jms.Message;
import javax.jms.MessageListener;
import javax.jms.Session;
import javax.naming.Context;
import javax.naming.NamingException;

import biz.hammurapi.config.Component;
import biz.hammurapi.config.ComponentBase;
import biz.hammurapi.config.ConfigurationException;
import biz.hammurapi.config.RestartCommand;
import biz.hammurapi.config.Restartable;
import biz.hammurapi.config.RuntimeConfigurationException;
import biz.hammurapi.logging.ConsoleLogger;
import biz.hammurapi.logging.Logger;
import biz.hammurapi.metrics.MeasurementCollector;
import biz.hammurapi.metrics.MeasurementConsumer;
import biz.hammurapi.util.Worker;

/**
 * Base class for JMS message processors.
 * @author Pavel Vlasov
 */
public abstract class MessageProcessor extends ComponentBase implements Restartable {
	
	private static final String INITIAL_CONTEXT_PREFIX = "initialContext/";
	protected Logger logger = new ConsoleLogger(ConsoleLogger.INFO);
	
	/**
	 * Sets logger.
	 * @param logger
	 */
	public void setLogger(Logger logger) {
		this.logger = logger;
	}
		
	protected Context initialContext;
	private String connectionFactoryName;
	private String destinationName;
	private int listeners = Runtime.getRuntime().availableProcessors();
	
	/**
	 * Number of message listening threads. Defaults to number of processes available to JVM.
	 * @param listeners
	 */
	public void setListeners(int listeners) {
		this.listeners = listeners;
	}
	
	public void setInitialContext(Context initialContext) {
		this.initialContext = initialContext;
	}
	
	/**
	 * Request destination (queue or topic) name
	 * @param destinationName
	 */
	public void setDestination(String destinationName) {
		this.destinationName = destinationName;
	}
	
	/**
	 * JMS connection name
	 * @param connectionName
	 */
	public void setConnectionFactory(String connectionFactoryName) {
		this.connectionFactoryName = connectionFactoryName;
	}
	
	/**
	 * Worker to process requests. It is optional. If worker is not set or cannot process requests,
	 * requests are processed in the message listener thread.
	 * @param worker
	 */
	public void setWorker(Worker worker) {
		this.worker = worker;
	}
	
	protected Connection connection;
	private Destination destination;
	protected boolean isTransacted;
	protected int acknowledgeMode = Session.AUTO_ACKNOWLEDGE;

	/**
	 * Acknoledge mode. Valid values: AUTO (default), CLIENT, DUPS_OK.
	 * @param acknowledgeModeName
	 */
	public void setAcknowledgeMode(String acknowledgeModeName) {
		try {
			this.acknowledgeMode = Session.class.getField(acknowledgeModeName+"_ACKNOWLEDGE").getInt(null);
		} catch (Exception e) {
			throw new IllegalArgumentException("Invalid acknowledge mode '"+acknowledgeModeName+"': "+e);
		}
	}
	
	/**
	 * Explicitly sets connection. This method is useful when several components share one connection.
	 * @param connection
	 */
	public void setConnection(Connection connection) {
		this.connection = connection;
	}
	
	/**
	 * Transactional attribute of JMS Session.
	 * @param isTransacted
	 */
	public void setTransacted(boolean isTransacted) {
		this.isTransacted = isTransacted;
	}
	
	protected Worker worker;	
	
	private String user;
	private String pwd;
	protected ConnectionFactory connectionFactory;
	
	/** 
	 * JMS Connection user name. Optional.
	 * @param user
	 */
	public void setUser(String user) {
		this.user = user;
	}
	
	/**
	 * JMS Connection password. Optional.
	 * @param pwd
	 */
	public void setPassword(String pwd) {
		this.pwd = pwd;
	}

	/**
	 * Message listener class.
	 * @author Pavel Vlasov
	 */
	private class ProcessingMessageListener implements MessageListener {
		
		private Session session;

		public ProcessingMessageListener(Session session) {
			this.session = session;
		}		
		
		public synchronized void onMessage(final Message request) {
			logger.debug(this, "Message received");
			addMeasurement("receive", 1, System.currentTimeMillis());
			
			if (worker==null) {
				_processMessage(request, session);
			} else {
				Runnable job = new Runnable() {

					public void run() {
						_processMessage(request, null);
					}
					
				};
				
				// Try to post job to worker. process in the current thread if worker doesn't accept it.
				if (!worker.post(job)) {
					_processMessage(request, session);
				}
			}						
		}
				
	};

	public void start() throws ConfigurationException {
		try {			
			logger.info(this, "Starting ...");
			
			connectionFactory = (ConnectionFactory) initialContext.lookup(connectionFactoryName);
			if (user==null) {
				connection = connectionFactory.createConnection();
			} else {
				connection = connectionFactory.createConnection(user, pwd);
			}
			connection.start();
			
			final ExceptionListener pel = connection.getExceptionListener();
			connection.setExceptionListener(new ExceptionListener() {

				public void onException(JMSException e) {
					if (pel!=null) {
						pel.onException(e);
					}
										
					logger.error(MessageProcessor.this, "JMS connection exception: " + e.toString());
					
					if (restartCommand!=null) {
						try {
							try {
								connection.setExceptionListener(pel); // To avoid double firing if there is another exception in stop().
								stop();
							} catch (Exception ex) {
								ex.printStackTrace();
								logger.error(MessageProcessor.this, "Cannot stop the processor: "+ex);
							}
						} finally {
							new Thread(restartCommand, "Connection restart thread "+restartCommand.getAttempt()).start();
						}
					}
					
				}
				
			});
			
			destination = (Destination) initialContext.lookup(destinationName);
			
			if (worker instanceof Component) {
				((Component) worker).start();
			}
			
			if (worker instanceof MeasurementCollector) {
				((MeasurementCollector) worker).setMeasurementConsumer(
						new MeasurementConsumer() {

							public void addMeasurement(String name, double value, long time) {
								MessageProcessor.this.addMeasurement("worker."+name, value, time);								
							}
							
						});
			}			
			
			for (int i=0; i<listeners; ++i) {
				final Session session = borrowSession();
				session.createConsumer(destination, messageSelector, noLocal).setMessageListener(new ProcessingMessageListener(session));
			}

			logger.info(this, "Started");
		} catch (Exception e) {
			if (e instanceof ConfigurationException) {
				throw (ConfigurationException) e;
			}
			
			throw new ConfigurationException("Could not start message processor: "+e, e);
		}		
	}
	
	/**
	 * Processes request message
	 * @param request Request message
	 * @param session Session if message is processed in the message listener thread (worker is null or cannot process jobs), null otherwise.
	 */
	protected abstract void processMessage(Message request, Session session);
	
	protected void _processMessage(Message request, Session session) {
		long start = System.currentTimeMillis();
		try {
			processMessage(request, session);
		} finally {
			long now = System.currentTimeMillis();
			addMeasurement("process", now - start, now);
		}
	}
	
	/**
	 * Stops worker (thread pool), if any, and connection.
	 */
	public void stop() throws ConfigurationException {
		try {			
			logger.info(this, "Stopping ...");
			
			// Stop worker
			if (worker instanceof Component) {
				((Component) worker).stop();
			}
						
			// No need to close individual sessions.
			if (connection!=null) {
				connection.close();
			}		
			
			logger.info(this, "Stopped");
		} catch (Exception e) {
			if (e instanceof ConfigurationException) {
				throw (ConfigurationException) e;
			}
			
			throw new ConfigurationException("Could not stop message processor");
		}
	}
	
	/**
	 * This implementation simply creates a new session.
	 * Subclasses can override this method to implement session pooling.
	 * @return
	 * @throws JMSException 
	 */
	protected Session borrowSession() throws JMSException {
		return connection.createSession(isTransacted, acknowledgeMode);
	}

	/**
	 * This implementation simply closes the session.
	 * Subclasses can override this method to implement session pooling.
	 * @throws JMSException 
	 */
	protected void releaseSession(Session session) throws JMSException {
		session.close();
	}

	/**
	 * Provides access to destination, initial context, connection, and worker.
	 * Bridges to initial context. For names in a form initialContext/&lt;name&gt; the name is looked up in initial context.
	 */
	protected Object getChild(String name) {
		if ("destination".equals(name)) {
			return destination;
		}
		
		if ("initialContext".equals(name)) {
			return initialContext;
		}
		
		if (name!=null && name.startsWith(INITIAL_CONTEXT_PREFIX)) {
			String jndiName = name.substring(INITIAL_CONTEXT_PREFIX.length());
			try {
				return initialContext.lookup(jndiName);
			} catch (NamingException e) {
				throw new RuntimeConfigurationException("Lookup in inital context failed for JNDI name '"+jndiName+"':"+e, e);
			}
		}
		
		if ("connection".equals(name)) {
			return connection;
		}
		
		if ("connectionFactory".equals(name)) {
			return connectionFactory;
		}
		
		if ("worker".equals(name)) {
			return worker;
		}		
		
		return super.getChild(name);
	}
	
	private String messageSelector;
	
	private boolean noLocal;
	
	/**
	 * Message selector
	 * @param messageSelector
	 */
	public void setMessageSelector(String messageSelector) {
		this.messageSelector = messageSelector;
	}
	
	/**
	 * If true and destination is topic then messages produced by this connection are not consumed by the message processor.
	 * @param noLocal
	 */
	public void setNoLocal(boolean noLocal) {
		this.noLocal = noLocal;
	}
	
	protected RestartCommand restartCommand;
		
	public void setRestartCommand(RestartCommand command) {
		this.restartCommand = command;		
	}
}
