package biz.hammurapi.jms.adapter;

import java.io.InputStream;
import java.lang.reflect.Constructor;
import java.net.URL;
import java.util.ArrayList;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Map;
import java.util.Timer;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.apache.xmlbeans.XmlObject;
import org.apache.xmlbeans.XmlOptions;

import biz.hammurapi.config.ConfigurationException;
import biz.hammurapi.config.DomConfigFactory;
import biz.hammurapi.config.GenericContainer;
import biz.hammurapi.config.MapContext;
import biz.hammurapi.config.MutableContext;
import biz.hammurapi.config.SimpleContext;
import biz.hammurapi.config.Wrapper;
import biz.hammurapi.convert.CompositeConverter;
import biz.hammurapi.jms.adapter.definition.GenericService;
import biz.hammurapi.jms.adapter.definition.JmsAdapterDocument;
import biz.hammurapi.jms.adapter.definition.NamedObjectSpecification;
import biz.hammurapi.jms.adapter.definition.ObjectSpecification;
import biz.hammurapi.jms.adapter.definition.Property;
import biz.hammurapi.metrics.MeasurementConsumer;
import biz.hammurapi.util.Worker;

public class JmsAdapter extends GenericContainer implements MutableContext, Worker {
	
	private static final String RESOURCE_PREFIX = "resource:";

	private static final Logger logger = Logger.getLogger(JmsAdapter.class.getName());
	
	private Timer timer;
	private biz.hammurapi.jms.adapter.definition.JmsAdapter definition;

	private Worker defaultWorker;
	
	/**
	 * Sets default worker. Overrides default worker read from
	 * configuration. This method is used by JCA wrapper.
	 * @param defaultWorker
	 */
	public void setDefaultWorker(Worker defaultWorker) {
		this.defaultWorker = defaultWorker;
	}
	
	public Timer getTimer() {
		return timer;
	}
	
	/**
	 * Instantiates JMS adapter from the definition.
	 * @param definition
	 */
	public JmsAdapter(biz.hammurapi.jms.adapter.definition.JmsAdapter definition) throws ConfigurationException {
		this.definition = definition;
		
		// Measurement consumer
		if (definition.getMeasurementConsumer()!=null) {
			setMeasurementConsumer((MeasurementConsumer) instantiate(definition.getMeasurementConsumer()));
		}
		
		// Bind types
		NamedObjectSpecification[] bindTypesArray = definition.getBindTypeArray();
		GenericContainer bindTypes=new GenericContainer();
		addComponent("bind-types", bindTypes);
		for (int i=0; i<bindTypesArray.length; ++i) {
			bindTypes.addComponent(bindTypesArray[i].getName(), instantiate(bindTypesArray[i]));
		}
		
		// Workers
		biz.hammurapi.jms.adapter.definition.Worker[] workerArray = definition.getWorkerArray();
		GenericContainer workers=new GenericContainer();
		addComponent("workers", workers);
		for (int i=0; i<workerArray.length; ++i) {
			Object worker = instantiate(workerArray[i]);
			workers.addComponent(workerArray[i].getName(), worker);
			if (workerArray[i].getDefault()) {
				defaultWorker = (Worker) worker;
			}
		}
		
		// Generic services
		GenericService[] serviceArray = definition.getServiceArray();
		GenericContainer services=new GenericContainer();
		addComponent("services", services);
		for (int i=0; i<serviceArray.length; ++i) {
			Object service = instantiate(serviceArray[i]);
			services.addComponent(serviceArray[i].getName(), service);
			for (int j=0; j<serviceArray[i].getAliasArray().length; ++j) {
				set(serviceArray[i].getAliasArray()[j], service);
			}
		}
		
		// Contexts
		biz.hammurapi.jms.adapter.definition.Context[] contextArray = definition.getContextArray();
		GenericContainer contexts=new GenericContainer();
		addComponent("contexts", contexts);
		for (int i=0; i<contextArray.length; ++i) {
			contexts.addComponent(contextArray[i].getName(), new JndiContext(this, contextArray[i]));
		}
		
		// Factory connections
		biz.hammurapi.jms.adapter.definition.FactoryConnection[] connectionArray = definition.getConnectionArray();
		GenericContainer connections=new GenericContainer();
		addComponent("connections", connections);
		for (int i=0; i<connectionArray.length; ++i) {
			connections.addComponent(connectionArray[i].getName(), new FactoryConnection(this, connectionArray[i]));
		}
		
		
	}
	
	/**
	 * Instantiates object from specifications and configures it by injecting properties.
	 * @param objSpec
	 * @return
	 * @throws ConfigurationException
	 */
	public static Object instantiate(ObjectSpecification objSpec) throws ConfigurationException {
		try {
			Class clazz = Class.forName(objSpec.getType());
			Constructor candidate = null;
			Constructor[] ca = clazz.getConstructors();
			for (int i = 0; i<ca.length; ++i) {
				Class[] parameterTypes = ca[i].getParameterTypes();
				if (parameterTypes.length==1 
						&& parameterTypes[0].isInstance(objSpec) 
						&& (candidate==null || candidate.getParameterTypes()[0].isAssignableFrom(parameterTypes[0]))) {
					candidate = ca[i];
				}
			}
			
			if (candidate!=null) {
				return candidate.newInstance(new Object[] {objSpec});
			}
			
			Object ret = clazz.newInstance();
			Map properties = instantiate(objSpec.getPropertyArray());
			if (!properties.isEmpty()) {
				DomConfigFactory.inject(ret, new MapContext(properties));
			}
			return ret;
		} catch (ConfigurationException e) {
			throw e;
		} catch (Exception e) {
			throw new ConfigurationException("Could not instantiate object from specification: "+e, e);
		}
	}

	/**
	 * Instantiates object from specifications and configures it by injecting properties.
	 * @param objSpec
	 * @return
	 * @throws ConfigurationException
	 */
	public static Object instantiate(NamedObjectSpecification objSpec) throws ConfigurationException {
		try {
			Class clazz = Class.forName(objSpec.getType());
			Constructor candidate = null;
			Constructor[] ca = clazz.getConstructors();
			for (int i = 0; i<ca.length; ++i) {
				Class[] parameterTypes = ca[i].getParameterTypes();
				if (parameterTypes.length==1 
						&& parameterTypes[0].isInstance(objSpec) 
						&& (candidate==null || candidate.getParameterTypes()[0].isAssignableFrom(parameterTypes[0]))) {
					candidate = ca[i];
				}
			}
			
			if (candidate!=null) {
				return candidate.newInstance(new Object[] {objSpec});
			}
			
			Object ret = clazz.newInstance();
			Map properties = instantiate(objSpec.getPropertyArray());
			if (!properties.isEmpty()) {
				DomConfigFactory.inject(ret, new MapContext(properties));
			}
			return ret;
		} catch (ConfigurationException e) {
			throw e;
		} catch (Exception e) {
			throw new ConfigurationException("Could not instantiate object from specification: "+e, e);
		}
	}

	/**
	 * Instantiates properties from XML defintion
	 * @param property
	 * @return
	 * @throws ConfigurationException
	 */
	public static Hashtable instantiate(Property[] properties) throws ConfigurationException {
		Hashtable ret = new Hashtable();
		if (properties!=null) {
			for (int i=0; i<properties.length; ++i) {
				if (properties[i].getType()==null || properties[i].getType().trim().length()==0) {
					ret.put(properties[i].getName(), properties[i].getStringValue());
				} else {				
					try {
						ret.put(
								properties[i].getName(), 
								CompositeConverter
									.getDefaultConverter()
									.convert(
											properties[i].getStringValue(), 
											Class.forName(properties[i].getType()), true));
					} catch (ClassNotFoundException e) {
						throw new ConfigurationException("Cannot instantiate property of type "+properties[i].getType()+" from value "+properties[i].getStringValue());
					}
				}
			}
		}
		return ret;
	}

	public void start() throws ConfigurationException {
		timer = new Timer();
		if (definition.getStartInBackground()) {
			new Thread(definition.getName()+" startup thread") {
				public void run() {
					try {
						super.start();
					} catch (Exception e) {
						logger.log(Level.SEVERE, "Failed to start JMS Adapter "+definition.getName()+": "+e, e);
					}
				}
			}.start();
		} else {
			super.start();
		}

	}
	
	public void stop() throws ConfigurationException {
		if (timer!=null) {
			timer.cancel();
		}
		super.stop();
	}

	private SimpleContext aliases = new SimpleContext();

	public void remove(String name) {
		aliases.remove(name);		
	}

	public void set(String name, Object value) {
		aliases.set(name, value);		
	}
	
	/**
	 * Returns service. Unwraps wrappers.
	 */
	public Object get(String name) {		
		Object ret = super.get(name);
		if (ret==null) {
			ret = aliases.get(name);
		}
		return ret instanceof Wrapper ? ((Wrapper) ret).getMaster() : ret;
	}

	/**
	 * The first and only argument is definition URL.
	 * If URL starts with "resource:" then definition is loaded from classpath resource.
	 * If the first argument is not present then definition is read from System.in
	 * Adapter is stopped by shutdown hook on JVM termination (Ctrl-C).
	 * @param args
	 */
	public static void main(String[] args) {
		System.out.println("Usage: java <options> biz.hammurapi.jms.adapter.JmsAdapter [<definition URL>]");
		
		final long start=System.currentTimeMillis();
		
		XmlObject parsed;
		try {
			if (args.length==0) {
				parsed = XmlObject.Factory.parse(System.in);
			} else if (args[0].startsWith(RESOURCE_PREFIX)) {
				InputStream resourceStream = JmsAdapter.class.getClassLoader().getResourceAsStream(
						args[0].substring(RESOURCE_PREFIX.length()));
				parsed = XmlObject.Factory.parse(resourceStream);
			} else {
				parsed = XmlObject.Factory.parse(new URL(args[0]));
			}
		} catch (Exception e) {
			System.err.println("Could not load definition: "+e);
			e.printStackTrace();	
			System.exit(3);
			return; // Stupid, but needed for proper compilation.
		}
		
		try {
			if (parsed instanceof JmsAdapterDocument) {
				ArrayList validationErrors = new ArrayList();
				XmlOptions validationOptions = new XmlOptions();
				validationOptions.setErrorListener(validationErrors);
				if (!parsed.validate(validationOptions)) {
					System.err.println("Invalid adapter definition:");
				    Iterator iter = validationErrors.iterator();
				    while (iter.hasNext()) {
				        System.err.println("\t>> " + iter.next());
				    }
					
					System.exit(1);
				}
				biz.hammurapi.jms.adapter.definition.JmsAdapter definition = ((JmsAdapterDocument) parsed).getJmsAdapter();
				final JmsAdapter adapter = new JmsAdapter(definition);
				System.out.println("Adapter started in "+((System.currentTimeMillis()-start)/1000)+" seconds");
				Runtime.getRuntime().addShutdownHook(
						new Thread() {
							public void run() {
								try {							
									adapter.stop();
								} catch (ConfigurationException e) {
									System.err.println("Could not properly stop JMS adapter.");
									e.printStackTrace();
								}
								System.out.println("Total execution time: "+((System.currentTimeMillis()-start)/1000)+" sec.");							
							}
						});			
				
				adapter.start();
			} else {
				System.out.println("Invalid adapter definition.");
				System.exit(2);
			}
		} catch (Exception e) {
			System.err.println("Could not start adapter: "+e);
			e.printStackTrace();	
			System.exit(4);			
		}
		
	}
	
	public static boolean isBlank(String str) {
		return str==null || str.trim().length()==0;
	}

	/**
	 * @return Adapter XML definition.
	 */
	public biz.hammurapi.jms.adapter.definition.JmsAdapter getDefinition() {
		return definition;
	}

	/**
	 * Adapter is always naming root. Owner is available through 'owner' alias.
	 */
	public void setOwner(Object owner) {
		aliases.set("owner", owner);
	}

	/**
	 * Executes job in the current thread if there is no default worker.
	 */
	public boolean post(Runnable job) {
		if (defaultWorker==null) {
			job.run();
			return true;
		}
		
		return defaultWorker.post(job);
	}
	
}
