package biz.hammurapi.jms.adapter;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.Timer;
import java.util.TimerTask;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.jms.Connection;
import javax.jms.ConnectionFactory;
import javax.jms.ExceptionListener;
import javax.jms.JMSException;
import javax.jms.Session;

import org.apache.commons.pool.ObjectPool;
import org.apache.commons.pool.PoolableObjectFactory;
import org.apache.commons.pool.impl.StackObjectPool;

import biz.hammurapi.config.ConfigurationException;
import biz.hammurapi.config.GenericContainer;
import biz.hammurapi.jms.JMSExceptionEx;

/**
 * Base class for Jndi and Factory connections. 
 * @author Pavel
 */
public abstract class JmsConnection extends GenericContainer {
	
	private static final Logger logger = Logger.getLogger(JmsConnection.class.getName());
	
	private static final int ACKNOWLEDGE_MODE = Session.AUTO_ACKNOWLEDGE;

	private static final long DEFAULT_RECOVERY_DELAY = 10000; // 10 seconds default recovery delay

	protected biz.hammurapi.jms.adapter.definition.Connection definition;

	public JmsConnection(JmsAdapter jmsAdapter, biz.hammurapi.jms.adapter.definition.Connection definition) throws ConfigurationException {
		this.definition = definition;
		
		// Services
		biz.hammurapi.jms.adapter.definition.JmsService[] serviceArray = definition.getServiceArray();
		GenericContainer services=new GenericContainer();
		addComponent("services", services);
		for (int i=0; i<serviceArray.length; ++i) {
			JmsService service = new JmsService(serviceArray[i]);
			services.addComponent(serviceArray[i].getName(), service);
			for (int j=0; j<serviceArray[i].getAliasArray().length; ++j) {
				jmsAdapter.set(serviceArray[i].getAliasArray()[j], service);
			}
		}
		
		// Proxy services
		biz.hammurapi.jms.adapter.definition.ProxyService[] proxyServiceArray = definition.getProxyServiceArray();
		for (int i=0; i<proxyServiceArray.length; ++i) {
			ProxyService service = new ProxyService(proxyServiceArray[i]);
			services.addComponent(proxyServiceArray[i].getName(), service);
			for (int j=0; j<proxyServiceArray[i].getAliasArray().length; ++j) {
				jmsAdapter.set(proxyServiceArray[i].getAliasArray()[j], service);
			}
		}
		
		// Listeners
		biz.hammurapi.jms.adapter.definition.Listener[] listenerArray = definition.getListenerArray();
		ListenerContainer listeners=new ListenerContainer();
		addComponent("listeners", listeners);
		for (int i=0; i<listenerArray.length; ++i) {
			Listener listener = new Listener(listenerArray[i]);
			listeners.addComponent(listenerArray[i].getName(), listener);
			for (int j=0; j<listenerArray[i].getAliasArray().length; ++j) {
				jmsAdapter.set(listenerArray[i].getAliasArray()[j], listener);
			}
		}
		
	}
	
	private class ListenerContainer extends GenericContainer {
		
		void enroll(Session session) throws JMSException {
			Iterator it = getComponents().iterator();
			while (it.hasNext()) {
				((Listener) it.next()).enroll(session);
			}
		}
	}
	
	/**
	 * Subclasses shall implement this method to either look up connection factory in JNDI or create using
	 * vendor-specific API's.
	 * @return
	 * @throws JMSException
	 */
	protected abstract ConnectionFactory getConnectionFactory() throws ConfigurationException;
	
	
	private class ThreadSessionEntry {
		int counter;
		Session session;
	}
	
	private ThreadLocal sessionTL = new ThreadLocal() {
		protected Object initialValue() {
			return new ThreadSessionEntry();
		}
	};
	
	private int connectionIndex;
	
	private static final long CONNECTION_WAIT = 30000; // 30 seconds to wait for connection availability.
	
	/**
	 * @return Valid connection is a round-robin fashion
	 */
	private Connection getConnection() throws JMSException {
		for (int k=0; k<2; ++k) {
			synchronized (connections) {
				for (int i=0; i<connections.length; ++i) {
					connectionIndex = (connectionIndex+1) % connections.length;
					if (connections[connectionIndex]!=null) {
						return connections[connectionIndex];
					}
				}
				
				try {
					connections.wait(CONNECTION_WAIT);
				} catch (InterruptedException e) {
					throw new JMSExceptionEx("Waiting for connection was interrupted: "+e, e);
				}
			}
		}
		
		throw new JMSException("No connection is available at this time.");
	}
	
	public Session borrowSession() throws JMSException {
		ThreadSessionEntry tse = (ThreadSessionEntry) sessionTL.get();
		
		if (definition.getReuseThreadSession() && tse.counter>0) {
			++tse.counter;
			return tse.session;
		}
		
		if (sessionPool == null) {
			/*
			 *  TODO Use keyed pool to allow services to specify transactionality and acknowledge mode
			 *  Do not reuse thread session if transactional.
			 *  More complex ThreadSessionEntry - key sessions by acknowledge mode.
			 */
			return getConnection().createSession(false, ACKNOWLEDGE_MODE);
		} else {
			try {
				return (Session) sessionPool.borrowObject();
			} catch (Exception e) {
				throw new JMSExceptionEx("Failed to get session from pool: "+e, e);
			}			
		}
	}
	
	private ObjectPool sessionPool;
	
	public void returnSession(Session session) throws JMSException {
		ThreadSessionEntry tse = (ThreadSessionEntry) sessionTL.get();
		
		if (definition.getReuseThreadSession() && --tse.counter>0) {
			return;
		}
		
		tse.counter=0;
		tse.session=null;
		
		if (sessionPool == null) {
			session.close();
		} else {
			try {
				sessionPool.returnObject(session);
			} catch (Exception e) {
				throw new JMSExceptionEx("Failed to return session to pool: "+e, e);
			}			
		}
	}
	
	public void invalidateSession(Session session) throws JMSException {
		ThreadSessionEntry tse = (ThreadSessionEntry) sessionTL.get();
		tse.counter=0;
		tse.session=null;
		
		if (sessionPool == null) {
			session.close();
		} else {
			try {
				sessionPool.invalidateObject(session);
			} catch (Exception e) {
				throw new JMSExceptionEx("Failed to invalidate session in pool: "+e, e);
			}			
		}
		
	}
	
	private Connection[] connections;
	
	private Collection connectionTasks = new ArrayList();
	
	/**
	 * Starts connection, enrolls listener sessions
	 * @author Pavel
	 *
	 */
	private class ConnectionTask extends TimerTask {
		
		private int index;
		private boolean cancelOnSuccess;

		ConnectionTask(int index, boolean cancelOnSuccess) {
			this.index = index;
			this.cancelOnSuccess = cancelOnSuccess;
			synchronized (connectionTasks) {
				connectionTasks.add(this);
			}
		}
		
		public boolean cancel() {
			synchronized (connectionTasks) {
				connectionTasks.remove(this);
			}
			return super.cancel();
		}

		public void run() {
			try {
				ConnectionFactory factory = getConnectionFactory();				
				Connection connection;
				if (JmsAdapter.isBlank(definition.getUser())) {
					connection = factory.createConnection();
				} else {
					connection = factory.createConnection(definition.getUser(), definition.getPassword());
				}
				
				connection.setExceptionListener(
					new ExceptionListener() {

						public void onException(JMSException e) {
							logger.log(Level.SEVERE, "Connection failure: "+e+", re-creating connection.", e);
							synchronized (connections) {
								connections[index]=null;
							}
							Timer timer = ((JmsAdapter) getOwner(JmsAdapter.class)).getTimer();
							timer.schedule(new ConnectionTask(index, true), 0, definition.getRecoveryDelay()==0 ? DEFAULT_RECOVERY_DELAY : definition.getRecoveryDelay());
						}
						
					});
				
				int listeningSessions = Math.max(1, definition.getListeningSessions());
				for (int i=0; i<listeningSessions; ++i) {
					((ListenerContainer) get("listeners")).enroll(connection.createSession(false, ACKNOWLEDGE_MODE));
				}
				
				connection.start();
				
				logger.fine("Connection started");
				
				synchronized (connections) {
					if (connections[index]!=null) {
						try {
							connections[index].close();
						} catch (JMSException e) {
							logger.log(Level.WARNING, "Could not close connection: "+e, e);							
						}
					}
					connections[index] = connection;
					connections.notifyAll(); // Notify pending threads about availability of a connection.					
				}
				
				if (cancelOnSuccess) {
					cancel();
				} else {
					index = index+1 % connections.length;
				}
			} catch (Exception e) {
				logger.log(Level.SEVERE, "Could not create connection: "+e, e);
			}			
		}
		
	}
	
	public void start() throws ConfigurationException {
		// Start connections and schedule refresh tasks.
		connections = new Connection[Math.max(1, definition.getConnectionPool())];
		Timer timer = ((JmsAdapter) getOwner(JmsAdapter.class)).getTimer();
		for (int i=0; i<connections.length; ++i) {
			timer.schedule(new ConnectionTask(i, true), 0, definition.getRecoveryDelay()==0 ? DEFAULT_RECOVERY_DELAY : definition.getRecoveryDelay());
		}
		
		// Regular reconnection
		if (definition.getRefreshInterval()>0) {
			timer.schedule(new ConnectionTask(0, false), definition.getRefreshInterval(), definition.getRefreshInterval());
		}
		
		// Create session pool
		if (definition.getSessionPool()!=null) {
			PoolableObjectFactory sessionFactory = new PoolableObjectFactory() {

				public void activateObject(Object obj) throws Exception {
					// Nothing to do					
				}

				public void destroyObject(Object obj) throws Exception {					
					((Session) obj).close();
				}

				public Object makeObject() throws Exception {
					return getConnection().createSession(false, ACKNOWLEDGE_MODE);
				}

				public void passivateObject(Object obj) throws Exception {
					// Nothing to do					
				}

				public boolean validateObject(Object obj) {
					// Nothing to do
					return true;
				}
				
			};
			
			sessionPool = new StackObjectPool(
					sessionFactory, 
					definition.getSessionPool().getMaxIdle(), 
					definition.getSessionPool().getInitial());
		}
		
		super.start();
	}
	
	public void stop() throws ConfigurationException {
		super.stop();
		
		if (sessionPool!=null) {
			try {
				sessionPool.close();
			} catch (Exception e) {
				throw new ConfigurationException("Could not stop session pool: "+e, e);
			}
		}
		
		Iterator it;
		synchronized (connectionTasks) {
			it = new ArrayList(connectionTasks).iterator();
		}
		
		while (it.hasNext()) {
			((ConnectionTask) it.next()).cancel();
		}
		
		synchronized (connections) {
			for (int i=0; i<connections.length; ++i) {
				if (connections[i]!=null) {
					try {
						connections[i].close();
					} catch (JMSException e) {
						logger.log(Level.INFO, "Connection closing failed: "+e, e);
					}
				}
			}
		}
	}

}
