package biz.hammurapi.jms.adapter;

import java.util.Hashtable;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.jms.DeliveryMode;
import javax.jms.Destination;
import javax.jms.JMSException;
import javax.jms.Message;
import javax.jms.MessageConsumer;
import javax.jms.MessageProducer;
import javax.jms.Session;
import javax.jms.TemporaryQueue;

import biz.hammurapi.config.ConfigurationException;
import biz.hammurapi.config.ServiceBase;
import biz.hammurapi.jms.JMSExceptionEx;
import biz.hammurapi.util.ExceptionSink;
import biz.hammurapi.util.Worker;

public class JmsService extends ServiceBase {
	
	private static final Logger logger = Logger.getLogger(JmsService.class.getName());

	private biz.hammurapi.jms.adapter.definition.JmsService definition;

	private Hashtable properties;
	
	public JmsService(biz.hammurapi.jms.adapter.definition.JmsService definition) throws ConfigurationException {
		this.definition = definition;
		this.properties = JmsAdapter.instantiate(definition.getPropertyArray());
	}
	
	private Destination requestDestination;
	
	private Destination replyDestination;
	
	private Converter converter;
	
	private Worker worker;
	
	public Worker getWorker() {
		return worker;
	}

	protected void startInternal() throws ConfigurationException {
		if (!definition.getQueueFromSession() && !definition.getTopicFromSession()) {		
			requestDestination = ((DestinationResolver) getOwner(DestinationResolver.class)).lookupDestination(definition.getRequestDestination());
			
			if (!JmsAdapter.isBlank(definition.getReplyDestination())) {
				replyDestination = ((DestinationResolver) getOwner(DestinationResolver.class)).lookupDestination(definition.getReplyDestination());
			}
		}
		
		converter = (Converter) get("/bind-types/"+definition.getBindType());
		
		if (definition.getWorker()==null) {
			worker = (Worker) getOwner(Worker.class);
		} else {
			worker = (Worker) get("/workers/"+definition.getWorker());
		}		
		
	}

	protected void stopInternal() throws ConfigurationException {
		// Nothing to do
	}
	
	private Destination getRequestDestination(Session session) throws JMSException {
		if (definition.getQueueFromSession()) {
			return session.createQueue(definition.getRequestDestination());
		}
		
		if (definition.getTopicFromSession()) {
			return session.createTopic(definition.getRequestDestination());						
		}
		
		return requestDestination;
	}
	
	private Destination getReplyDestination(Session session) throws JMSException {
		if (JmsAdapter.isBlank(definition.getReplyDestination())) {
			return session.createTemporaryQueue();
		}
		
		if (definition.getQueueFromSession()) {
			return session.createQueue(definition.getReplyDestination());
		}
		
		if (definition.getTopicFromSession()) {
			return session.createTopic(definition.getReplyDestination());						
		}
		
		return replyDestination;
	}
	
	/**
	 * Sends object to JMS destination.
	 * @param obj Object to send
	 * @param async If true all operations are performed in a worker thread.
	 * @throws JMSException
	 */
	public void send(final Object obj, boolean async) throws JMSException {
		if (async) {
			worker.post(new Runnable() {

				public void run() {
					try {
						send(obj, false);
					} catch (JMSException e) {
						logger.log(Level.SEVERE, "send() failed: "+e, e);
					}					
				}
				
			});
		} else {
			long start = System.currentTimeMillis();
			JmsConnection connection = (JmsConnection) getOwner(JmsConnection.class);
			Session session = connection.borrowSession();
			try {
				try {
					Message message = converter.convert(obj, session, properties, null);
					MessageProducer producer = session.createProducer(getRequestDestination(session));
					
					if (definition.getTimeToLive()>0) {
						producer.setTimeToLive(definition.getTimeToLive());
					}
					
					producer.setDeliveryMode(definition.getPersistent() ? DeliveryMode.PERSISTENT : DeliveryMode.NON_PERSISTENT);
					
					if (definition.getPriority()!=null) {
						producer.setPriority(definition.getPriority().intValue());
					}
					
					try {
						producer.send(message);
					} finally {
						producer.close();
					}
				} finally {
					connection.returnSession(session);
				}
			} catch (Exception e) {
				addMeasurement("error", 1, System.currentTimeMillis());
				connection.invalidateSession(session);
				if (e instanceof RuntimeException) {
					throw (RuntimeException) e;
				}
				throw e instanceof JMSException ? (JMSException) e : new JMSExceptionEx(e);
			} finally {
				long now = System.currentTimeMillis();
				addMeasurement("send", now - start, now);
			}
		}
	}
	
	/**
	 * Sends object JMS destination, waits for reply, receives reply converts to Object
	 * and returns. Reply message is correlated on send message ID if reply destination is set.
	 * If reply destination is not specified in configuration then temporary reply destination is 
	 * used. If wait expires this method returns null.
	 * @param obj Object to send.
	 * @return Reply object.
	 * @throws JMSException
	 */
	public Object request(Object obj) throws JMSException {
		long start = System.currentTimeMillis();
		JmsConnection connection = (JmsConnection) getOwner(JmsConnection.class);
		Session session = connection.borrowSession();
		try {
			try {
				Message message = converter.convert(obj, session, properties, null);
				MessageProducer producer = session.createProducer(getRequestDestination(session));
				
				if (definition.getTimeToLive()>0) {
					producer.setTimeToLive(definition.getTimeToLive());
				}
				
				message.setJMSReplyTo(getReplyDestination(session));
							
				producer.setDeliveryMode(definition.getPersistent() ? DeliveryMode.PERSISTENT : DeliveryMode.NON_PERSISTENT);
				
				if (definition.getPriority()!=null) {
					producer.setPriority(definition.getPriority().intValue());
				}
								
				try {
					producer.send(message);
				} finally {
					producer.close();
				}
				
				long waitStart = System.currentTimeMillis();
				
				MessageConsumer consumer;
	
				if (message.getJMSReplyTo() instanceof TemporaryQueue) {
					consumer = session.createConsumer(message.getJMSReplyTo());
				} else {
					String selector = "JMSCorrelationID='"+message.getJMSMessageID()+"'";
					consumer = session.createConsumer(message.getJMSReplyTo(), selector);
				}
				
				try {
					Message reply = consumer.receive(definition.getTimeout());
					if (reply == null) {
						addMeasurement("timeout", 1, System.currentTimeMillis());
						throw new JMSException("Waiting for reply timed out");
					}
					
					long waitEnd = System.currentTimeMillis();
					addMeasurement("wait", waitEnd - waitStart, waitEnd);
					
					return converter.convert(reply, properties);
				} finally {
					consumer.close();
					if (message.getJMSReplyTo() instanceof TemporaryQueue) {
						((TemporaryQueue) message.getJMSReplyTo()).delete();
					}
				}		
			} catch (Exception e) {
				addMeasurement("error", 1, System.currentTimeMillis());
				connection.invalidateSession(session);
				if (e instanceof RuntimeException) {
					throw (RuntimeException) e;
				}
				throw e instanceof JMSException ? (JMSException) e : new JMSExceptionEx(e);
			} finally {
				connection.returnSession(session);
			}
		} finally {
			long now = System.currentTimeMillis();
			addMeasurement("request", now - start, now);
		}
	}
	
	/**
	 * Asynchronously performs request/reply operation in worker thread.
	 * @param obj Object to send
	 * @param receptor Processor for reply. In the case of failure, exception is passed to the processor.
	 */
	public void request(final Object obj, final Processor receptor) {
		worker.post(new Runnable() {

			public void run() {
				try {
					receptor.process(request(obj));
				} catch (Exception e) {
					if (receptor instanceof ExceptionSink) {
						((ExceptionSink) receptor).consume(obj, e);
					} else {
						logger.log(Level.SEVERE, "Request processing failed: "+e, e);
					}
				}				
			}
			
		});
	}

}
