package biz.hammurapi.jms.adapter;

import java.util.Hashtable;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.jms.Destination;
import javax.jms.JMSException;
import javax.jms.Message;
import javax.jms.MessageConsumer;
import javax.jms.MessageListener;
import javax.jms.MessageProducer;
import javax.jms.Session;

import org.apache.commons.pool.ObjectPool;
import org.apache.commons.pool.PoolableObjectFactory;
import org.apache.commons.pool.impl.SoftReferenceObjectPool;

import biz.hammurapi.config.Component;
import biz.hammurapi.config.ConfigurationException;
import biz.hammurapi.config.PoolableComponent;
import biz.hammurapi.config.ServiceBase;
import biz.hammurapi.config.Wrapper;
import biz.hammurapi.metrics.MeasurementCollector;
import biz.hammurapi.metrics.MeasurementConsumer;
import biz.hammurapi.util.ExceptionSink;
import biz.hammurapi.util.Worker;

/**
 * Listens for JMS messages, converts them to objects using Converter and processes
 * them using processor. Processors can implement Observer/Observable pattern and pass received
 * objects to registered observers/listeners. 
 * @author Pavel
 */
public class Listener extends ServiceBase implements MessageListener, Wrapper {

	private Logger logger = Logger.getLogger(Listener.class.getName()); 
	
	private biz.hammurapi.jms.adapter.definition.Listener definition;
	private Hashtable properties;
	private Processor processor;
	private ObjectPool pool;
	
	public Listener(final biz.hammurapi.jms.adapter.definition.Listener definition) throws ConfigurationException {
		this.definition = definition;
		properties = JmsAdapter.instantiate(definition.getPropertyArray());
		if (definition.getPooled()) {
			PoolableObjectFactory objectFactory = new PoolableObjectFactory() {

				public void activateObject(Object obj) throws Exception {
					if (obj instanceof PoolableComponent) {
						((PoolableComponent) obj).activate();
					}					
				}

				public void destroyObject(Object obj) throws Exception {
					if (obj instanceof Component) {
						((Component) obj).stop();
					}					
				}

				public Object makeObject() throws Exception {
					Object ret = (Processor) JmsAdapter.instantiate(definition.getProcessor());
					if (ret instanceof Component) {
						((Component) ret).setOwner(this);
						((Component) ret).start();
					}
					if (ret instanceof MeasurementCollector) {
						MeasurementConsumer cmc = new MeasurementConsumer() {
							public void addMeasurement(String mName, double value, long time) {
								MeasurementConsumer measurementConsumer = getMeasurementConsumer();
								if (measurementConsumer != null) {
									measurementConsumer.addMeasurement("processor."+mName, value, time==0 ? System.currentTimeMillis() : time);
								}						
							}
						};
						((MeasurementCollector) ret).setMeasurementConsumer(cmc);
					}					
					return ret;
				}

				public void passivateObject(Object obj) throws Exception {
					if (obj instanceof PoolableComponent) {
						((PoolableComponent) obj).passivate();
					}					
				}

				public boolean validateObject(Object obj) {
					if (obj instanceof PoolableComponent) {
						return ((PoolableComponent) obj).validate();
					}	
					
					return true;
				}
				
			};
			pool = new SoftReferenceObjectPool(objectFactory);
		} else {
			processor = (Processor) JmsAdapter.instantiate(definition.getProcessor());
			if (processor instanceof Component) {
				((Component) processor).setOwner(this);
			}
			if (processor instanceof MeasurementCollector) {
				MeasurementConsumer cmc = new MeasurementConsumer() {
					public void addMeasurement(String mName, double value, long time) {
						MeasurementConsumer measurementConsumer = getMeasurementConsumer();
						if (measurementConsumer != null) {
							measurementConsumer.addMeasurement("processor."+mName, value, time==0 ? System.currentTimeMillis() : time);
						}						
					}
				};
				((MeasurementCollector) processor).setMeasurementConsumer(cmc);
			}					
		}
	}
	
	private Destination getDestination(Session session) throws JMSException {
		if (definition.getQueueFromSession()) {
			return session.createQueue(definition.getDestination());
		}
		
		if (definition.getTopicFromSession()) {
			return session.createTopic(definition.getDestination());						
		}
		
		return destination;
	}
	
	private Destination getReplyDestination(Session session) throws JMSException {
		if (definition.getQueueFromSession()) {
			return session.createQueue(definition.getReplyDestination());
		}
		
		if (definition.getTopicFromSession()) {
			return session.createTopic(definition.getReplyDestination());						
		}
		
		return replyDestination;
	}
	
	private Destination destination;
	private Destination replyDestination;
	private Converter converter;
	private Worker worker;

	protected void stopInternal() throws ConfigurationException {
		if (processor instanceof Component) {
			((Component) processor).stop();
		}
		if (pool!=null) {
			try {
				pool.close();
			} catch (Exception e) {
				throw new ConfigurationException("Could not close pool: "+e, e);
			}
		}
	}

	public void startInternal() throws ConfigurationException {
		if (!definition.getQueueFromSession() && !definition.getTopicFromSession()) {		
			destination = ((DestinationResolver) getOwner(DestinationResolver.class)).lookupDestination(definition.getDestination());
			if (!JmsAdapter.isBlank(definition.getReplyDestination())) {
				replyDestination = ((DestinationResolver) getOwner(DestinationResolver.class)).lookupDestination(definition.getReplyDestination());
			}
		}
		
		converter = (Converter) get("/bind-types/"+definition.getBindType());
		if (processor instanceof Component) {
			((Component) processor).start();
		}
		
		if (definition.getWorker()!=null) {
			worker = (Worker) get("/workers/"+definition.getWorker());
		}
	}

	/**
	 * Adds listener to JMS session
	 * @param session
	 * @throws JMSException
	 */
	public void enroll(Session session) throws JMSException {
		MessageConsumer consumer = session.createConsumer(getDestination(session), definition.getMessageSelector(), definition.getNoLocal());
		consumer.setMessageListener(this);
	}
	
	private Processor borrowProcessor() throws Exception {
		if (definition.getPooled()) {
			return (Processor) pool.borrowObject();
		}
		
		return processor; 
	}
	
	private void returnProcessor(Processor processor) throws Exception {
		if (definition.getPooled()) {
			pool.returnObject(processor);
		}
	}

	/**
	 * Converts message to object and passes to processor.
	 */
	public void onMessage(final Message message) {	
		long start = System.currentTimeMillis();
		try {
			final Processor prc = borrowProcessor();
			final Destination replyTo = message.getJMSReplyTo();
			final int deliveryMode = message.getJMSDeliveryMode();
			final int priority = message.getJMSPriority();
			
			try {
				final Object obj = converter.convert(message, properties);
				
				if (worker==null) {
					try {
						try {
							Object ret;
							long prcStart = System.currentTimeMillis();
							try {
								ret = prc.process(obj);
							} finally {
								long now = System.currentTimeMillis();
								addMeasurement("message.process", now - prcStart, now);
							}
							if (ret!=null) {
								reply(replyTo, deliveryMode, priority, message, ret);
							}
						} finally {
							returnProcessor(prc);
						}
					} catch (Exception e) {
						if (prc instanceof ExceptionSink) {
							((ExceptionSink) prc).consume(message, e);
						} else {
							logger.log(Level.SEVERE, "Processing error, sending error: "+e, e);
							replyError(replyTo, deliveryMode, priority, message, e);
						}
					}
				} else {
					Runnable job = new Runnable() {
		
						public void run() {
							try {
								try {
									Object ret;
									long prcStart = System.currentTimeMillis();
									try {
										ret = prc.process(obj);
									} finally {
										long now = System.currentTimeMillis();
										addMeasurement("message.process", now - prcStart, now);
									}
									if (ret!=null) {
										reply(replyTo, deliveryMode, priority, message, ret);
									}
								} finally {
									returnProcessor(prc);
								}
							} catch (Exception e) {
								if (prc instanceof ExceptionSink) {
									((ExceptionSink) prc).consume(Listener.this, e);
								} else {
									logger.log(Level.SEVERE, "Processing error, sending error: "+e, e);
									replyError(replyTo, deliveryMode, priority, message, e);
								}
							}
						}
						
					};
					
					// Execute in the current thread if worker doesn't accept job.
					if (!worker.post(job)) {
						job.run();
					}
				}
			} catch (Exception e) {
				if (prc instanceof ExceptionSink) {
					((ExceptionSink) prc).consume(message, e);
				} else {
					logger.log(Level.SEVERE, "Processing error: "+e, e);
					replyError(replyTo, deliveryMode, priority, message, e);
				}
			}
		} catch (Exception e) {
			logger.log(Level.SEVERE, "Processing error: "+e, e);			
		} finally {
			long now = System.currentTimeMillis();
			addMeasurement("message", now - start, now);
		}
	}

	private void replyError(final Destination replyTo, final int deliveryMode, final int priority, Message request, Exception e) {
		try {
			JmsConnection connection = (JmsConnection) getOwner(JmsConnection.class);
			Session session = connection.borrowSession();
			try {
				Message message = converter.convert(e, session, properties, request);
				sendReply(
						session, 
						message, 
						replyTo, 
						deliveryMode, 
						priority);
			} finally {
				connection.returnSession(session);
			}									
		} catch (Exception ex) {
			logger.log(Level.SEVERE, "Could not send error: "+e, e);									
		}
	}
	
	private void reply(final Destination replyTo, final int deliveryMode, final int priority, Message request, Object ret) throws Exception {
		JmsConnection connection = (JmsConnection) getOwner(JmsConnection.class);
		Session session = connection.borrowSession();
		try {
			Message message = converter.convert(ret, session, properties, request);
			sendReply(
					session, 
					message, 
					replyTo, 
					deliveryMode, 
					priority);
		} finally {
			connection.returnSession(session);
		}									
	}
	
	private void sendReply(Session session, Message message, Destination replyTo, int deliveryMode, int priority) throws JMSException {
		Destination finalDestination = replyTo==null ? getReplyDestination(session) : replyTo;
		if (finalDestination != null) {
			MessageProducer producer = session.createProducer(finalDestination);
			
			if (definition.getReplyTimeToLive()>0) {
				producer.setTimeToLive(definition.getReplyTimeToLive());
			}
			
			producer.setDeliveryMode(deliveryMode);		
			producer.setPriority(priority);
							
			try {
				producer.send(message);
			} finally {
				producer.close();
			}
		}
	}

	/**
	 * Returns processor.
	 */
	public Object getMaster() {
		return processor;
	}

}
