package biz.hammurapi.jms.adapter;

import java.io.Serializable;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationHandler;
import java.lang.reflect.Proxy;

import biz.hammurapi.config.Component;
import biz.hammurapi.config.ConfigurationException;
import biz.hammurapi.config.GenericContainer;
import biz.hammurapi.config.RuntimeConfigurationException;
import biz.hammurapi.config.Wrapper;
import biz.hammurapi.jms.adapter.definition.ProxyServiceDocument;
import biz.hammurapi.metrics.MeasurementCollector;
import biz.hammurapi.metrics.MeasurementConsumer;

public class ProxyService extends GenericContainer implements Wrapper {

	private final class ProxyServiceInvocationHandler implements InvocationHandler {
		private LocalDelegate delegate;

		private ProxyServiceInvocationHandler(LocalDelegate delegate) {
			this.delegate = delegate;
		}
		
		Object getState() {
			return delegate == null ? null : delegate.getState();
		}
		
		void setState(Object state) {
			if (delegate!=null) {
				delegate.setState(state);
			}
		}

		public Object invoke(final Object proxy, final java.lang.reflect.Method method, final Object[] args) throws Throwable {
			Method jmsMethod = (Method) get(signature(method));
			if (jmsMethod==null || jmsMethod.getWrapperException()==null) {
				return delegate.invoke(proxy, method, args, jmsMethod);
			} else {
				try {
					return delegate.invoke(proxy, method, args, jmsMethod);
				} catch (Exception e) {
					boolean isWrapperDeclared = false;
					Class[] declaredExceptions = method.getExceptionTypes();
					for (int i=0; i<declaredExceptions.length; ++i) {
						if (declaredExceptions[i].isInstance(e)) {
							throw e;
						}
						if (declaredExceptions[i].isAssignableFrom(jmsMethod.getWrapperException())) {
							isWrapperDeclared = true;
						}
					}
					
					if (e instanceof RuntimeException) {
						throw e;
					}
					
					if (!isWrapperDeclared) {
						throw e; // Wrapper shall be in the list of declared exceptions, we don't do double-wrapping.
					}
					
					Constructor[] constructors = jmsMethod.getWrapperException().getConstructors();
					
					/*
					 * 0 - String, Throwable
					 * 1 - Throwable, String
					 * 2 - Throwable
					 * 3 - String
					 * 4 - Default
					 */
					Constructor[] candidates = new Constructor[5];
					for (int i=0; i<constructors.length; ++i) {
						Class[] cpt = constructors[i].getParameterTypes();
						switch (cpt.length) {
						case 0:
								candidates[4] = constructors[i];
								break;
						case 1:
								if (String.class.equals(cpt[0])) {
									candidates[3] = constructors[i];
								} else if (Throwable.class.equals(cpt[0])) {
									candidates[2] = constructors[i];
								}
								break;
						case 2:
							if (String.class.equals(cpt[0]) && Throwable.class.equals(cpt[1])) {
								candidates[0] = constructors[i];
							} else if (Throwable.class.equals(cpt[0]) && String.class.equals(cpt[1])) {
								candidates[1] = constructors[i];
							}
							break;																		
						}
					}
					
					if (JmsAdapter.isBlank(e.getMessage())) {
						if (candidates[2]!=null) {
							throw (Exception) candidates[2].newInstance(new Object[] {e});
						}
						
						if (candidates[0]!=null) {
							throw (Exception) candidates[0].newInstance(new Object[] {e.getMessage(), e});
						}
						
						if (candidates[1]!=null) {
							throw (Exception) candidates[1].newInstance(new Object[] {e, e.getMessage()});
						}
						
						if (candidates[3]!=null) {
							throw (Exception) candidates[3].newInstance(new Object[] {e.toString()});
						}
					} else {
						if (candidates[0]!=null) {
							throw (Exception) candidates[0].newInstance(new Object[] {e.getMessage(), e});
						}
						
						if (candidates[1]!=null) {
							throw (Exception) candidates[1].newInstance(new Object[] {e, e.getMessage()});
						}
						
						if (candidates[2]!=null) {
							throw (Exception) candidates[2].newInstance(new Object[] {e});
						}
						
						if (candidates[3]!=null) {
							throw (Exception) candidates[3].newInstance(new Object[] {e.toString()});
						}
					}
					
					if (candidates[4]!=null) {
						throw (Exception) candidates[4].newInstance(null);
					}
											
					throw e;
				}
			}
		}
	}
	
	/**
	 * @param proxy
	 * @return Local delegate of the proxy if it is serializable
	 */
	public static Object getProxyState(Object proxy) {
		if (proxy!=null && Proxy.isProxyClass(proxy.getClass())) {
			InvocationHandler ih = Proxy.getInvocationHandler(proxy);
			if (ih  instanceof ProxyServiceInvocationHandler) {
				Object ret = ((ProxyServiceInvocationHandler) ih).getState();
				if (ret instanceof Serializable) {
					return ret;
				}
			}
		}
		return null;
	}
	
	/**
	 * @param proxy
	 * @return Local delegate of the proxy if it is serializable
	 */
	public static void setProxyState(Object proxy, Object state) {
		if (proxy!=null && Proxy.isProxyClass(proxy.getClass())) {
			InvocationHandler ih = Proxy.getInvocationHandler(proxy);
			if (ih  instanceof ProxyServiceInvocationHandler) {
				((ProxyServiceInvocationHandler) ih).setState(state);
			}
		}
	}
	
	private biz.hammurapi.jms.adapter.definition.ProxyService definition;
	
	private Class[] interfaces;
	
	private JmsService jmsService;
	
	JmsService getJmsService() {
		if (definition.getService()!=null) {
			return jmsService;
		} else if (definition.getServiceRef()!=null) {
			return (JmsService) get(definition.getServiceRef());
		} 
		
		return null;
	}
	
	public ProxyService(biz.hammurapi.jms.adapter.definition.ProxyService definition) throws ConfigurationException {
		this.definition = definition;
		if (definition.getService()!=null) {
			jmsService = new JmsService(definition.getService());
			addComponent("jms-service", jmsService);
		}
		
		interfaces = new Class[definition.getInterfaceArray().length];
		for (int i=0; i<interfaces.length; ++i) {
			try {
				interfaces[i] = Class.forName(definition.getInterfaceArray()[i]);
			} catch (ClassNotFoundException e) {
				throw new ConfigurationException("Interface not found: "+definition.getInterfaceArray()[i]+": "+e, e);
			}
		}
		
		for (int i=0; i<definition.getMethodArray().length; ++i) {
			biz.hammurapi.jms.adapter.definition.Method method = definition.getMethodArray(i);
			StringBuffer methodSignature = new StringBuffer(method.getName());
			methodSignature.append("(");
			for (int j=0; j<method.getParameterArray().length; ++j) {
				if (j>0) {
					methodSignature.append(",");
				}
				methodSignature.append(method.getParameterArray(j));
			}
			methodSignature.append(")");
			addComponent(methodSignature.toString(), new Method(method));
		}
		
	}
	
	/**
	 * Creates new proxy instance. Local state is associated with 
	 * the instance.
	 * @return New proxy instance.
	 */
	public Object createProxy() {
		final LocalDelegate localDelegate;
		if (definition.getLocalDelegate()==null) {
			localDelegate = new MapLocalDelegate();
		} else {
			try {
				localDelegate = (LocalDelegate) JmsAdapter.instantiate(definition.getLocalDelegate());
				if (localDelegate instanceof Component) {
					((Component) localDelegate).setOwner(this);
					((Component) localDelegate).start();
				}
				if (localDelegate instanceof MeasurementCollector) {
					MeasurementConsumer cmc = new MeasurementConsumer() {
						public void addMeasurement(String mName, double value, long time) {
							MeasurementConsumer measurementConsumer = getMeasurementConsumer();
							if (measurementConsumer != null) {
								measurementConsumer.addMeasurement("local-delegate."+mName, value, time==0 ? System.currentTimeMillis() : time);
							}						
						}
					};
					((MeasurementCollector) localDelegate).setMeasurementConsumer(cmc);
				}					
			} catch (ConfigurationException e) {
				throw new RuntimeConfigurationException("Cannot create local delegate: "+e, e);
			}
		}

		InvocationHandler ih = new ProxyServiceInvocationHandler(localDelegate);
		
		return Proxy.newProxyInstance(getClass().getClassLoader(), interfaces, ih);
	}

	public static String signature(final java.lang.reflect.Method method) {
		final StringBuffer methodSignature = new StringBuffer(method.getName());
		methodSignature.append("(");
		Class[] parameterTypes=method.getParameterTypes();
		for (int i=0; i<parameterTypes.length; ++i) {
			if (i>0) {
				methodSignature.append(",");
			}
			methodSignature.append(parameterTypes[i].getName());					
		}
		methodSignature.append(")");
		return methodSignature.toString();
	}
	
	/**
	 * Generates proxy service XML by introspecting class which name is passed in the first parameter.
	 * @param args
	 */
	public static final void main(String[] args) throws Exception {
		System.out.println("Usage: java [<options>] "+ProxyService.class.getName()+" <interface name>");
		if (args.length!=1) {
			System.out.println("Invalid number of parameters");
		} else {
			Class iClass = Class.forName(args[0]);
			if (iClass.isInterface()) {
				ProxyServiceDocument doc = ProxyServiceDocument.Factory.newInstance();
				biz.hammurapi.jms.adapter.definition.ProxyService ps = doc.addNewProxyService();
				ps.addInterface(iClass.getName());				
				
				java.lang.reflect.Method[] methods = iClass.getMethods();
				for (int i=0; i<methods.length; ++i) {
					biz.hammurapi.jms.adapter.definition.Method method = ps.addNewMethod();
					method.setName(methods[i].getName());
					Class[] parameterTypes = methods[i].getParameterTypes();
					for (int j=0; j<parameterTypes.length; ++j) {
						method.addParameter(parameterTypes[j].getName());
					}
				}
				System.out.println(doc);
			} else {
				System.out.println(args[0]+" is not an interface");
			}
		}				
	}

	/**
	 * Invokes createProxy()
	 * @return
	 */
	public Object getMaster() {
		return createProxy();
	}
}
