/*
  * hgee 2
 * Hammurapi Group Enterprise Extensions 
 * Copyright (C) 2006  Hammurapi Group
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * URL: http://www.hammurapi.biz/products/enterpriseextensions/index.html
 * e-Mail: support@hammurapi.biz 
 */
package biz.hammurapi.jms.adapter.converters;

import java.io.IOException;
import java.io.StringReader;
import java.io.StringWriter;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.HashMap;
import java.util.Map;

import javax.jms.Message;
import javax.jms.Session;
import javax.jms.TextMessage;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.transform.Result;
import javax.xml.transform.Source;
import javax.xml.transform.Templates;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerConfigurationException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;
import javax.xml.transform.stream.StreamSource;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import biz.hammurapi.config.ComponentBase;
import biz.hammurapi.config.ConfigurationException;
import biz.hammurapi.jms.adapter.Converter;
import biz.hammurapi.jms.adapter.JmsAdapter;
import biz.hammurapi.xml.dom.DOMUtils;

/**
 * This converter supports the following optional conversion properties:
 * <UL>
 * <LI>from-xml-style - URL of stylesheet to apply before XML -&gt; Object conversion. 
 * If URL starts with <code>resource:</code> then it is loaded from classloader resource.</LI>
 * <LI>to-xml-style - URL of stylesheet to apply after Object -&gt; XML conversion. 
 * If URL starts with <code>resource:</code> then it is loaded from classloader resource.</LI>
 * <LI>error-style - URL of stylesheet to apply after Exception -&gt; XML conversion. 
 * If URL starts with <code>resource:</code> then it is loaded from classloader resource.</LI>
 * </UL>
 * @author Tatyana Konukova
 *
 */
public abstract class StylingXmlConverter extends ComponentBase implements Converter {

	private static final String RESOURCE_PREFIX = "resource:";
	private Map templates = new HashMap();
	
	private DocumentBuilder documentBuilder;

	protected synchronized Document createDocument() {
		return documentBuilder.newDocument();
	}
	
	/**
	 * Creates result to receive transformed input.
	 * @return Result instance.
	 */
	protected abstract Result createResult();
	
	/**
	 * Converts reply result to object.
	 * @param result
	 * @return
	 */
	protected abstract Object processResult(Result result) throws Exception;
//		return ((StringWriter) ((StreamResult) result).getWriter()).toString();
	
	/**
	 * Converts request object to Source for styling/serialization
	 * @param request
	 * @return
	 */
	protected abstract Source convert(Object request) throws Exception; 
		
	protected TransformerFactory transformerFactory;

	public void start() throws ConfigurationException {
		try {			
			documentBuilder = DocumentBuilderFactory.newInstance().newDocumentBuilder();
			transformerFactory = TransformerFactory.newInstance();
		} catch (Exception e) {
			throw new ConfigurationException("Startup failed: "+e, e);
		}
	}
	
	protected Transformer getTransformer(String url) throws TransformerConfigurationException, MalformedURLException, IOException {
		if (JmsAdapter.isBlank(url)) {
			return null;
		}
		
		synchronized (templates) {
			Templates template = (Templates) templates.get(url);
			if (template==null) {
				if (url.startsWith(RESOURCE_PREFIX)) {
					template = transformerFactory.newTemplates(
							new StreamSource(
									getClass()
									.getClassLoader()
									.getResourceAsStream(
											url.substring(RESOURCE_PREFIX.length()))));
				} else {
					template = transformerFactory.newTemplates(new StreamSource(new URL(url).openStream()));
				}
				templates.put(url, template);
			}
			return template.newTransformer();
		}
	}

	/**
	 * This method supports <code>from-xml-style</code> optional property - URL of stylesheet to apply before XML -&gt; Object conversion. 
	 * If URL starts with <code>resource:</code> then it is loaded from classloader resource.</LI>
	 */
	public Object convert(Message message, Map properties) throws Exception {
		Result result = createResult();		
		Transformer transformer = getTransformer((String) properties.get("from-xml-style")); 
		if (transformer==null) {
			transformer = transformerFactory.newTransformer();
		}
		transformer.transform(new StreamSource(new StringReader(((TextMessage) message).getText())), result);
		return processResult(result);
	}

	/**
	 * This method supports <code>to-xml-style</code> optional property - URL of stylesheet to apply after Object -&gt; XML conversion. 
	 * If URL starts with <code>resource:</code> then it is loaded from classloader resource.</LI>
	 * Correlates by message ID.
	 */
	public Message convert(Object obj, Session session, Map properties, Message request) throws Exception {
		TextMessage ret = session.createTextMessage();
		Transformer transformer = getTransformer((String) properties.get("to-xml-style")); 
		if (transformer==null) {
			transformer = transformerFactory.newTransformer();
		}
		StringWriter sw = new StringWriter();
		transformer.transform(convert(obj), new StreamResult(sw));
		sw.close();
		ret.setText(sw.toString());
		if (request!=null) {
			ret.setJMSCorrelationID(request.getJMSMessageID());
		}
		return ret;
	}

	/**
	 * This method supports <code>error-style</code> optional property - URL of stylesheet to apply after Exception -&gt; Conversion.
	 * If URL starts with <code>resource:</code> then it is loaded from classloader resource.</LI>
	 * This implementation uses ThrowableSerializer to convert exception to XML.
	 * Correlates by message ID.
	 */
	public Message convert(Exception ex, Session session, Map properties, Message request) throws Exception {
		TextMessage ret = session.createTextMessage();
		Document doc = createDocument();
		Element root = doc.createElement("error");
		doc.appendChild(root);
		DOMUtils.toDom(ex, root);
		
		StringWriter sw = new StringWriter();
		Transformer transformer = getTransformer((String) properties.get("error-style")); 
		if (transformer==null) {
			DOMUtils.serialize(doc, sw);
//			XMLSerializer xmlSerializer = new XMLSerializer(sw, new OutputFormat("xml","UTF-8",false));
//			xmlSerializer.setNamespaces(true);
//			xmlSerializer.serialize(doc);
		} else {
			transformer.transform(new DOMSource(doc), new StreamResult(sw));
		}
		sw.close();
		ret.setText(sw.toString());
		if (request!=null) {
			ret.setJMSCorrelationID(request.getJMSMessageID());
		}
		return ret;
	}

	public void stop() throws ConfigurationException {
		// Nothing to do		
	}

}
