package biz.hammurapi.jms.adapter.converters;

import java.io.StringReader;
import java.io.StringWriter;
import java.util.Map;

import javax.jms.JMSException;
import javax.jms.Message;
import javax.jms.Session;
import javax.jms.TextMessage;

import org.apache.xmlbeans.XmlObject;

import biz.hammurapi.jms.adapter.Converter;
import biz.hammurapi.jms.adapter.JmsAdapter;
import biz.hammurapi.xml.dom.DOMUtils;
import biz.hammurapi.xmlbeans.XmlObjectSerializable;

/**
 * Converts object to message and vice versa using XML Beans.
 * Falls back to dom serialization and domConfigFactory if no XML Beans types are available.
 * @author Pavel
 *
 */
public class XmlBeansConverter implements Converter {

	/**
	 * Parses message content to XmlObject. If message content contains excepiton then this method
	 * tries to throw exception of the same type with the same message.
	 */
	public Object convert(Message message, Map properties) throws Exception {		
		XmlObject ret = XmlObject.Factory.parse(new StringReader(((TextMessage) message).getText()));
		if (ret instanceof biz.hammurapi.invocation.ExceptionDocument) {
			biz.hammurapi.invocation.Exception xex = ((biz.hammurapi.invocation.ExceptionDocument) ret).getException();
			if (!JmsAdapter.isBlank(xex.getType())) {
				Exception toThrow = null;				
				try {
					Class exceptionClass = Class.forName(xex.getType());
					if (Exception.class.isAssignableFrom(exceptionClass)) {
						if (!JmsAdapter.isBlank(xex.getMessage())) {
							toThrow = (Exception) exceptionClass 
								.getConstructor(new Class[] {String.class})
								.newInstance(new String[] {xex.getMessage()});
						} else {
							toThrow = (Exception) exceptionClass.newInstance();
						}
					}
				} catch (Exception e) {
					toThrow = new JMSException("Remote exception "+xex.getType()+": "+xex.getMessage());
				}					
				throw toThrow == null ? new JMSException("Remote exception "+xex.getType()+": "+xex.getMessage()) : toThrow;
			}
			throw new JMSException("Remote exception: "+xex.getMessage());
		}
		return ret;
	}

	/**
	 * Writes XmlObjects using save() method. Uses DOM serialization for other objects.
	 * Converts XmlBeansSerializabe to XmlObjects.
	 */
	public Message convert(Object obj, Session session, Map properties, Message request) throws Exception {
		TextMessage ret = session.createTextMessage();
		if (obj instanceof XmlObject) {
			StringWriter sw = new StringWriter();
			((XmlObject) obj).save(sw);
			sw.close();
			ret.setText(sw.toString());
		} else if (obj instanceof XmlObjectSerializable) {
			StringWriter sw = new StringWriter();
			((XmlObjectSerializable) obj).toXmlObject().save(sw);
			sw.close();
			ret.setText(sw.toString());
		} else {
			ret.setText(DOMUtils.toXmlString(obj, "request"));
		}
		if (request!=null) {
			ret.setJMSCorrelationID(request.getJMSMessageID());
		}
		return ret;
	}

	public Message convert(Exception e, Session session, Map properties, Message request) throws Exception {
		biz.hammurapi.invocation.ExceptionDocument doc = biz.hammurapi.invocation.ExceptionDocument.Factory.newInstance(); 
		biz.hammurapi.invocation.Exception xex = doc.addNewException();
		exceptionToXml(e, xex);
				
		StringWriter sw = new StringWriter();
		doc.save(sw);
		sw.close();
		
		TextMessage ret = session.createTextMessage();
		ret.setText(sw.toString());
		if (request!=null) {
			ret.setJMSCorrelationID(request.getJMSMessageID());
		}
		return ret;
	}
	
	static void exceptionToXml(Throwable ex, biz.hammurapi.invocation.Exception xex) {
		xex.setType(ex.getClass().getName());
		if (ex.getMessage()!=null) {
			xex.setMessage(ex.getMessage());
		}
		
		StackTraceElement[] trace = ex.getStackTrace();
		if (trace!=null) {
			for (int i=0; i<trace.length; ++i) {
				biz.hammurapi.invocation.StackFrame xFrame = xex.addNewStackFrame();
				xFrame.setDeclaringClass(trace[i].getClassName());
				xFrame.setFile(trace[i].getFileName());
				xFrame.setLine(trace[i].getLineNumber());
				xFrame.setMethod(trace[i].getMethodName());
			}
		}
		
		if (ex.getCause()!=null) {
			exceptionToXml(ex.getCause(), xex.addNewCause());
		}
		
	}	

}
