package biz.hammurapi.jms.adapter.jca;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.jms.Message;
import javax.jms.MessageListener;
import javax.resource.spi.endpoint.MessageEndpoint;
import javax.resource.spi.endpoint.MessageEndpointFactory;
import javax.resource.spi.work.Work;
import javax.resource.spi.work.WorkManager;

import biz.hammurapi.jms.adapter.Processor;
import biz.hammurapi.remoting.Invocation;

/**
 * Passes objects to endpoints to be processed in worker threads.
 * If endpoint is instance of MessageListener and object is instance of message then
 * message is passed to the listener.
 * If endpoint is instance of Processor then object is passed to the processor.
 * Otherwise, if object is instance of Invocation then the invocation is executed
 * against the endpoint. 
 * If none of this is true then message is discarded and a warning is logged.
 * In the case of invocation, if the target method returns value and there are multiple endpoints
 * listening then there are two options - a) return value from one of registered endpoints is returned (default) 
 * b) all values are assembled into a collection and returned. This behavior is controlled by "collectAllResults" parameter.
 * The same is true if endpoints provide processor instances. Nulls are not collected and not sent back. Empty results collection
 * is not sent back either. 
 * @author Pavel
 *
 */
public class ActivationProcessor implements Processor {
	
	private static final Logger logger = Logger.getLogger(ActivationProcessor.class.getName());
	
	private WorkManager workManager;
	private Set activations;
	
	public void configure(Set activations, WorkManager workManager) {
		this.workManager = workManager;
		this.activations = activations;
	}
	
	private boolean collectAllResults;

	private long timeout = 30000;
	
	/**
	 * Timeout for collection of results from all work items.
	 * @param timeout
	 */
	public void setTimeout(long timeout) {
		this.timeout = timeout;
	}
	
	public long getTimeout() {
		return timeout;
	}
	
	/**
	 * If set to true then value returned from process is a collection of all 
	 * return
	 * @param collectAllReturns
	 */
	public void setCollectAllResults(boolean collectAllReturns) {
		this.collectAllResults = collectAllReturns;
	}
	
	public boolean isCollectAllResults() {
		return collectAllResults;
	}
	
	/**
	 * Collects non-null returns from registered endpoints and returns them as collection
	 * Returns null if collection is empty.
	 */
	public Object process(final Object obj) throws Exception {
		Collection activationsCopy;
		synchronized (activations) {
			activationsCopy = new ArrayList(activations);
		}
		
		final Collection allResults = collectAllResults ? new ArrayList() : null;
		final int[] workCounter = {0};
		Iterator it = activationsCopy.iterator();
		while (it.hasNext()) {
			Object key = it.next();
			final MessageEndpointFactory factory = ResourceAdapterImpl.factoryFromKey(key);			
			
			if (it.hasNext()) {
				synchronized (workCounter) {
					++workCounter[0];
				}
				
				workManager.scheduleWork(new Work() {

					public void release() {
						// Not implemented							
					}

					public void run() {
						try {
							Object result = processInternal(factory, obj); // By the spec object shall be cloned for each endpoint. we assume the object immutable.
							if (collectAllResults && result!=null) {
								allResults.add(result);
							}
						} catch (Exception e) {
							logger.log(Level.SEVERE, "Exception processing "+obj, e);
						} finally {
							synchronized (workCounter) {
								if (--workCounter[0]==0) {
									workCounter.notifyAll();
								}
							}
						}
					}
					
				});
			} else {
				Object result = processInternal(factory, obj); // By the spec object shall be cloned for each endpoint. we assume the object immutable.
				if (collectAllResults && result!=null) {
					allResults.add(result);
				} else {				
					return processInternal(factory, obj);
				}
			}
		}
		
		synchronized (workCounter) {
			while (workCounter[0]>0) {
				workCounter.wait(timeout );
			}			
		}
		
		return allResults==null || allResults.isEmpty() ? null : allResults;
	}
	
	private Object processInternal(MessageEndpointFactory factory, Object obj) throws Exception {
		MessageEndpoint endpoint = factory.createEndpoint(null);
		try {
			if (endpoint instanceof MessageListener && obj instanceof Message) {
				((MessageListener) endpoint).onMessage((Message) obj);
				return null;
			}
			
			if (obj instanceof Invocation) {
				return ((Invocation) obj).invoke(endpoint);
			}
			
			if (endpoint instanceof Processor) {
				return ((Processor) endpoint).process(obj);
			}
			
			logger.warning("Don't know how to process "+obj.getClass().getName()+" with "+endpoint.getClass().getName());
			return null;
		} finally {
			endpoint.release();
		}		
	}

}
