package biz.hammurapi.jms.adapter.jca;

import java.net.URL;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.resource.ResourceException;
import javax.resource.spi.ActivationSpec;
import javax.resource.spi.BootstrapContext;
import javax.resource.spi.ResourceAdapter;
import javax.resource.spi.ResourceAdapterInternalException;
import javax.resource.spi.endpoint.MessageEndpointFactory;
import javax.resource.spi.work.Work;
import javax.resource.spi.work.WorkEvent;
import javax.resource.spi.work.WorkException;
import javax.resource.spi.work.WorkListener;
import javax.resource.spi.work.WorkManager;
import javax.transaction.xa.XAResource;

import org.apache.xmlbeans.XmlObject;

import biz.hammurapi.config.ConfigurationException;
import biz.hammurapi.jms.adapter.JmsAdapter;
import biz.hammurapi.jms.adapter.definition.JmsAdapterDocument;
import biz.hammurapi.util.Worker;

public class ResourceAdapterImpl implements ResourceAdapter {
	
	private Logger logger = Logger.getLogger(ResourceAdapterImpl.class.getName());

	public SharedAdapter getSharedAdapter() {
		// TODO Add reference counting when auto-reload of configuration is implemented.
		return new SharedAdapter() {

			public void close() {
				// TODO Decrement counter and stop adapter if it is scheduled to stop.
				
			}

			public Object get(String name) {				
				return adapter.get(name);
			}
			
		};
	}
	
	private Map activationMap = new HashMap();
	private WorkManager workManager;
	
	private static Object activationKey(MessageEndpointFactory factory, ActivationSpec spec) {
		ArrayList ret = new ArrayList();
		ret.add(factory);
		ret.add(spec);
		return ret;
	}
	
	static MessageEndpointFactory factoryFromKey(Object key) {
		return (MessageEndpointFactory) ((List) key).get(0);
	}

	public void endpointActivation(MessageEndpointFactory factory, ActivationSpec spec) throws ResourceException {
		if (!(spec instanceof ActivationSpecImpl)) {
			throw new ResourceException("Invalid activation specification: "+spec);
		}		
		
		if (adapter==null) {
			throw new ResourceException("Adapter is not properly initialized");
		}
		
		String listenerName = ((ActivationSpecImpl) spec).getListenerName();
		synchronized (activationMap) {
			Set endpointSet = (Set) activationMap.get(listenerName);
			if (endpointSet==null) {				
				Object processor = adapter.get(listenerName);
				if (processor == null) {
					throw new ResourceException("Listener not found: "+listenerName);
				}
				
				if (!(processor instanceof ActivationProcessor)) {
					throw new ResourceException("Unexpected listener processor type: "+processor.getClass());
				}
				
				endpointSet = new HashSet();				
				activationMap.put(listenerName, endpointSet);
				((ActivationProcessor) processor).configure(endpointSet, workManager);
			}
			
			synchronized (endpointSet) {
				endpointSet.add(activationKey(factory, spec));
			}
		}
				
	}

	/**
	 * Removes endpoint entry from activation maps.
	 */
	public void endpointDeactivation(MessageEndpointFactory factory, ActivationSpec spec) {
		if (spec instanceof ActivationSpecImpl) {
			synchronized (activationMap) {
				Set endpointSet = (Set) activationMap.get(((ActivationSpecImpl) spec).getListenerName());
				if (endpointSet!=null) {
					synchronized (endpointSet) {
						endpointSet.remove(activationKey(factory, spec));
					}					
				}
			}
		}
	}

	public XAResource[] getXAResources(ActivationSpec[] speca) throws ResourceException {
		// Not implemented - transactions are not supported
		return null;
	}
	
	private String configUrl;
	private JmsAdapter adapter;
	
	/**
	 * URL of configuration. If it starts with <code>resource:</code> then it is read from classloader resource.
	 * @param configUrl
	 */
	public void setConfigUrl(String configUrl) {
		this.configUrl = configUrl;
	}
	
	// TODO - reload of configuration if changed - reload interval.
	
	private static final String RESOURCE_PREFIX = "resource:";

	public void start(BootstrapContext context) throws ResourceAdapterInternalException {
        if (configUrl==null) {
        	throw new ResourceAdapterInternalException("Configuration location is not set");
        }
        
        try {
        	URL confURL = configUrl.startsWith(RESOURCE_PREFIX) ? getClass().getClassLoader().getResource(configUrl.substring(RESOURCE_PREFIX.length())) : new URL(configUrl);
			XmlObject document = XmlObject.Factory.parse(confURL);
	        if (!(document instanceof JmsAdapterDocument)) {
	        	throw new ResourceAdapterInternalException("Configuration document is not of expected type");
	        }
	        
	        biz.hammurapi.jms.adapter.definition.JmsAdapter definition = ((JmsAdapterDocument) document).getJmsAdapter();
	        adapter = new JmsAdapter(definition);
	        adapter.start();
	            
			workManager = context.getWorkManager();
			
			if (workManager!=null) {
				final WorkListener workListener = new WorkListener() {
					
					public void workAccepted(WorkEvent workEvent) {
						logger.fine("Work accepted: "+workEvent.getWork());
					}
	
					public void workCompleted(WorkEvent workEvent) {
						logger.fine("Work completed: "+workEvent.getWork());
					}
	
					public void workRejected(WorkEvent workEvent) {
						logger.log(Level.SEVERE, "Work rejected: "+workEvent.getWork(), workEvent.getException());
					}
	
					public void workStarted(WorkEvent workEvent) {
						logger.fine("Work started: "+workEvent.getWork());
					}
					
				};
				
				Worker raWorker = new Worker() {
	
					public boolean post(final Runnable job) {
						
						Work work = new Work() {
							
							public void run() {
								job.run();
							}
	
							public void release() {
								// Not supported							
							}
							
							public String toString() {
								return "Work for "+job;
							}
						};
											
						try {
							workManager.scheduleWork(work, WorkManager.INDEFINITE, null, workListener);
						} catch (WorkException e) {
							logger.log(Level.SEVERE, "Failed to schedule work", e);
							return false;
						}
						return true;
					}
					
				};
				
				adapter.setDefaultWorker(raWorker);
			}
        } catch (Exception e) {
        	throw new ResourceAdapterInternalException("Failed to start adapter: "+e, e);
        }
	}

	public void stop() {
		if (adapter!=null) {
			try {
				adapter.stop();
			} catch (ConfigurationException e) {
				logger.log(Level.SEVERE, "Failed to properly stop adapter: "+e, e);
			}
		}
		
	}
	
}
