package biz.hammurapi.metrics;

import java.io.Serializable;
import java.util.LinkedList;
import java.util.List;
import java.util.Timer;
import java.util.TimerTask;

import biz.hammurapi.config.Component;
import biz.hammurapi.config.ConfigurationException;
import biz.hammurapi.metrics.SimpleSlice;
import biz.hammurapi.metrics.Slice;
import biz.hammurapi.metrics.SliceConsumer;


/**
 * Collects slices for a period of time and processes them in batches.
 * @author Pavel Vlasov
 */
public abstract class BatchingSliceConsumer implements SliceConsumer, Component {
	private static final int MAX_SLICES = 10000; // maximum number of slices to keep in the send queue
	
	private int maxSlices = MAX_SLICES;
	private long interval = 60000; // Default batching period is one minute.
	
	/**
	 * Sets maximum number of slices to be retained if processing fails.
	 * @param maxSlices
	 */
	public void setMaxSlices(int maxSlices) {
		this.maxSlices = maxSlices;
	}
	
	/**
	 * Sets batching interval. Default interval is one minute.
	 * @param interval
	 */
	public void setInterval(long interval) {
		this.interval = interval;
	}
	
	private LinkedList slices = new LinkedList();
	
	private static final Timer timer = new Timer(true);
	private Thread shutdownHook;

	private TimerTask task;
	
	/**
	 * Helper class to serialize accumulated slices.
	 * @author Pavel Vlasov
	 */
	public static class SliceEntry implements Serializable {
		String category;
		Slice slice;
		
		SliceEntry(String category, Slice slice) {
			this.category = category;
			if (slice instanceof SimpleSlice) {
				this.slice = slice;
			} else {
				this.slice = new SimpleSlice(slice);
			}
		}
		
		public String getCategory() {
			return category;
		}
		
		public Slice getSlice() {
			return slice;
		}
	}

	public synchronized boolean consumeSlice(String category, Slice slice) {
		slices.add(new SliceEntry(category, slice));
		return true;
	}

	public void setOwner(Object owner) {
		// Nothing
		
	}

	public void start() throws ConfigurationException {
		if (interval<=0) {
			throw new ConfigurationException("Invalid interval: "+interval);
		}
		task = new TimerTask() {

			public void run() {
				LinkedList slicesToProcess;
				synchronized (BatchingSliceConsumer.this) {
					if (slices.isEmpty()) {
						return;
					}
					slicesToProcess = slices;
					slices = new LinkedList();
				}
				
				if (!processSlices(slicesToProcess)) {
					synchronized (BatchingSliceConsumer.this) {
						slicesToProcess.addAll(slices);
						slices = slicesToProcess;
						while (slicesToProcess.size()>maxSlices) {
							slicesToProcess.removeFirst();
						}
					}
				}				
			}
			
		};
		timer.schedule(task, interval, interval);
		shutdownHook = new Thread(task);
		Runtime.getRuntime().addShutdownHook(shutdownHook);
	}

	public void stop() throws ConfigurationException {
		if (task!=null) {
			task.cancel();
			if (shutdownHook!=null) {
				try {
					Runtime.getRuntime().removeShutdownHook(shutdownHook);
				} catch (Exception e) {
					// We don't care
				}
			}
			task.run();
		}		
	}
	
	/**
	 * Subclasses shall implement this method.
	 * @param slices
	 * @return true if slices was successfully processed.
	 */
	protected abstract boolean processSlices(List slices);

}
