package biz.hammurapi.metrics;

import java.io.IOException;
import java.io.ObjectInputStream;
import java.net.InetAddress;
import java.net.ServerSocket;
import java.net.Socket;
import java.util.Iterator;
import java.util.List;

import biz.hammurapi.config.Component;
import biz.hammurapi.config.ConfigurationException;
import biz.hammurapi.metrics.ConsoleSliceConsumer;
import biz.hammurapi.metrics.SliceConsumer;
import biz.hammurapi.metrics.BatchingSliceConsumer.SliceEntry;


public class SocketSliceConsumerServer implements Component {
	
	private SliceConsumer delegate = new ConsoleSliceConsumer();
	
	private int port = SocketSliceConsumer.DEFAULT_PORT;
	
	/**
	 * Set server listening port
	 * @param port
	 */
	public void setPort(int port) {
		this.port = port;
	}
	
	ServerSocket serverSocket;

	public void setOwner(Object owner) {
		// Nothing

	}

	public void start() throws ConfigurationException {
		try {
			serverSocket = new ServerSocket(port);
			Thread listeningThread = new Thread() {
				public void run() {
					while (true) {
						try {
							process(serverSocket.accept());
						} catch (IOException e) {
							consumeException(e);
							return;
						}
					}
				}
			};
			listeningThread.setDaemon(true);
			listeningThread.setName("Slice consumer listener");
			listeningThread.start();
		} catch (IOException e) {
			throw new ConfigurationException("Could not create server socket: "+e,e);
		}
	}

	public void stop() throws ConfigurationException {
		if (serverSocket!=null) {
			try {
				serverSocket.close();
			} catch (IOException e) {
				throw new ConfigurationException("Could not close server socket: "+e, e);
			}
		}
	}
	
	/**
	 * @return SliceConsumer to delegate processing. This class returns ConsoleSliceConsumer.
	 * Subclasses can override this method.
	 */
	protected SliceConsumer getDelegate() {
		return delegate;
	}
	
	/**
	 * This method iterates over slices in the batch
	 * and sends them to the delegate SliceConsumer returned
	 * by getDelegate() method.  
	 * @param slices List of slice entries obtained from clients
	 */
	protected void processSlices(List slices, InetAddress address, String id) {
		Iterator it = slices.iterator();
		while (it.hasNext()) {
			SliceEntry se = (SliceEntry) it.next();
			getDelegate().consumeSlice("["+address+":"+id+"] "+se.getCategory(), se.getSlice());
		}		
	}
	
	/**
	 * Spawns a new thread for each socket
	 * @param socket
	 */
	protected void process(final Socket socket) {
		Runnable job = new Runnable() {
			public void run() {
				try {
					ObjectInputStream ois = new ObjectInputStream(socket.getInputStream());
					try {
						String id = (String) ois.readObject();
						processSlices((List) ois.readObject(), socket.getInetAddress(), id);
					} finally {
						ois.close();
						socket.close();
					}
				} catch (Exception e) {
					consumeException(e);
				}
			}
		};
		process(job);		
	}
	
	/**
	 * Spawns a new thread for each job. 
	 * Override to use thread pools on high loads.
	 * @param job
	 */
	protected void process(Runnable job) {
		new Thread(job, "Slice consumer").start();
	}
	
	/**
	 * Consumes exceptions in processing threads.
	 * @param e
	 */
	protected void consumeException(Exception e) {
		e.printStackTrace();		
	}
	
	public static void main(String[] args) throws ConfigurationException {
		System.out.println("Usage: java [options] biz.hammurapi.metrics.SocketSliceConsumerServer [port]");
		int port = SocketSliceConsumer.DEFAULT_PORT;
		System.out.println("Default port: "+port);
		if (args.length>0) {
			try {
				port = Integer.parseInt(args[0]);
			} catch (NumberFormatException e) {
				System.err.println("Invalid port number: "+args[0]);
				System.exit(1);
			}
		}
		
		SocketSliceConsumerServer server = new SocketSliceConsumerServer();
		server.setPort(port);
		server.start();
	}

}
