/*
 * hgee 2
 * Hammurapi Group Enterprise Extensions 
 * Copyright (C) 2006  Hammurapi Group
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * URL: http://www.hammurapi.biz/products/enterpriseextensions/index.html
 * e-Mail: support@hammurapi.biz
 */
package biz.hammurapi.metrics.persistent;

import java.sql.SQLException;
import java.text.MessageFormat;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import org.w3c.dom.Element;

import biz.hammurapi.sql.SQLRuntimeException;
import biz.hammurapi.util.Visitable;
import biz.hammurapi.util.Visitor;
import biz.hammurapi.xml.dom.DomSerializable;


/**
 * @author Pavel Vlasov
 * @revision $Revision$
 */
public abstract class AbstractPeriod implements DomSerializable, Visitable {

	protected long timeFrom;
	protected long timeTo;
	protected Slice[] slices;
	protected PeriodFactory factory;
	private double maxValue;
	private double minValue;
	private double deviation;
	private double total;
	private double measurements;

	/**
	 * @return Returns the deviation.
	 */
	public double getDeviation() {
		return measurements==0 ? 0 : deviation/measurements;
	}

	/**
	 * @return Returns the from.
	 */
	public long getTimeFrom() {
		return timeFrom;
	}

	/**
	 * @return Returns the maxValue.
	 */
	public double getMaxValue() {
		return maxValue;
	}

	/**
	 * @return Returns the measurements.
	 */
	public double getMeasurements() {
		return measurements;
	}

	/**
	 * @return Returns the minValue.
	 */
	public double getMinValue() {
		return minValue;
	}

	/**
	 * @return Returns the to.
	 */
	public long getTimeTo() {
		return timeTo;
	}

	/**
	 * @return Returns the total.
	 */
	public double getTotal() {
		return total;
	}

	/**
	 * @return Returns the slices.
	 */
	public Slice[] getSlices() {
		return slices;
	}
	
    private List subCategories;	

	public Collection getSubCategories() throws SQLException {
		if (subCategories==null) {
			subCategories=new ArrayList();
			loadSubCategories(subCategories);
			Collections.sort(subCategories, new Comparator() {

				public int compare(Object period0, Object period1) {
					double total0 = ((AbstractPeriod) period0).getTotal();
					double total1 = ((AbstractPeriod) period1).getTotal();
					return total0==total1 ? 0 : total0>total1 ? -1 : 1;
				}
				
			});
			
			Iterator it=subCategories.iterator();
			for (int i=1; it.hasNext(); i++) {
				Period p=(Period) it.next();
				labelMap.put(new Integer(p.getId()), p.getName().length()>15 ? "Category "+i : p.getName());
			}
			
		}
		
		return subCategories;
	}

	/**
	 * This method should add subcategories (of type Period) to this category.
	 * @throws SQLException
	 */
	protected abstract void loadSubCategories(List subCategories) throws SQLException;

	public void toDom(Element holder) {
		setAttributes(holder);
		for (int i=0; i<slices.length; i++) {
			Element se=holder.getOwnerDocument().createElement("slice");
			holder.appendChild(se);
			slices[i].toDom(se);
			se.setAttribute("avg", String.valueOf(slices[i].getMeasurements()==0 ? 0 : slices[i].getTotal()/slices[i].getMeasurements()));
			se.setAttribute("length", String.valueOf(slices[i].getTimeTo()-slices[i].getTimeFrom()));
		}
		
		try {
			Iterator it=getSubCategories().iterator();
			while (it.hasNext()) {
				Element e=holder.getOwnerDocument().createElement("category");
				holder.appendChild(e);
				AbstractPeriod period = (AbstractPeriod) it.next();
				period.setAttributes(e);
				e.setAttribute("pct", MessageFormat.format("{0,number,#.##}", new Object[] {new Double(100*period.getTotal()/getTotal())}));
			}
		} catch (SQLException e) {
			throw new SQLRuntimeException(e);
		}				
	}
	
	private Map labelMap=new HashMap();
	
	/**
	 * @param childId
	 * @return
	 */
	public String getLabel(int childId) {
		return (String) labelMap.get(new Integer(childId));
	}	

	/**
	 * @param holder
	 */
	protected void setAttributes(Element holder) {
		holder.setAttribute("from", new Date(timeFrom).toString());
		holder.setAttribute("from-msec", String.valueOf(timeFrom));
		holder.setAttribute("to", new Date(timeTo).toString());
		holder.setAttribute("to-msec", String.valueOf(timeTo));
		holder.setAttribute("max", String.valueOf(maxValue));
		holder.setAttribute("min", String.valueOf(minValue));
		holder.setAttribute("deviation", String.valueOf(getDeviation()));
		holder.setAttribute("total", String.valueOf(total));
		holder.setAttribute("measurements", String.valueOf(measurements));
		holder.setAttribute("avg", String.valueOf(total/measurements));
		holder.setAttribute("length", String.valueOf(getLength()));
		holder.setAttribute("name", name);
	}

	/**
	 * @param id
	 * @param slices
	 */
	protected void createSlices(int slices) {
		this.slices=new Slice[slices];
		sliceInitialized=new boolean[slices];
	    for (int i=0; i<slices; i++) {
	        this.slices[i]=new SliceImpl();
	        this.slices[i].setTimeFrom((long) ((((double) this.timeFrom)*(slices-i)+((double) this.timeTo)*i)/slices));
	        if (i==slices-1) {
	            this.slices[i].setTimeTo(this.timeTo);
	        } 
	        
	        if (i>0) {
	            this.slices[i-1].setTimeTo(this.slices[i].getTimeFrom()-1);
	        }
	        
	        this.slices[i].setId(i);
	    }
	}

    private boolean initialized;
    private boolean[] sliceInitialized;
    
	/**
	 * @param slice
	 */
	protected void aggregate(Slice slice) {
		long interval=Math.min(slice.getTimeTo(), this.timeTo)-Math.max(slice.getTimeFrom(), this.timeFrom);
		if (interval>=0) {
			if (!initialized) {
				minValue=slice.getMinValue();
				maxValue=slice.getMaxValue();
				initialized=true;
			} else {
				minValue=Math.min(minValue, slice.getMinValue());
				maxValue=Math.max(maxValue, slice.getMaxValue());
			}
			
			double coeff = interval==slice.getTimeTo()-slice.getTimeFrom() ? 1 : ((double) interval)/(double)(slice.getTimeTo()-slice.getTimeFrom());
			deviation+=slice.getDeviation()*slice.getMeasurements()*coeff;
			total+=slice.getTotal()*coeff;
			measurements+=slice.getMeasurements()*coeff;
		}
		
		for (int i=0; i<this.slices.length; i++) {
			Slice curSlice = this.slices[i];
			long sinterval=Math.min(slice.getTimeTo(), curSlice.getTimeTo())-Math.max(slice.getTimeFrom(), curSlice.getTimeFrom());
			if (sinterval>=0) {
		    	if (sliceInitialized[i]) {
		    		curSlice.setMinValue(Math.min(curSlice.getMinValue(), slice.getMinValue()));
		    		curSlice.setMaxValue(Math.max(curSlice.getMaxValue(), slice.getMaxValue()));
		    	} else {
		    		curSlice.setMinValue(slice.getMinValue());
		    		curSlice.setMaxValue(slice.getMaxValue());
		    		sliceInitialized[i]=true;
		    	}
		    	
				double scoeff = sinterval==slice.getTimeTo()-slice.getTimeFrom() ? 1 : ((double) sinterval)/(double)(slice.getTimeTo()-slice.getTimeFrom());
		    	curSlice.setDeviation(curSlice.getDeviation()+slice.getDeviation()*slice.getMeasurements()*scoeff);
		    	curSlice.setTotal(curSlice.getTotal()+slice.getTotal()*scoeff);
		    	curSlice.setMeasurements(curSlice.getMeasurements()+slice.getMeasurements()*scoeff);
			}			
		}
	}

	/**
	 * Invoke after all aggregations.
	 */
	protected void normalizeSliceDeviations() {
		for (int i=0; i<this.slices.length; i++) {
	    	Slice curSlice = this.slices[i];
			curSlice.setDeviation(curSlice.getMeasurements()==0 ? 0 : curSlice.getDeviation()/curSlice.getMeasurements());
	    }
	}  
	
	public long getLength() {
		return timeTo-timeFrom;
	}
	
	public boolean accept(Visitor visitor) {
		if (visitor.visit(this)) {
			try {
				Iterator it=getSubCategories().iterator();
				while (it.hasNext()) {
					((Visitable) it.next()).accept(visitor);
				}
			} catch (SQLException e) {
				throw new SQLRuntimeException(e);
			}
			return true;
		}
		
		return false;
	}
	
	/**
	 * @return Returns the name.
	 */
	public String getName() {
		return name;
	}
	
	protected String name;
		
}
