/*
 * hgee 2
 * Hammurapi Group Enterprise Extensions 
 * Copyright (C) 2006  Hammurapi Group
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * URL: http://www.hammurapi.biz/products/enterpriseextensions/index.html
 * e-Mail: support@hammurapi.biz
 */
package biz.hammurapi.metrics.persistent;

import java.sql.SQLException;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import org.w3c.dom.Element;

import biz.hammurapi.sql.SQLRuntimeException;

/**
 * Measurement period for a particular metric
 * @author Pavel Vlasov
 * @version $Revision$
 */
public class Period extends AbstractPeriod {
    private int id;
	/**
	 * Period primary key
	 */
	static class Key {
		private int id;
		private int hashCode;
		private long timeFrom;
		private long timeTo;
		private int slices;
		private boolean shrink;
		
		/**
		 * @param id
		 * @param timeFrom
		 * @param timeTo
		 */
		public Key(int id, long timeFrom, long timeTo, int slices, boolean shrink) {
			super();
			this.id = id;
			this.timeFrom = timeFrom;
			this.timeTo = timeTo;
	    	hashCode=id ^ new Long(timeFrom).hashCode() ^ new Long(timeTo).hashCode() ^ slices ^ new Boolean(shrink).hashCode();
		}
		
		public boolean equals(Object arg) {
			return (arg==this || 
					(arg instanceof Key 
							&& ((Key) arg).id==id 
							&& ((Key) arg).timeFrom==timeFrom 
							&& ((Key) arg).timeTo==timeTo
							&& ((Key) arg).slices==slices
							&& ((Key) arg).shrink==shrink));
		}
		
		public int hashCode() {
			return hashCode;
		}
	}
	
    /**
     * Actual from and to can differ from the passed if shrink is true slices don't cover all the period.
     * @param processor
     * @param id
     * @param from
     * @param to
     * @param periods
     * @param shrink Shrink period to exclude leading and trailing times without measurements.
     * @throws SQLException
     */
    Period(PeriodFactory factory, Key key) throws SQLException {
    	this(factory, null, key.id, key.timeFrom, key.timeTo, key.slices, key.shrink);    	
    	name=factory.getEngine().getMetric(id).getName();
    }
    
    /**
     * Actual from and to can differ from the passed if shrink is true slices don't cover all the period.
     * @param processor
     * @param id
     * @param from
     * @param to
     * @param periods
     * @param shrink Shrink period to exclude leading and trailing times without measurements.
     * @throws SQLException
     */
    Period(PeriodFactory factory, int id, long from, long to, int slices, boolean shrink) throws SQLException {
    	this(factory, null, id, from, to, slices, shrink);
    	name=factory.getEngine().getMetric(id).getName();
    }
    
    /**
     * Actual from and to can differ from the passed if shrink is true slices don't cover all the period.
     * @param processor
     * @param id
     * @param from
     * @param to
     * @param periods
     * @param shrink Shrink period to exclude leading and trailing times without measurements.
     * @throws SQLException
     */
    private Period(PeriodFactory factory, String name, int id, long from, long to, int slices, boolean shrink) throws SQLException {
    	this.name=name;
    	this.id=id;
    	MetricsEngine engine = factory.getEngine();
    	this.factory=factory;
    	
    	key=new Key(id, from, to, slices, shrink);
    	
        this.timeFrom=from;
        this.timeTo=to;
        if (shrink) {
            ActualPeriod actualPeriod = engine.getActualPeriod(id, to, from);
			if (actualPeriod!=null && actualPeriod.getTimeFrom()!=actualPeriod.getTimeTo()) {
            	if (actualPeriod.getTimeFrom()>from) {
            		this.timeFrom=actualPeriod.getTimeFrom()-1;            		
            	}
            	
            	if (actualPeriod.getTimeTo()<to) {
            		this.timeTo=actualPeriod.getTimeTo()+1;
            	}
            }
        }
        
        createSlices(slices);
        
        // calculate values for slices and totals here
        Iterator it=engine.getSlice(id, this.timeTo, this.timeFrom).iterator();
        while (it.hasNext()) {
			aggregate((Slice) it.next());
        }
                
        normalizeSliceDeviations();
    }
    
	private Key key;
	
	/**
	 * @return Returns the id.
	 */
	public int getId() {
		return id;
	}

	public boolean equals(Object arg) {
		return arg==this || (arg instanceof Period	&& key.equals(((Period) arg).toKey())); 
	}
	
	public int hashCode() {
		return key.hashCode;
	}
	
	/**
	 * Primary key of period to use in memory sensitive cache.
	 * @return
	 */
	public Object toKey() {
		return key;
	}

	/**
	 * @param holder
	 */
	protected void setAttributes(Element holder) {
		super.setAttributes(holder);
		holder.setAttribute("id", String.valueOf(id));
		holder.setAttribute("absolute-name", getAbsoluteName());
	}
	
	public void toDom(Element holder) {
		super.toDom(holder);
		MetricRef[] path = getPath();
		for (int i=0; i<path.length; i++) {
			Element pe=holder.getOwnerDocument().createElement("path");
			holder.appendChild(pe);
			pe.setAttribute("id", String.valueOf(path[i].getId()));
			pe.setAttribute("name", path[i].getName());
		}
	}

	/**
	 * @throws SQLException
	 */
	protected void loadSubCategories(List subCategories) throws SQLException {
		Iterator it = factory.getEngine().getActiveMetrics(getId(), getTimeTo(), getTimeFrom()).iterator();
		while (it.hasNext()) {
			Period subPeriod = factory.getPeriod(((Number) it.next()).intValue(), getTimeFrom(), getTimeTo(), slices.length, false);
			if (subPeriod.getMeasurements()!=0) {
				subCategories.add(subPeriod);
			}
		}		
	}

	public String getAbsoluteName() {
		StringBuffer ret=new StringBuffer();
		MetricRef[] path = getPath();
		for (int i=0; i<path.length; i++) {
			ret.append(path[i].getName());
			ret.append(".");
		}		
		ret.append(getName());
		return ret.toString();
	}
	
	private MetricRef[] path;
	
	public MetricRef[] getPath() {
		if (path==null) {
			LinkedList pathList=new LinkedList();
			try {
				for (Metric m=factory.getEngine().getMetric(getId()); m!=null; m = m.getParent()==null ? null : factory.getEngine().getMetric(m.getParent().intValue())) {
					if (m.getId()!=getId()) {
						pathList.addFirst(m);
					}
				}
			} catch (SQLException e) {
				throw new SQLRuntimeException(e);
			}
			path=(MetricRef[]) pathList.toArray(new MetricRef[pathList.size()]);
		}
		return path;
	}

}
