/*
  * hgee 2
 * Hammurapi Group Enterprise Extensions 
 * Copyright (C) 2006  Hammurapi Group
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * URL: http://www.hammurapi.biz/products/enterpriseextensions/index.html
 * e-Mail: support@hammurapi.biz
  */
package biz.hammurapi.metrics.persistent;

import java.sql.Connection;
import java.sql.SQLException;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.StringTokenizer;

import biz.hammurapi.convert.CompositeConverter;
import biz.hammurapi.metrics.HousekeepingSliceConsumer;
import biz.hammurapi.metrics.Slice;
import biz.hammurapi.sql.IdentityGenerator;
import biz.hammurapi.sql.IdentityManager;
import biz.hammurapi.sql.IdentityRetriever;
import biz.hammurapi.sql.SQLProcessor;


/**
 * @author Pavel Vlasov
 *
 * @version $Revision: 1.2 $
 */
public class PersistingSliceConsumer implements HousekeepingSliceConsumer {
    public static final long SECOND=1000;
    public static final long MINUTE=60*SECOND;
    public static final long HOUR=60*MINUTE;
    public static final long DAY=24*HOUR;
    public static final long WEEK=1000;
    public static final long MONTH=1000;
    public static final long YEAR=1000;

	private SQLProcessor processor;
	protected boolean isActive=false;
	protected Number rootId;
    private long history;
    private long maintenanceInterval;
	private IdentityManager identityManager;


	/**
	 * @param tick
	 * @param keepMeasurements
	 * @throws SQLException
	 */
	public PersistingSliceConsumer(
	        String category,
	        long history,
	        long maintenanceInterval,
	        SQLProcessor processor,
			IdentityManager identityManager) throws SQLException {
		this.processor=processor;
		if (identityManager==null) {
			this.identityManager=new IdentityGenerator() {

				public int generate(Connection con, String name) throws SQLException {
					return PersistingSliceConsumer.this.processor.nextPK(con, "PRIMARY_KEY", name);
				}
				
			};
		} else {
			this.identityManager=identityManager;
		}
		this.history=history;
		this.maintenanceInterval=maintenanceInterval;
		if (category!=null) {
			Connection con=processor.getConnection();
			boolean cac=con.getAutoCommit();
			try {
				con.setAutoCommit(false);
				MetricsEngine tEngine=new MetricsEngine(new SQLProcessor(con, processor.getNameMap()));
				StringTokenizer st=new StringTokenizer(category, ".");
				while (st.hasMoreTokens()) {
					String token=st.nextToken();
					Number newRoot;
					if (rootId==null) {
						newRoot=tEngine.getRootMetricId(token);
					} else {
						newRoot=tEngine.getMetricId(rootId.intValue(), token);
					}
					
					if (newRoot==null) {
						MetricImpl metric=new MetricImpl(true);
						metric.setName(token);
						if (rootId!=null) {
							metric.setParent((Integer) CompositeConverter.getDefaultConverter().convert(rootId, Integer.class, false));
						}
						
						insertMetric(con, tEngine, metric);						
						insertKindred(tEngine, rootId, metric.getId());
						newRoot=new Integer(metric.getId());
					}
					
					rootId=newRoot;
				}
				con.commit();
			} catch (SQLException e) {
				con.rollback();
				throw e;
			} finally {				
				con.setAutoCommit(cac);
				processor.releaseConnection(con);
			}
		}
		isActive=true;
	}

    /**
	 * @param con
	 * @param tEngine
	 * @param metric
	 * @throws SQLException
	 */
	private void insertMetric(Connection con, MetricsEngine tEngine, MetricImpl metric) throws SQLException {
		if (this.identityManager instanceof IdentityGenerator) {
			metric.setId(((IdentityGenerator) this.identityManager).generate(con, "METRIC"));
		}						
		tEngine.insertMetric(metric);
		if (this.identityManager instanceof IdentityRetriever) {
			metric.setId(((IdentityRetriever) this.identityManager).retrieve(con));
		}
	}

	/**
     * Breaks metric name into individual tokens. E.g. <code>biz.hammurapi.sql</code> could be
     * broken into <code>com, pavelvlasov</code>, and <code>sql</code>
     * Default implementation does nothing - it returns <code>new String[] {name}</code>
     * @return metric name broken into individual elements
     */
    protected String[] tokenize(String name) {
        return new String[] {name};
    }
    
    private Map categoryMap=new HashMap();
    
    private Number getCategoryId(Connection con, MetricsEngine engine, String category) throws SQLException {
    	if (category==null || category.length()==0) {
    		return rootId;
    	}
    	
    	Number ret = (Number) categoryMap.get(category);
    	if (ret==null) {
	    	int idx=category.lastIndexOf('.');
	    	if (idx==-1) {	    		
	    		ret = rootId==null ? engine.getRootMetricId(category) : engine.getMetricId(rootId.intValue(), category);
	    		if (ret==null) {
					MetricImpl metric=new MetricImpl(true);
					metric.setName(category);
					if (rootId!=null) {
						metric.setParent((Integer) CompositeConverter.getDefaultConverter().convert(rootId, Integer.class, false));
					}
					insertMetric(con, engine, metric);
					insertKindred(engine, metric.getParent(), metric.getId());
					ret=new Integer(metric.getId());
	    		}
	    	} else {
				Number parentId = getCategoryId(con, engine, category.substring(0, idx));
				MetricImpl metric=new MetricImpl(true);
				metric.setName(category.substring(idx+1));
				metric.setParent((Integer) CompositeConverter.getDefaultConverter().convert(parentId, Integer.class, false));
				insertMetric(con, engine, metric);
				insertKindred(engine, parentId, metric.getId());
				ret=new Integer(metric.getId());
	    	}
	    	categoryMap.put(category, ret);
    	}
    	
    	return ret;
    }

    /**
	 * @param parentId
	 * @param metricId
     * @param engine
	 * @throws SQLException
	 */
	private void insertKindred(MetricsEngine engine, Number parent, int metricId) throws SQLException {
		engine.insertKindred(metricId, metricId);
		if (parent!=null) {
			engine.newRelative(metricId,parent.intValue());
		}
	}

	public boolean consumeSlice(String category, Slice slice) {
		if (isActive) {
			try {
				Connection con=processor.getConnection();
				boolean cac=con.getAutoCommit();
				try {
					con.setAutoCommit(false);
					MetricsEngine tEngine=new MetricsEngine(new SQLProcessor(con, processor.getNameMap()));
				    Number metricId=getCategoryId(con, tEngine, category);
				    String[] tokens=tokenize(slice.getName()==null ? "(null)" : slice.getName());
				    for (int i=0; i<tokens.length; i++) {
			            Number newMetricId = metricId==null ? tEngine.getRootMetricId(tokens[i]) : tEngine.getMetricId(metricId.intValue(), tokens[i]);
						if (newMetricId==null) {
							MetricImpl metric=new MetricImpl(true);
							metric.setName(tokens[i]);
							metric.setParent((Integer) CompositeConverter.getDefaultConverter().convert(metricId, Integer.class, false));
							insertMetric(con, tEngine, metric);
							insertKindred(tEngine, metricId, metric.getId());
							newMetricId=new Integer(metric.getId());
						}
						metricId=newMetricId;
				    }

				    SliceImpl simpl=new SliceImpl(true);
				    
				    simpl.setMetricId(metricId.intValue());
					simpl.setTimeFrom(slice.getFrom()); 
					simpl.setTimeTo(slice.getTo());
					simpl.setMeasurements(slice.getNumber());
					simpl.setTotal(slice.getTotal());
					simpl.setMinValue(slice.getMin());
					simpl.setMaxValue(slice.getMax());
					simpl.setDeviation(slice.getDeviation());

					if (this.identityManager instanceof IdentityGenerator) {
						simpl.setId(((IdentityGenerator) this.identityManager).generate(con, "SLICE"));
					}						
					tEngine.insertSlice(simpl);
					if (this.identityManager instanceof IdentityRetriever) {
						simpl.setId(((IdentityRetriever) this.identityManager).retrieve(con));
					}
					
					Collection measurements = slice.getMeasurements();
					if (measurements!=null && !measurements.isEmpty()) {
						Iterator it=measurements.iterator();
						while (it.hasNext()) {
						    biz.hammurapi.metrics.Metric.Measurement m=(biz.hammurapi.metrics.Metric.Measurement) it.next();
							Measurement dm=new MeasurementImpl();
							dm.setMeasurementTime(m.getTime());
							dm.setMeasurementValue(m.getValue());
							dm.setSliceId(simpl.getId());
							if (this.identityManager instanceof IdentityGenerator) {
								dm.setId(((IdentityGenerator) this.identityManager).generate(con, "MEASUREMENT"));
							}						
							tEngine.insertMeasurement(dm);
						}
					}
					
					con.commit();
					return true;
				} catch (SQLException e) {
					con.rollback();
					throw e;
				} finally {				
					con.setAutoCommit(cac);
					processor.releaseConnection(con);
				}
			} catch (SQLException e) {
				System.err.println("WARN ["+getClass().getName()+"] Exception - deactivated.");
				isActive=false;
				e.printStackTrace();
				return false;
			}
		}
		
		return true;
	}
    
    private long lastMaintenance;
    
    public void onTick(long now) {
        if (isActive && maintenanceInterval>0 && history>0 && lastMaintenance<now-maintenanceInterval) {
            lastMaintenance=now;
            try {
                new MetricsEngine(processor).deleteSliceTimeToLE(now-history);
            } catch (SQLException e) {
				System.err.println("WARN ["+getClass().getName()+"] Exception - deactivated.");
				isActive=false;
                e.printStackTrace();
            }
        }
    }
    
    public Number getRootId() {
    	return rootId;
    }

}
