/*
 * hgee 2
 * Hammurapi Group Enterprise Extensions 
 * Copyright (C) 2006  Hammurapi Group
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * URL: http://www.hammurapi.biz/products/enterpriseextensions/index.html
 * e-Mail: support@hammurapi.biz
 */
package biz.hammurapi.persistence;

import java.util.Timer;

import javax.sql.DataSource;

import biz.hammurapi.cache.AppendableMemoryCache;
import biz.hammurapi.cache.Entry;
import biz.hammurapi.cache.MemoryCache;
import biz.hammurapi.config.ComponentBase;
import biz.hammurapi.config.ConfigurationException;
import biz.hammurapi.config.Context;
import biz.hammurapi.metrics.MeasurementCategory;
import biz.hammurapi.metrics.MeasurementCategoryFactory;
import biz.hammurapi.sql.IdentityManager;
import biz.hammurapi.sql.SQLProcessor;


/**
 * This class manages storage and retrieval of strings to and from relational database.
 * Strings can be of arbitrary size, long strings are cut into chunks. Strings are keyed by integers.
 * Often used strings are cached in memory using soft references.
 * @author Pavel Vlasov
 * @revision $Revision$
 */
public class ChunkingStringStorage extends ComponentBase implements StringStorage {
	private static final MeasurementCategory mc=MeasurementCategoryFactory.getCategory(ChunkingStringStorage.class);

	private Timer timer;
	
	/**
	 * Timer for cleaning up memory entries, can be null.
	 * @param timer
	 */
	public void setTimer(Timer timer) {
		this.timer = timer;
	}
	
	private DataSource datasource;
	
	/**
	 * Data source
	 * @param datasource
	 */
	public void setDatasource(DataSource datasource) {
		this.datasource = datasource;
	}
	
	private AppendableMemoryCache sharedTextCache;
	
	/* (non-Javadoc)
	 * @see biz.hammurapi.persistence.StringStorage#getText(java.lang.Integer)
	 */
	public String getText(Integer key) {
		Entry entry = sharedTextCache.get(key);
		return (String) (entry==null ? null : entry.get());
	}
	
	/* (non-Javadoc)
	 * @see biz.hammurapi.persistence.StringStorage#addText(java.lang.String)
	 */
	public Integer addText(String text) {
		return (Integer) (text==null ? null : sharedTextCache.add(text, 0, 0));
	}
	
	private boolean isOwnTimer;

	/**
	 * Stops memory cache and times.
	 */
	public void stop() {
		sharedTextCache.stop();		
		if (isOwnTimer) {
			timer.cancel();
		}
	}
	
	private Context nameMap;

	private IdentityManager identityManager;

	private int chunkSize;
	
	/**
	 * Set nameMap if database object names are different from default.
	 * @param nameMap
	 */
	void setNameMap(Context nameMap) {
		this.nameMap = nameMap;
	}
	
	/**
	 * Identity manager is used to generate/retrieve identity of newely inserted entries.
	 * @param identityManager
	 */
	public void setIdentityManager(IdentityManager identityManager) {
		this.identityManager = identityManager;
	}
	
	/**
	 * Size of individual text chunk.
	 * @param chunkSize
	 */
	public void setChunkSize(int chunkSize) {
		this.chunkSize = chunkSize;
	}

	/**
	 * Starts timer and cache.
	 */
	public void start() throws ConfigurationException {
		if (timer==null) {
			timer=new Timer();
			isOwnTimer=true;
		}
		
		SQLProcessor sqlProcessor = new SQLProcessor(datasource, nameMap);
		SharedTextAppendableProducer stap = new SharedTextAppendableProducer(
				sqlProcessor,
				identityManager,
				chunkSize);
		
		sharedTextCache=new AppendableMemoryCache(
				stap,
				null,
				mc,
				timer,
				MemoryCache.CLEANUP_INTERVAL);
	}
	
	/**
	 * Alternative way to access text entries.
	 */
	protected Object getChild(String name) {		
		return getText(new Integer(name));
	}
	

}
