/*
 * hgee 2
 * Hammurapi Group Enterprise Extensions 
 * Copyright (C) 2006  Hammurapi Group
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * URL: http://www.hammurapi.biz/products/enterpriseextensions/index.html
 * e-Mail: support@hammurapi.biz
 */
package biz.hammurapi.persistence;

import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.Set;

import biz.hammurapi.cache.AppendableProducer;
import biz.hammurapi.cache.Cache;
import biz.hammurapi.cache.Entry;
import biz.hammurapi.sql.IdentityGenerator;
import biz.hammurapi.sql.IdentityManager;
import biz.hammurapi.sql.IdentityRetriever;
import biz.hammurapi.sql.RowProcessor;
import biz.hammurapi.sql.SQLProcessor;
import biz.hammurapi.sql.SQLRuntimeException;
import biz.hammurapi.sql.Transaction;


/**
 * Manages storing and retrieving shared text from the database.
 * @author Pavel Vlasov
 * @revision $Revision$
 */
class SharedTextAppendableProducer implements AppendableProducer {
	
	private SQLProcessor processor;
	private ChunkingStringStorageEngine engine;
	private IdentityManager identityManager;
	private int chunkSize;

	public SharedTextAppendableProducer(SQLProcessor processor, IdentityManager identityManager, int chunkSize) {
		this.processor =  processor;
		engine = new ChunkingStringStorageEngine(processor);
		this.identityManager = identityManager;
		this.chunkSize = chunkSize;
	}

	public Object add(final Object value) {
		if (value==null) {
			return null;
		}
		
		try {
			final Integer[] key={null};
			final Integer valueHashCode=new Integer(value.hashCode());
			processor.executeTransaction(
					new Transaction() {

						public boolean execute(SQLProcessor processor) throws SQLException {
							final ChunkingStringStorageEngine tEngine=new ChunkingStringStorageEngine(processor);
							tEngine.processSharedTextHashCodeEQ(
									valueHashCode,
									new RowProcessor() {

										public boolean process(ResultSet rs) throws SQLException {
											int id = rs.getInt("ID");
											String dbText=retrieveText(tEngine, id);
											if (value.equals(dbText)) {
												key[0]=new Integer(id);
												return false;
											}
											return true;
										}
										
									});
							
							// No values were found
							if (key[0]==null) {
								String textValue=(String) value;
								int valueLength=textValue.length();
								int numberOfChunks = (int) Math.ceil(textValue.length()/chunkSize);
								String[] chunks=new String[numberOfChunks==0 ? 1 : numberOfChunks];
								
								for (int i=0; i<chunks.length; i++) {
									int startIndex = i*chunkSize;
									int endIndex = startIndex+chunkSize;
									chunks[i]=textValue.substring(startIndex, endIndex>valueLength ? valueLength : endIndex);
								}
								
								SharedTextImpl head=new SharedTextImpl(true);
								head.setChunkPosition(0);
								head.setHashCode(valueHashCode);
								head.setTextChunk(chunks[0]);
								if (identityManager instanceof IdentityGenerator) {
									head.setId(((IdentityGenerator) identityManager).generate(processor.getConnection(), "SHARED_TEXT"));
								}
							
								tEngine.insertSharedText(head);
								
								if (identityManager instanceof IdentityRetriever) {
									head.setId(((IdentityRetriever) identityManager).retrieve(processor.getConnection()));
								}
								key[0]=new Integer(head.getId());
								
								for (int i=1; i<chunks.length; i++) {
									SharedTextImpl sti=new SharedTextImpl(true);
									sti.setTextHead(key[0]);
									sti.setChunkPosition(i);
									sti.setTextChunk(chunks[i]);
									tEngine.insertSharedText(sti);
								}
							}
							return true;
						}
						
					});
			return key[0];
		} catch (SQLException e) {
			throw new SQLRuntimeException("Couldn't store shared text value", e);
		}
	}

	public Entry get(Object key) {
		if (key==null) {
			return null;
		}
		
		try {
			final String value=retrieveText(engine, ((Number) key).intValue());
			return new Entry() {
	
				public long getExpirationTime() {
					return 0;
				}
	
				public long getTime() {
					return 0;
				}
	
				public Object get() {
					return value;
				}
				
			};
		} catch (SQLException e) {
			throw new SQLRuntimeException("Couldn't retrieve shared text value from the database", e);
		}
	}
	
	private String retrieveText(ChunkingStringStorageEngine tEngine, int key) throws SQLException {
		final StringBuffer sb=new StringBuffer();
		tEngine.processTextChunks(
				key, 
				key,
				new RowProcessor() {

					public boolean process(ResultSet rs) throws SQLException {
						sb.append(rs.getString(1));
						return true;
					}
					
				});
		return sb.toString();
	}

	public void addCache(Cache cache) {
		// This functinality is not needed.

	}

	public Set keySet() {
		// This functionality is not needed.
		return null;
	}

}
