package biz.hammurapi.remoting;

import java.io.Serializable;
import java.lang.reflect.InvocationTargetException;

import org.apache.xmlbeans.XmlObject;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

import biz.hammurapi.invocation.Invocation.Arg;
import biz.hammurapi.invocation.Invocation.State;
import biz.hammurapi.xml.dom.AbstractDomObject;
import biz.hammurapi.xml.dom.CompositeDomSerializer;
import biz.hammurapi.xml.dom.DOMUtils;
import biz.hammurapi.xml.dom.DomSerializable;
import biz.hammurapi.xmlbeans.XmlObjectSerializable;

/**
 * Method invocation.
 * @author Pavel
 */
public class Invocation implements DomSerializable, Serializable, XmlObjectSerializable {
	
	/**
	 * Interface to inject client state into the server
	 * object before invocation and reset it after.
	 * Implementations shall store state in a thread local 
	 * variable.
	 * @author Pavel
	 *
	 */
	public interface Stateful {
		
		/**
		 * Sets client state before method invocation.
		 * @param state
		 */
		void setState(Object state);
		
		/**
		 * Resets client state after invocation.
		 * @param state
		 */
		void resetState();
	}

	private Object state;
	private String methodName;
	private String[] parameterTypes;
	private Object[] args;
	private String declaringClass;
	private String id;

	public Invocation(Object state,	String declaringClass, String methodName, String[] parameterTypes,	Object[] args, String id) {
		this.state = state;
		this.declaringClass = declaringClass;
		this.methodName = methodName;
		this.parameterTypes = parameterTypes;
		this.args = args;
		this.id = id;
	}
	
	public Invocation(Object state,	java.lang.reflect.Method method, Object[] args, String id) {
		this.state = state;
		this.declaringClass = method.getDeclaringClass().getName();
		this.methodName = method.getName();
		Class[] mpt = method.getParameterTypes();
		this.parameterTypes = new String[mpt.length];
		for (int i=0; i<mpt.length; ++i) {
			parameterTypes[i] = mpt[i].getName();
		}
		this.args = args;
		this.id = id;
	}
	
	public Object getState() {
		return state;
	}
	
	public String getId() {
		return id;
	}
	
	public String getMethodName() {
		return methodName;
	}
	
	public String getDeclaringClass() {
		return declaringClass;
	}
	
	public String[] getParameterTypes() {
		return parameterTypes;
	}
	
	public Object[] getArguments() {
		return args;
	}
	
	public void toDom(Element holder) {
		holder.setAttribute("method-name", methodName);
		if (declaringClass!=null) {
			holder.setAttribute("declaring-class", declaringClass);
		}
		holder.setAttribute("type", getClass().getName());
		
		if (id!=null) {
			holder.setAttribute("id", id);
		}
		
		if (state!=null) {
			DOMUtils.toDom(state, "state", holder);
		}
		
		if (parameterTypes!=null) {
			for (int i=0; i<parameterTypes.length; ++i) {
				AbstractDomObject.addTextElement(holder, "parameter-type", parameterTypes[i]);				
			}
		}
		
		if (args!=null) {
			for (int i=0; i<args.length; ++i) {
				if (args[i]==null) {
					AbstractDomObject.addElement(holder, "argument").setAttribute("is-null", "true");
				} else {
					DOMUtils.toDom(args[i], "argument", holder);
				}
			}
		}
		
	}
	
	public XmlObject toXmlObject() {
		biz.hammurapi.invocation.InvocationDocument ret = biz.hammurapi.invocation.InvocationDocument.Factory.newInstance();
		biz.hammurapi.invocation.Invocation invocation = ret.addNewInvocation();
		if (declaringClass!=null) {
			invocation.setDeclaringClass(declaringClass);			
		}
		
		invocation.setMethodName(methodName);
		
		if (id!=null) {
			invocation.setId(id);
		}
		
		if (parameterTypes!=null) {
			for (int i=0; i<parameterTypes.length; ++i) {
				invocation.addParameterType(parameterTypes[i]);
			}
		}
		
		if (state!=null) {
			State xmlState = invocation.addNewState();
			if (state instanceof XmlObject) {
				xmlState.set((XmlObject) state);
			} else {
				Node domNode = xmlState.getDomNode();
				CompositeDomSerializer.getThreadInstance().toDomSerializable(state).toDom((Element) domNode);				
			}
		}
		
		if (args!=null) {
			for (int i=0; i<args.length; ++i) {
				Arg arg = invocation.addNewArg();
				if (args[i] instanceof XmlObject) {
					arg.set((XmlObject) args[i]);
				} else {
					Node domNode = arg.getDomNode();
					CompositeDomSerializer.getThreadInstance().toDomSerializable(args[i]).toDom((Element) domNode);				
				}
			}			
		}
			
		return ret;
	}
	
	/**
	 * Finds method to invoke.
	 * @param invocation
	 * @param classLoader Classloader, can be null.
	 * @return
	 * @throws ClassNotFoundException 
	 */
	public static java.lang.reflect.Method findMethod(biz.hammurapi.invocation.Invocation invocation, ClassLoader classLoader) throws ClassNotFoundException, NoSuchMethodException {
		Class declaringClass = classLoader==null ? Class.forName(invocation.getDeclaringClass()) : classLoader.loadClass(invocation.getDeclaringClass());
		String[] pta = invocation.getParameterTypeArray();
		Class[] parameterTypes = new Class[pta.length];
		for (int i=0; i<pta.length; ++i) {
			if ("int".equals(pta[i])) {
				parameterTypes[i] = int.class;
			} else if ("short".equals(pta[i])) {
				parameterTypes[i] = short.class;
			} else if ("long".equals(pta[i])) {
				parameterTypes[i] = long.class;
			} else if ("boolean".equals(pta[i])) {
				parameterTypes[i] = boolean.class;
			} else if ("byte".equals(pta[i])) {
				parameterTypes[i] = byte.class;
			} else if ("double".equals(pta[i])) {
				parameterTypes[i] = double.class;
			} else if ("float".equals(pta[i])) {
				parameterTypes[i] = float.class;
			} else if ("char".equals(pta[i])) {
				parameterTypes[i] = char.class;
			} else {
				parameterTypes[i] = classLoader==null ? Class.forName(pta[i]) : classLoader.loadClass(pta[i]);
			}
		}
		return declaringClass.getMethod(invocation.getMethodName(), parameterTypes);
	}
	
	/**
	 * Finds method to invoke.
	 * @param invocation
	 * @param classLoader Classloader, can be null.
	 * @return
	 * @throws ClassNotFoundException 
	 */
	public java.lang.reflect.Method findMethod(ClassLoader classLoader) throws ClassNotFoundException, NoSuchMethodException {
		Class declaringClass = classLoader==null ? Class.forName(getDeclaringClass()) : classLoader.loadClass(getDeclaringClass());
		Class[] parameterTypes = new Class[this.parameterTypes.length];
		for (int i=0; i<this.parameterTypes.length; ++i) {
			if ("int".equals(this.parameterTypes[i])) {
				parameterTypes[i] = int.class;
			} else if ("short".equals(this.parameterTypes[i])) {
				parameterTypes[i] = short.class;
			} else if ("long".equals(this.parameterTypes[i])) {
				parameterTypes[i] = long.class;
			} else if ("boolean".equals(this.parameterTypes[i])) {
				parameterTypes[i] = boolean.class;
			} else if ("byte".equals(this.parameterTypes[i])) {
				parameterTypes[i] = byte.class;
			} else if ("double".equals(this.parameterTypes[i])) {
				parameterTypes[i] = double.class;
			} else if ("float".equals(this.parameterTypes[i])) {
				parameterTypes[i] = float.class;
			} else if ("char".equals(this.parameterTypes[i])) {
				parameterTypes[i] = char.class;
			} else {
				parameterTypes[i] = classLoader==null ? Class.forName(this.parameterTypes[i]) : classLoader.loadClass(this.parameterTypes[i]);
			}
		}
		return declaringClass.getMethod(getMethodName(), parameterTypes);
	}	
	
	/**
	 * Method for simple invocation.
	 * @param instance
	 * @return
	 * @throws IllegalAccessException
	 * @throws InvocationTargetException
	 * @throws ClassNotFoundException
	 * @throws NoSuchMethodException
	 */
	public Object invoke(Object instance) throws IllegalAccessException, InvocationTargetException, ClassNotFoundException, NoSuchMethodException {
		if (instance instanceof Stateful) {
			((Stateful) instance).setState(state);
		}
		try {
			return findMethod(instance==null ? getClass().getClassLoader() : instance.getClass().getClassLoader()).invoke(instance, args);
		} finally {
			if (instance instanceof Stateful) {
				((Stateful) instance).resetState();
			}			
		}
	}
	
}
