/*
 * hgcommons 7
 * Hammurapi Group Common Library 
 * Copyright (C) 2003  Hammurapi Group
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * URL: http://www.hammurapi.biz/hammurapi-biz/ef/xmenu/hammurapi-group/products/products/hgcommons/index.html
 * e-Mail: support@hammurapi.biz
 */
package biz.hammurapi.sql;

import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.Map;

/**
 * This projector constructs objects using database field values
 * @author Pavel Vlasov
 * @version $Revision: 1.2 $
 */
public class ConstructorProjector extends BaseReflectionProjector implements Projector {
	public static class ColumnName {
		private String name;

		public ColumnName(String name) {
			this.name=name;
		}
	}
	
	private Object[] args;
	private Constructor constructor;

	/**
	 * Use this constructor if target constructor shall take not only values from database fields as
	 * parameter, but other objects as well.
	 * @param constructor Constructor to instantiate object.
	 * @param args Constructor arguments. Arguments of type {@link ColumnName} will be replaced
	 * with values from corresponding fields from the database.
	 * @param typeMap See {@link java.sql.ResultSet#getObject(java.lang.String, java.util.Map)}. Can be null.
	 */
	public ConstructorProjector(Constructor constructor, Object[] args, Map typeMap) {
		super(typeMap);
		if (constructor.getParameterTypes().length!=args.length) {
			throw new IllegalArgumentException("argNames length shall be equal to the number of constructor parameters");
		}
		this.constructor=constructor;
		this.args=args;
	}
	
	/**
	 * Use this constructor if only database fields values are used as target constructor parameters.
	 * @param constructor Constructor to instantiate object
	 * @param columnNames Names of columns that shall be passed as parameters to constructor.
	 * @param typeMap See {@link java.sql.ResultSet#getObject(java.lang.String, java.util.Map)}. Can be null.
	 */
	public ConstructorProjector(Constructor constructor, String[] columnNames, Map typeMap) {
		this(constructor, columnNamesToArgs(columnNames), typeMap);
	}
	
	/**
	 * Use constructor for positioned projection. Column values will be passed to as constructor
	 * arguments according to their position.
	 * @param constructor
	 * @param typeMap
	 */
	public ConstructorProjector(Constructor constructor, Map typeMap) {
		super(typeMap);
		this.constructor=constructor;
	}

	/**
	 * 
	 * @param fieldNames
	 * @return
	 */
	private static Object[] columnNamesToArgs(String[] fieldNames) {
		Object[] ret=new Object[fieldNames.length];
		for (int i=0; i<fieldNames.length; i++) {
			ret[i]=new ColumnName(fieldNames[i]);
		}
		return ret;
	}

	public Object project(ResultSet rs) throws SQLException {
		Object[] params=new Object[constructor.getParameterTypes().length];
		for (int i=0; i<params.length; i++) {
			if (args==null) {
				params[i]=getColumn(rs, i+1);
			} else if (args[i] instanceof ColumnName) {
				params[i]=getColumn(rs, ((ColumnName) args[i]).name);	
			} else {
				params[i]=args[i];
			}			
		}
		
		try {
			return constructor.newInstance(params);
		} catch (InvocationTargetException e) {
			throw new SQLExceptionEx(e);
		} catch (InstantiationException e) {
			throw new SQLExceptionEx(e);
		} catch (IllegalAccessException e) {
			throw new SQLExceptionEx(e);
		}
	}

}
