/*
  * hgcommons 7
 * Hammurapi Group Common Library 
 * Copyright (C) 2003  Hammurapi Group
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * URL: http://www.hammurapi.biz/hammurapi-biz/ef/xmenu/hammurapi-group/products/products/hgcommons/index.html
 * e-Mail: support@hammurapi.biz
 */
package biz.hammurapi.sql;

import java.io.Serializable;
import java.lang.reflect.Constructor;
import java.sql.PreparedStatement;
import java.sql.SQLException;
import java.sql.Types;
import java.util.AbstractList;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Properties;
import java.util.Set;
import java.util.Map.Entry;

import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import biz.hammurapi.config.ConfigurationException;
import biz.hammurapi.config.Context;
import biz.hammurapi.config.ContextConfigurable;
import biz.hammurapi.config.DomConfigFactory;
import biz.hammurapi.config.DomConfigurable;
import biz.hammurapi.convert.CompositeConverter;
import biz.hammurapi.sql.columns.Column;
import biz.hammurapi.sql.columns.ColumnChangeListener;
import biz.hammurapi.util.Attributable;
import biz.hammurapi.util.ClassResourceLoader;
import biz.hammurapi.util.Observable;
import biz.hammurapi.util.Observer;
import biz.hammurapi.util.Versioned;
import biz.hammurapi.xml.dom.DOMUtils;
import biz.hammurapi.xml.dom.AbstractDomObject;
import biz.hammurapi.xml.dom.CompositeDomSerializer;
import biz.hammurapi.xml.dom.DomSerializable;


/**
 * SQLC-generated interface implementations implement this method to achieve
 * differential update functionality - inserting and updating only modified fields.
 * @author Pavel Vlasov
 * @version $Revision: 1.11 $
 */
public class DatabaseObject 
	implements 
		DomSerializable, 
		ColumnChangeListener, 
		Cloneable, 
		ContextConfigurable, 
		Context,
		DomConfigurable,
		Attributable,
		Versioned,
		Observable, 
		IDatabaseObject,
		Serializable {
    
    protected Collection columns=new ArrayList();
    private Map columnMap=new HashMap();
    private boolean force;
	private boolean isDeleted;
	
	protected Column getColumn(String name) {
		return (Column) columnMap.get(name);
	}
	
	/**
	 * Default constructor
	 */
	public DatabaseObject() {
		// Default constructor
	}
	
	/**
	 * 
	 * @param force Forces columns to be marked as 
	 * modified if setter method is invoked even if value being set equals to existing column
	 * value. Useful during inserts with non-nullable columns which map to primitive 
	 * types and as such have default values.
	 */
	public DatabaseObject(boolean force) {
		this.force=force;
	}
	
    protected void addColumn(Column column) {
    	column.setForce(force);
        columns.add(column);
        columnMap.put(column.getName(), column);
        column.setListener(this);
    }
       
	/* (non-Javadoc)
	 * @see biz.hammurapi.sql.IDatabaseObject#update(biz.hammurapi.sql.SQLProcessor, java.lang.String)
	 */
	public int update(SQLProcessor processor, String tableName) throws SQLException {
	    StringBuffer sb=new StringBuffer("UPDATE ");
	    sb.append(tableName);
	    sb.append(" SET ");
	    boolean hasColumns=false;
	    Iterator it=columns.iterator();
	    while (it.hasNext()) {
	        Column column = (Column) it.next();
			String colName=(column).listName();
	        if (colName!=null) {
	            if (hasColumns) {
	                sb.append(", ");
	            }
	            sb.append(colName);
	            sb.append("=?");
	            hasColumns=true;
	            
	        }
	    }
	    
	    if (!hasColumns) {
	        return 0;
	    }
	    
	    sb.append(" WHERE ");
	    
	    boolean hasPkColumns=false;
	    it=columns.iterator();
	    while (it.hasNext()) {
	        Column column = (Column) it.next();
	        if (column.isPrimaryKey()) {
				String colName=(column).getName();
	            if (hasPkColumns) {
	                sb.append(" AND ");
	            }
	            sb.append(colName);
	            sb.append("=?");
	            hasPkColumns=true;
	        }
	    }
	    
	    int ret=processor.processUpdate(sb.toString(), new Parameterizer() {
            public void parameterize(PreparedStatement ps) throws SQLException {
            	int idx=1;
        	    Iterator it=columns.iterator();
        	    while (it.hasNext()) {
        	        Column column = (Column) it.next();
       	        	idx=column.parameterize(ps, idx, false);
        	    }
        	    
        	    it=columns.iterator();
        	    while (it.hasNext()) {
        	        Column column = (Column) it.next();
        	        if (column.isPrimaryKey()) {
        	        	column.parameterizeOriginal(ps, idx++);
        	        }
        	    }
            }	        
	    });

	    it=columns.iterator();
	    while (it.hasNext()) {
	        ((Column) it.next()).clearModified();
	    }	  
	    
	    originalVersion=objectVersion;	    
	    isModified=false;
	    
	    storeRelationships(processor);
	    
		it=relationships.iterator();
		while (it.hasNext()) {
			RelationshipEntry re=(RelationshipEntry) it.next();
			if (re.name!=null) {				
				Iterator iit=re.items.iterator();
				while (iit.hasNext()) {
					IDatabaseObject subItem = (IDatabaseObject) iit.next();
					if (subItem.isModified()) {
						re.items.update(processor, subItem);
						subItem.update(processor, null);
					}
				}
			}
		}
	    return ret;
	}
	
	/* (non-Javadoc)
	 * @see biz.hammurapi.sql.IDatabaseObject#delete(biz.hammurapi.sql.SQLProcessor, java.lang.String)
	 */
	public int delete(SQLProcessor processor, String tableName) throws SQLException {
		storeRelationships(processor);
		
	    StringBuffer sb=new StringBuffer("DELETE FROM ");
	    sb.append(tableName);
	    sb.append(" WHERE ");
	    
	    boolean hasPkColumns=false;
	    Iterator it=columns.iterator();
	    while (it.hasNext()) {
	        Column column = (Column) it.next();
	        if (column.isPrimaryKey()) {
				String colName=(column).getName();
	            if (hasPkColumns) {
	                sb.append(" AND ");
	            }
	            sb.append(colName);
	            sb.append("=?");
	            hasPkColumns=true;
	        }
	    }
	    
	    int ret=processor.processUpdate(sb.toString(), new Parameterizer() {
            public void parameterize(PreparedStatement ps) throws SQLException {
            	int idx=1;
        	    Iterator it=columns.iterator();
        	    while (it.hasNext()) {
        	        Column column = (Column) it.next();
        	        if (!column.isPrimaryKey()) {
        	        	idx=column.parameterize(ps, idx, false);
        	        }
        	    }
        	    
        	    it=columns.iterator();
        	    while (it.hasNext()) {
        	        Column column = (Column) it.next();
        	        if (column.isPrimaryKey()) {
        	        	idx=column.parameterize(ps, idx, true);
        	        }
        	    }
            }	        
	    });

	    it=columns.iterator();
	    while (it.hasNext()) {
	        ((Column) it.next()).clearModified();
	    }	    
	    
	    originalVersion=objectVersion;
	    isDeleted=true;
	    return ret;
	}
	
	/* (non-Javadoc)
	 * @see biz.hammurapi.sql.IDatabaseObject#insert(biz.hammurapi.sql.SQLProcessor, java.lang.String)
	 */
	public int insert(SQLProcessor processor, String tableName) throws SQLException {
	    StringBuffer sb=new StringBuffer("INSERT INTO ");
	    sb.append(tableName);
	    sb.append(" (");
	    int columnsCounter=0;
	    Iterator it=columns.iterator();
	    while (it.hasNext()) {
	        String colName=((Column) it.next()).listName();
	        if (colName!=null) {
	            if (columnsCounter>0) {
	                sb.append(", ");
	            }
	            sb.append(colName);
	            columnsCounter++;
	            
	        }
	    }
	    
	    if (columnsCounter==0) {
	        return 0;
	    }
	    
	    sb.append(") VALUES (");
	    for (int i=0; i<columnsCounter; i++) {
	        if (i>0) {
	            sb.append(", ");
	        }
	        sb.append("?");
	    }
	    sb.append(")");
	    
	    int ret=processor.processUpdate(sb.toString(), new Parameterizer() {
            public void parameterize(PreparedStatement ps) throws SQLException {
        	    Iterator it=columns.iterator();
        	    for (int i=1; it.hasNext(); i=((Column) it.next()).parameterize(ps, i, false));
            }	        
	    });

	    it=columns.iterator();
	    while (it.hasNext()) {
	        ((Column) it.next()).clearModified();
	    }	    
	    
	    originalVersion=objectVersion;
	    isModified=false;
	    
	    storeRelationships(processor);
	    return ret;
	}
	
	/* (non-Javadoc)
	 * @see biz.hammurapi.sql.IDatabaseObject#fromDom(org.w3c.dom.Element)
	 */
	public void fromDom(Element holder) throws ConfigurationException {
		fromDom(holder, getNameMap(getClass()));
	}
	
	private void loadAttributes(Element holder) throws ConfigurationException {
		DomConfigFactory dcf=new DomConfigFactory();
		attributes.clear();
		try {
			Node attributesNode = DOMUtils.selectSingleNode(holder, "object-attributes");
			if (attributesNode!=null) {
				attributes.putAll((Map) dcf.create(attributesNode));
			}
		} catch (Exception e) {
			throw new ConfigurationException("Cannot load database object attributes: "+e, e);
		}
				
	}
	
	/* (non-Javadoc)
	 * @see biz.hammurapi.sql.IDatabaseObject#fromDom(org.w3c.dom.Element, java.util.Properties)
	 */
	public void fromDom(Element holder, Properties nameMap) throws ConfigurationException {
		Iterator it=columns.iterator();
		while (it.hasNext()) {
			Column column=(Column) it.next();
			String nodeName = nameMap.getProperty(column.getName(), column.getName()).trim();			

			if (nodeName.length()==0) {
				// Zero mapping
				continue;
			} else if (nodeName.startsWith("@")) {
				// Attribute mapping
				if (holder.hasAttribute(nodeName.substring(1))) {
					column.load(holder.getAttribute(nodeName.substring(1)));
				}
			} else if (nodeName.equals(".")) {
				// Text content mapping
				column.load(AbstractDomObject.getElementText(holder));
			} else if (nodeName.startsWith("!")) {
				// Simple node mapping
				try {
					Node cNode=DOMUtils.selectSingleNode(holder, nodeName.substring(1));
					if (cNode!=null) {
						column.load(AbstractDomObject.getElementText(cNode));
					}
				} catch (Exception e) {
					throw new ConfigurationException("Cannot load column "+column.getName());
				}				
			} else {
				// Regular node mapping
				try {
					Node cNode=DOMUtils.selectSingleNode(holder, nodeName);
					if (cNode!=null) {
						column.configure(cNode, null);
					}
				} catch (Exception e) {
					throw new ConfigurationException("Cannot load column "+column.getName());
				}
			}
		}				
		
		it=relationships.iterator();
		while (it.hasNext()) {
			RelationshipEntry re=(RelationshipEntry) it.next();
			if (re.name!=null) {
				String domName=re.domName==null ? re.name : re.domName;
				try {
					Constructor constructor=re.items.relationship.getItemType().getConstructor(new Class[] {Element.class, boolean.class});
					Element rel = ".".equals(domName) ? holder : (Element) DOMUtils.selectSingleNode(holder, domName);
					String itemName = re.itemName==null ? "element" : re.itemName;
					NodeList nl=DOMUtils.selectNodeList(rel, itemName);
					for (int i=0, l=nl.getLength(); i<l; ++i) {
						Element itemElement = (Element) nl.item(i);
						re.items.add(constructor.newInstance(new Object[] {itemElement, force ? Boolean.TRUE : Boolean.FALSE}));
					}
				} catch (Exception e) {
					throw new ConfigurationException("Cannot load relationship "+re.name, e);
				}
			}
		}		
		loadAttributes(holder);
	}
		
	public void toDom(Element holder) {
		toDom(holder, getNameMap(getClass()), false);
	}
	
	/* (non-Javadoc)
	 * @see biz.hammurapi.sql.IDatabaseObject#toDom(org.w3c.dom.Element, java.util.Properties, boolean)
	 */
	public void toDom(Element holder, Properties nameMap, boolean originals) {
		if (nameMap==null) {
			nameMap=new Properties();
		}
		
		String cna = nameMap.getProperty("@type", "type").trim();
		if (!"".equals(cna)) {
			holder.setAttribute(cna, getClass().getName());
		}
		
		Iterator it=columns.iterator();
		while (it.hasNext()) {
			Column column = (Column) it.next();
			column.toDom(holder, nameMap.getProperty(column.getName(), column.getName()).trim(), originals);
		}
		
		it=relationships.iterator();
		while (it.hasNext()) {
			RelationshipEntry re=(RelationshipEntry) it.next();
			if (re.name!=null) {
				String domName=re.domName==null ? re.name : re.domName;
				Element rel = ".".equals(domName) ? holder : holder.getOwnerDocument().createElement(domName);
				if (rel!=holder) {
					holder.appendChild(rel);
				}
				
				Iterator iit=re.items.iterator();
				while (iit.hasNext()) {
					String itemName = re.itemName==null ? "element" : re.itemName;
					Element ie=holder.getOwnerDocument().createElement(itemName);
					rel.appendChild(ie);
					((DatabaseObject) iit.next()).toDom(ie);
				}
			}
		}
		
		if (!attributes.isEmpty()) {
			CompositeDomSerializer
			.getThreadInstance()
			.toDomSerializable(attributes)
			.toDom(AbstractDomObject.addElement(holder, "object-attributes"));
		}		
	}
	
	public String toString() {
		StringBuffer ret=new StringBuffer(getClass().getName());
		ret.append("[");
		Iterator it=columns.iterator();
		while (it.hasNext()) {
			ret.append(it.next());
			if (it.hasNext()) {
				ret.append(", ");
			}
		}		
		
		ret.append("]");		
		return ret.toString();
	}

	/**
	 * Sets modified flag to true and increments version number.
	 * Also broadcasts the change to observers. Changed column is
	 * passed as second argument of update() method.
	 * Override this method in subclasses to react on 
	 * change events, but don't forget to invoke 
	 * super.onChange().
	 * @param column Changed column
	 */
	public void onChange(Column column) {
		isModified=true;
		
		++objectVersion;
		
		if (column.isPrimaryKey()) {
			isDeleted=false;
		}
		
		Iterator it=relationships.iterator();
		while (it.hasNext()) {
			Relationship relationship=((RelationshipEntry) it.next()).items.relationship;
			if (relationship instanceof ColumnChangeListener) {
				((ColumnChangeListener) relationship).onChange(column);
			}
		}
		
		synchronized (observers) {
			it=observers.iterator();
			while (it.hasNext()) {
				((Observer) it.next()).update(this, column);
			}
		}		

	}
	
	/* (non-Javadoc)
	 * @see biz.hammurapi.sql.IDatabaseObject#setOriginal()
	 */
	public void setOriginal() {
		objectVersion=originalVersion;
		Iterator it=columns.iterator();
		while (it.hasNext()) {
			((Column) it.next()).setOriginal();
		}		
	}
	
	private boolean isModified=false;
	
	/* (non-Javadoc)
	 * @see biz.hammurapi.sql.IDatabaseObject#isModified()
	 */
	public boolean isModified() {
		return isModified;
	}
	
	/* (non-Javadoc)
	 * @see biz.hammurapi.sql.IDatabaseObject#isDeleted()
	 */
	public boolean isDeleted() {
		return isDeleted;
	}
	
	/**
	 * Two objects are considered equal and all their fields are equal.
	 * @param otherBean Other object
	 * @return true if object classes are equal and all member column values are
	 * equal.
	 */
    public boolean equals(Object otherBean) {
        if (otherBean==null) {
            return false;
        } else if (getClass().equals(otherBean.getClass())) {
        	Collection myColumns=new ArrayList(columns);
        	Collection otherColumns=new ArrayList(((DatabaseObject) otherBean).columns);
        	Iterator mcit=myColumns.iterator();
        	Z:
        	while (mcit.hasNext()) {
        		Column mc=(Column) mcit.next();
        		Iterator ocit=otherColumns.iterator();
				while (ocit.hasNext()) {
					Column oc=(Column) ocit.next();
					if (mc.getName().equals(oc.getName())) {
						if (mc.equals(oc)) {
							ocit.remove();
							mcit.remove();
							continue Z;
						}
						
						return false;
					}
				}
        	}
        	
			return myColumns.isEmpty() && otherColumns.isEmpty();
        } else {
            return false;
        }
    }
    
	public int hashCode() {
		int ret=0;
		Iterator cit=columns.iterator();
		while (cit.hasNext()) {
			ret^=cit.next().hashCode();
		}
		return ret;
	}

	/**
	 * Clones object, clears columns collection, clears isDeleted and isModified flags. 
	 * Subclasses shall add cloned columns.
	 */
	public Object clone() throws CloneNotSupportedException {
		DatabaseObject ret = (DatabaseObject) super.clone();
		ret.columns.clear();
		ret.isDeleted=false;
		ret.isModified=false;
		return ret;
	}
	
	/* (non-Javadoc)
	 * @see biz.hammurapi.sql.IDatabaseObject#clear()
	 */
	public void clear() {
		Iterator it = columns.iterator();
		while (it.hasNext()) {
			((Column) it.next()).clear();
		}
		isModified=false;
		isDeleted=false;
	}

	public void configure(Context context, CompositeConverter converter) throws ConfigurationException {
		Iterator it = columns.iterator();
		while (it.hasNext()) {
			((Column) it.next()).configure(context, converter);
		}
	}

    public Object get(String name) {
        Column col=(Column) columnMap.get(name);
        return col==null ? null : col.getObjectValue(false);
    }

	public void configure(Node configNode, Context context) throws ConfigurationException {
		fromDom((Element) configNode);
	}
	
	private static Map nnMap=new HashMap();
	
	private static Properties getNameMap(Class clazz) {
		synchronized (nnMap) {
			Properties ret=(Properties) nnMap.get(clazz.getName());
			if (ret==null) {
				ret=new ClassResourceLoader(clazz).getProperties(null, "namemap");
				nnMap.put(clazz.getName(), ret);
			}
			return ret;
		}
	}
	
	private class RelationshipEntry {
		String name;
		String itemName;
		String domName;
		RelationshipListImpl items;
	}
	
	private class RelationshipListImpl extends AbstractList implements RelationshipList {
		private ArrayList master=new ArrayList();
		private Relationship relationship;
		private boolean alreadyLoaded;

		private RelationshipListImpl(Relationship relationship) {
			this.relationship=relationship;
			relationship.setMaster(master);
		}
		
		private void update(SQLProcessor processor, IDatabaseObject subItem) throws SQLException {
			relationship.update(processor, subItem);			
		}
		
		private void load(SQLProcessor processor) throws SQLException {
			if (!alreadyLoaded || relationship.isModified()) {
				master.clear();
				relationship.load(processor, master);
				alreadyLoaded=true;
			}
		}
		
		private void store(SQLProcessor processor) throws SQLException {
			relationship.store(processor);
		}
		
		public Object get(int index) {
			return master.get(index);
		}

		public int size() {
			return master.size();
		}
		
		public boolean add(Object o) {
			relationship.add((DatabaseObject) o);
			return master.add(o);
		}
		
		public boolean remove(Object o) {
			relationship.remove((IDatabaseObject) o);
			return master.remove(o);
		}

		public boolean isLazy() {
			return relationship.isLazy();
		}

		public boolean isModified() {
			return relationship.isModified();
		}

		public IDatabaseObject add() {
			try {
				IDatabaseObject item = (IDatabaseObject) relationship.getItemType().newInstance();
				add(item);
				return item;
			} catch (InstantiationException e) {
				throw new RuntimeException("Cannot instantiate "+relationship.getItemType());
			} catch (IllegalAccessException e) {
				throw new RuntimeException("Cannot instantiate "+relationship.getItemType());
			}
		}		
	}
		
	private Collection relationships=new ArrayList();
	private Map relationshipMap=new HashMap();
	
	protected RelationshipList getRelationship(String name) {
		RelationshipEntry entry=(RelationshipEntry) relationshipMap.get(name);
		if (entry.items.isLazy() || entry.items.isModified()) {
			try {
				entry.items.load(((Lazy) this).getProcessor());
			} catch (SQLException e) {
				throw new SQLRuntimeException(e);
			}
		}
		return entry.items;
	}
	
	/**
	 * @param name Name for composite relationships to render in XML, null for shared relationships.
	 * @param itemName name of item element in XML.
	 * @param itemClass item class. This class shall have constructor from Element,boolean in order to load from 
	 * XML docs.
	 * @param relationship
	 */
	protected RelationshipList addRelationship(String name, String itemName, Relationship relationship) {
		if (relationship.isLazy() && !(this instanceof Lazy)) {
			throw new IllegalArgumentException("Cannot add lazy relationship to class which doesn't implement Lazy interface");
		}
		
		if (!DatabaseObject.class.isAssignableFrom(relationship.getItemType())) {
			throw new IllegalArgumentException("Relationship can be established only between DatabaseObject subclasses");
		}
		RelationshipEntry entry=new RelationshipEntry();
		relationships.add(entry);
		entry.name=name;
		entry.itemName=itemName;
		entry.items=new RelationshipListImpl(relationship);
        Properties nameMap=getNameMap(getClass());
        String cName = name==null ? null : nameMap.getProperty(name);        
		entry.domName = cName==null ? null : cName.trim();
		relationshipMap.put(name, entry);
		return entry.items;
	}
	
	/**
	 * Use this method to eagerly load relationships in constructors.
	 * @param processor
	 * @throws SQLException
	 */
	protected void loadRelationships(SQLProcessor processor) throws SQLException {
		Iterator it=relationships.iterator();
		while (it.hasNext()) {
			RelationshipEntry relationshipEntry = (RelationshipEntry) it.next();
			if (!relationshipEntry.items.isLazy()) {
				(relationshipEntry).items.load(processor);
			}
		}
	}
	
	/**
	 * Use this method to eagerly load relationships in constructors.
	 * @param processor
	 * @throws SQLException
	 */
	private void storeRelationships(SQLProcessor processor) throws SQLException {
		Iterator it=relationships.iterator();
		while (it.hasNext()) {
			((RelationshipEntry) it.next()).items.store(processor);
		}
	}	
	
	
	/* (non-Javadoc)
	 * @see biz.hammurapi.sql.IDatabaseObject#copy(biz.hammurapi.sql.DatabaseObject)
	 */
	public void copy(DatabaseObject source) {
		Iterator it=columns.iterator();
		while (it.hasNext()) {
			Column targetColumn=(Column) it.next();
			Column sourceColumn=(Column) source.columnMap.get(targetColumn.getName());
			if (targetColumn.getClass().isInstance(sourceColumn)) { // Copy values only from compatible columns
				targetColumn.set(sourceColumn);
			}
		}
	}
	
	private Map attributes=new HashMap();
	
	public void setAttribute(Object key, Object value) {
		attributes.put(key, value);		
	}
	
	public Object getAttribute(Object key) {
		return attributes.get(key);
	}
	
	public Object removeAttribute(Object key) {
		return attributes.remove(key);
	}
	
	/* (non-Javadoc)
	 * @see biz.hammurapi.sql.IDatabaseObject#setColumnAttribute(java.lang.String, java.lang.Object, java.lang.Object)
	 */
	public void setColumnAttribute(String columnName, Object key, Object value) {
		Column column=(Column) columnMap.get(columnName);
		if (column==null) {
			throw new IllegalArgumentException("Column not found: "+columnName);
		}
		column.setAttribute(key, value);		
	}
	
	/* (non-Javadoc)
	 * @see biz.hammurapi.sql.IDatabaseObject#getColumnAttribute(java.lang.String, java.lang.Object)
	 */
	public Object getColumnAttribute(String columnName, Object key) {
		Column column=(Column) columnMap.get(columnName);
		if (column==null) {
			throw new IllegalArgumentException("Column not found: "+columnName);
		}
		return column.getAttribute(key);
	}
	
	/* (non-Javadoc)
	 * @see biz.hammurapi.sql.IDatabaseObject#removeColumnAttribute(java.lang.String, java.lang.Object)
	 */
	public Object removeColumnAttribute(String columnName, Object key) {
		Column column=(Column) columnMap.get(columnName);
		if (column==null) {
			throw new IllegalArgumentException("Column not found: "+columnName);
		}
		return column.removeAttribute(key);
	}
	
	private static Map sqlTypes=new HashMap();
	
	/**
	 * Allows to override generated column types with &lt;class name&gt;.sqltypes resource.
	 * Subclasses shall use this method when dealing with Object columns.
	 * @param columnName
	 * @param generatedType
	 * @return
	 */
	protected int getSqlType(String columnName, int generatedType) {
		Map typeMap;
		synchronized (sqlTypes) {
			String className = getClass().getName();
			if (sqlTypes.containsKey(className)) {
				typeMap=(Map) sqlTypes.get(className);
			} else {
				Properties literalMap=new ClassResourceLoader(getClass()).getProperties(null, "sqltypes");
				typeMap=new HashMap();
				sqlTypes.put(className, typeMap);
				Iterator it=literalMap.entrySet().iterator();				
				while (it.hasNext()) {
					Entry entry=(Entry) it.next();
					try {
						Integer type = (Integer) Types.class.getDeclaredField((String) entry.getValue()).get(null) ;
						typeMap.put(entry.getKey(), type);
					} catch (IllegalArgumentException e) {
						System.err.println("Invalid SQL Type "+entry.getValue()+", ignored. Cause: "+e);
						e.printStackTrace();
					} catch (SecurityException e) {
						System.err.println("Invalid SQL Type "+entry.getValue()+", ignored. Cause: "+e);
						e.printStackTrace();
					} catch (IllegalAccessException e) {
						System.err.println("Invalid SQL Type "+entry.getValue()+", ignored. Cause: "+e);
						e.printStackTrace();
					} catch (NoSuchFieldException e) {
						System.err.println("Invalid SQL Type "+entry.getValue()+", ignored. Cause: "+e);
						e.printStackTrace();
					}
				}
			}
		}
		
		Integer st=(Integer) typeMap.get(columnName);
		return st==null ? generatedType : st.intValue();
	}
	
	/**
	 * Subclasses can choose to read object version from the database.
	 */
	protected int objectVersion;
	protected int originalVersion;

	public int getObjectVersion() {
		return objectVersion;
	}
	
	private Set observers = new HashSet();

	public void addObserver(Observer observer) {
		synchronized (observers) {
			observers.add(observer);
		}		
	}

	public void removeObserver(Observer observer) {
		synchronized (observers) {
			observers.remove(observer);
		}		
	}
	
}
