/*
 * hgcommons 7
 * Hammurapi Group Common Library 
 * Copyright (C) 2003  Hammurapi Group
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * URL: http://www.hammurapi.biz/hammurapi-biz/ef/xmenu/hammurapi-group/products/products/hgcommons/index.html
 * e-Mail: support@hammurapi.biz
 */
package biz.hammurapi.sql;

import java.lang.reflect.InvocationTargetException;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;

import biz.hammurapi.sql.columns.Column;
import biz.hammurapi.sql.columns.ColumnChangeListener;

/**
 * Base class for composite relationships where subitems exist only as part of the whole and removal of item from 
 * relationship is equivalent to deletion of the item.
 * @author Pavel Vlasov
 * @revision $Revision$
 */
public class SimpleRelationship implements Relationship, ColumnChangeListener {
	protected Collection inserted=new ArrayList();
	protected Collection deleted=new ArrayList();
	private DatabaseObject owner;
	protected String tableName;
	private boolean isLazy;
	private String[][] keyInfo;
	private Class itemClass;
	private boolean isModified;
	private Collection master;

	/**
	 * 
	 * @param owner Relationship owner.
	 * @param tableName Table where child item are stored.
	 * @param keyInfo Mapping of owner's primary key columns to child foreign key columns.
	 * @param isLazy indicates whether relationship is lazy.
	 */
	public SimpleRelationship(DatabaseObject owner, Class itemClass, String tableName, String[][] keyInfo, boolean isLazy) {
		this.owner=owner;
		this.tableName=tableName;
		this.isLazy=isLazy;
		this.keyInfo=keyInfo;
		this.itemClass=itemClass;
	}
	
	public boolean isLazy() {
		return isLazy;
	}
	
	public void store(SQLProcessor processor) throws SQLException {
		Iterator it=inserted.iterator();
		while (it.hasNext()) {
			IDatabaseObject item=(IDatabaseObject) it.next();
			item.insert(processor, tableName);
			it.remove();
		}
		
		it=deleted.iterator();
		while (it.hasNext()) {
			((IDatabaseObject) it.next()).delete(processor, tableName);
			it.remove();
		}		
	}
	
	public void add(DatabaseObject item) {
		if (!deleted.remove(item)) {
			link(item);
			inserted.add(item);
		} 
	}

	/**
	 * Sets foreign key fields.
	 * @param item
	 */
	private void link(DatabaseObject item) {
		for (int i=0; i<keyInfo.length; i++) {
			Column source=owner.getColumn(keyInfo[i][0]);
			Column target=item.getColumn(keyInfo[i][1]);
			target.set(source);
		}
	}

	public void remove(IDatabaseObject item) {
		if (!inserted.remove(item)) {
			deleted.add(item);
		} 
	}

	public void load(SQLProcessor processor, Collection receiver) throws SQLException {
		StringBuffer sql=new StringBuffer("SELECT * FROM ");
		sql.append(tableName);
		sql.append(" WHERE ");
		for (int i=0; i<keyInfo.length; i++) {
			if (i>0) {
				sql.append("AND ");
			}
			
			sql.append(keyInfo[i][1]);
			sql.append("=?");
		}
		
		processor.project(
				sql.toString(),
				new Parameterizer() {

					public void parameterize(PreparedStatement ps) throws SQLException {
						for (int i=0; i<keyInfo.length; i++) {
							owner.getColumn(keyInfo[i][0]).parameterize(ps, i+1, true);
						}
					}
					
				},
				new Projector() {

					public Object project(ResultSet rs) throws SQLException {
						try {
							return itemClass.getConstructor(new Class[] {ResultSet.class}).newInstance(new Object[] {rs});
						} catch (SecurityException e) {
							throw new SQLExceptionEx("Cannot project to "+itemClass, e);
						} catch (InstantiationException e) {
							throw new SQLExceptionEx("Cannot project to "+itemClass, e);
						} catch (IllegalAccessException e) {
							throw new SQLExceptionEx("Cannot project to "+itemClass, e);
						} catch (InvocationTargetException e) {
							throw new SQLExceptionEx("Cannot project to "+itemClass, e);
						} catch (NoSuchMethodException e) {
							throw new SQLExceptionEx("Cannot project to "+itemClass, e);
						}
					}
					
				},
				receiver);
	}

	public boolean isModified() {
		return isModified;
	}

	public void update(SQLProcessor processor, IDatabaseObject subItem) throws SQLException {
		subItem.update(processor, tableName);
	}

	public Class getItemType() {
		return itemClass;
	}

	public void onChange(Column column) {
		for (int i=0; i<keyInfo.length; i++) {
			if (keyInfo[i][0].equals(column.getName())) {
				isModified=true;
			}
		}
		
		if (isModified) {
			Iterator it=master.iterator();
			while (it.hasNext()) {
				link((DatabaseObject) it.next());
			}
		}
	}

	public void setMaster(Collection master) {
		this.master=master;
	}
}
