/*
 * hgcommons 7
 * Hammurapi Group Common Library 
 * Copyright (C) 2003  Hammurapi Group
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * URL: http://www.hammurapi.biz/hammurapi-biz/ef/xmenu/hammurapi-group/products/products/hgcommons/index.html
 * e-Mail: support@hammurapi.biz
 */
package biz.hammurapi.sql.columns;

import java.io.Serializable;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.ResultSetMetaData;
import java.sql.SQLException;
import java.util.HashMap;
import java.util.Map;

import javax.xml.transform.TransformerException;

import org.w3c.dom.Element;
import org.w3c.dom.Node;

import biz.hammurapi.config.ConfigurationException;
import biz.hammurapi.config.Context;
import biz.hammurapi.config.ContextConfigurable;
import biz.hammurapi.config.DomConfigFactory;
import biz.hammurapi.config.DomConfigurable;
import biz.hammurapi.sql.SQLExceptionEx;
import biz.hammurapi.util.Attributable;
import biz.hammurapi.xml.dom.DOMUtils;
import biz.hammurapi.xml.dom.AbstractDomObject;
import biz.hammurapi.xml.dom.CompositeDomSerializer;


/**
 * Base class for different column types
 * @author Pavel Vlasov
 * @version $Revision: 1.14 $
 */
public abstract class Column 
implements 
	Cloneable, 
	ContextConfigurable,
	DomConfigurable,
	Attributable,
	Serializable {
	
	private ColumnChangeListener listener;
    private String name;
    private boolean isModified;
    private boolean isPrimaryKey;
    private String label;
    
	public abstract void setOriginal();
	
    /**
     * Clears modified flag.
     */
    public void clearModified() {
    	isModified=false;
    }
    
    /**
     * Sets field value to default and clears modified flag.
     *
     */
    public abstract void clear();
    
	/**
	 * @param listener The listener to set.
	 */
	public void setListener(ColumnChangeListener listener) {
		this.listener = listener;
	}
	
	/**
	 * @return Returns the listener.
	 */
	public ColumnChangeListener getListener() {
		return listener;
	}
	
	protected void onChange() {
		isModified=true;
		if (listener!=null) {
			listener.onChange(this);
		}
	}
	
    /**
     * @param name Column name
     * @param isPrimaryKey 'true' if column is part of primary key
     */
    public Column(String name, boolean isPrimaryKey) {
        super();
        this.name = name;
        this.isPrimaryKey = isPrimaryKey;
    }
    
    /**
     * @return Column name if value was modified, null otherwise.
     */
    public String listName() {
        return isModified ? name : null;
    }
    
    /**
     * Parameterizes prepared statement. 
     * @param ps
     * @param idx Parameter index
     * @param force If true column is treated a modified
     * @return idx+1 if column was modified and prepared statement was parameterized, idx otherwise.
     */
    public int parameterize(PreparedStatement ps, int idx, boolean force) throws SQLException {
        if (isModified || force) {
        	try {
        		parameterizeInternal(ps, idx);
        	} catch (SQLException e) {
        		throw new SQLExceptionEx("Could not parameterize "+getName()+": "+e, e);
        	}
            return idx+1;
        }
        
		return idx;
    }

    /**
     * Implement this method in subclasses. 
     */
    protected abstract void parameterizeInternal(PreparedStatement ps, int idx) throws SQLException;
    
    /**
     * Implement this method in subclasses. 
     */
    public abstract void parameterizeOriginal(PreparedStatement ps, int idx) throws SQLException;
    
    /**
     * @return Returns the isPrimaryKey.
     */
    public boolean isPrimaryKey() {
        return isPrimaryKey;
    }
    
	/**
	 * @return name
	 */
	public String getName() {
		return name;
	}
	
	/**
	 * Returns column value as object
	 * @param originalValue if true this method returns original value instead of current value.
	 * @return
	 */
	public abstract Object getObjectValue(boolean originalValue);

	/**
	 * @param owner
	 */
	public void toDom(Element owner, String nodeName, boolean originalValue) {
//		String nodeName = domName==null ? name : domName;
		if (nodeName==null) {
			nodeName=getName();
		}
		
		Object objectValue=getObjectValue(false);
		
		if (nodeName.length()==0) {
			// Zero mapping
			return;
		}
		
		if (nodeName.startsWith("@")) {
			// Attribute mapping
			if (objectValue!=null) {
				owner.setAttribute(nodeName.substring(1), objectValue.toString());
			}			
		} else if (nodeName.equals(".")) {	
			// Text mapping
			if (objectValue!=null) {
				owner.appendChild(owner.getOwnerDocument().createTextNode(objectValue.toString()));
			}
		} else if (nodeName.startsWith("!")) {
			// Simple mapping
			if (objectValue!=null) {
				Element el=owner.getOwnerDocument().createElement(nodeName.substring(1));
				owner.appendChild(el);
				el.appendChild(owner.getOwnerDocument().createTextNode(objectValue.toString()));
			}			
		} else {
			// full mapping
			Element el=owner.getOwnerDocument().createElement(nodeName);
			owner.appendChild(el);
			
			CompositeDomSerializer cds = CompositeDomSerializer.getThreadInstance();
			
			if (objectValue==null) {
				el.setAttribute("is-null", "yes");
			} else {
				cds.toDomSerializable(objectValue).toDom(el);
			}
			
			if (isPrimaryKey) {
				el.setAttribute("primary-key", "yes");
			}
			if (isModified) {
				el.setAttribute("modified", "yes");
			}
			
			if (label!=null) {
			    el.setAttribute("label", label);
			}
			
			el.setAttribute("column-type", getType());
			
			el.setAttribute("align", getAlignment());	
			
			if (!attributes.isEmpty()) {
				cds.toDomSerializable(attributes).toDom(AbstractDomObject.addElement(el, "column-attributes"));
			}
		}
	}
	
	protected abstract String getType();
	
	/**
	 * @return
	 */
	protected String getAlignment() {
		return "left";
	}

	/**
	 * @param typeName 
	 * @return Column type for Java type
	 */
	public static Class columnType(String typeName) {
		if ("boolean".equals(typeName)) {
			return BooleanColumn.class;
		} else if ("byte".equals(typeName)) {
			return ByteColumn.class;
		} else if ("char".equals(typeName)) {
			return CharColumn.class;
		} else if ("double".equals(typeName)) {
			return DoubleColumn.class;
		} else if ("float".equals(typeName)) {
			return FloatColumn.class;
		} else if ("int".equals(typeName)) {
			return IntColumn.class;
		} else if ("long".equals(typeName)) {
			return LongColumn.class;
		} else if ("short".equals(typeName)) {
			return ShortColumn.class;
		} else {
			return ObjectColumn.class;
		}
	}
	
	/**
	 * @return Returns the isModified.
	 */
	public boolean isModified() {
		return isModified;
	}

	protected boolean force;
//	private String domName;
	
	/**
	 * @param force
	 */
	public void setForce(boolean force) {
		this.force=force;
	}
	
	/**
	 * @param rs
	 * @param name
	 * @return True if result set has a column with specified name.
	 * @throws SQLException
	 */
	public static boolean hasColumn(ResultSet rs, String name) throws SQLException {
		ResultSetMetaData metaData = rs.getMetaData();
		for (int i=1, c=metaData.getColumnCount(); i<=c; i++) {
			if (name.equals(metaData.getColumnName(i))) {
				return true;
			}
		}
		return false;
	}

	/**
	 * Loads value from XML Element
	 * @param textValue Text value
	 */
	public abstract void load(String textValue);
	
	public Object clone() throws CloneNotSupportedException {
		return super.clone();
	}	

	/**
	 * @return display label for column
	 */
    public String getLabel() {
        return label;
    }
    
    /**
     * Sets display label for column
     * @param label
     */
    public void setLabel(String label) {
        this.label = label;
    }

//	/**
//	 * @param domName
//	 */
//	public void setDomName(String domName) {
//		this.domName=domName;		
//	}
//	
//	public String getDomName() {
//		return domName;
//	}

	/**
	 * Copies values from source column
	 * @param source
	 */
	public abstract void set(Column source);
	
	private Map attributes=new HashMap();
	
	public void setAttribute(Object key, Object value) {
		attributes.put(key, value);		
	}
	
	public Object getAttribute(Object key) {
		return attributes.get(key);
	}
	
	public Object removeAttribute(Object key) {
		return attributes.remove(key);
	}
	
	public void configure(Node configNode, Context context) throws ConfigurationException {
		if (!((Element) configNode).getAttribute("is-null").equals("yes")) {
			load(AbstractDomObject.getElementText(configNode));
		}
		
		DomConfigFactory dcf=new DomConfigFactory();
		attributes.clear();
		try {
			Node attributesNode = DOMUtils.selectSingleNode(configNode, "column-attributes");
			if (attributesNode!=null) {
				attributes.putAll((Map) dcf.create(attributesNode));
			}
		} catch (Exception e) {
			throw new ConfigurationException("Cannot load column attributes: "+e, e);
		}
		
	}	
}
