/*
 * hgcommons 7
 * Hammurapi Group Common Library 
 * Copyright (C) 2003  Hammurapi Group
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * URL: http://www.hammurapi.biz/hammurapi-biz/ef/xmenu/hammurapi-group/products/products/hgcommons/index.html
 * e-Mail: support@hammurapi.biz
 */
package biz.hammurapi.sql.hypersonic;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.Reader;
import java.sql.SQLException;
import java.text.SimpleDateFormat;
import java.util.Date;


/**
 * Hypersonic temporary data source. It is similar to standalone datasource, but
 * data files are created in system temporary directory and are scheduled to deletion 
 * on JVM exit. Use this datasource if you need to keep large amount of temporary data,
 * for example data pumping applications can benefit from such data source.
 * @author Pavel Vlasov
 * @version $Revision: 1.6 $
 */
public class HypersonicTmpDataSource extends HypersonicDataSource {
	
	/** 
	 * @param initScript Fully qualified name of database initialization script to be loaded by 
	 * classloader. Can be null.
	 * @throws ClassNotFoundException
	 * @throws IOException
	 * @throws SQLException
	 */
	public HypersonicTmpDataSource(Reader scriptReader) throws ClassNotFoundException, IOException, SQLException {		
		super("jdbc:hsqldb:"+createTmpDir(), "sa", "", null);
		initDB(scriptReader, null);
	}
	
	/** 
	 * @param initScript Fully qualified name of database initialization script to be loaded by 
	 * classloader. Can be null.
	 * @throws ClassNotFoundException
	 * @throws IOException
	 * @throws SQLException
	 */
	public HypersonicTmpDataSource(String initScript) throws ClassNotFoundException, IOException, SQLException {		
		super("jdbc:hsqldb:"+createTmpDir(), "sa", "", null);
		InputStream in = getClass().getClassLoader().getResourceAsStream(initScript);
		if (in==null) {
			throw new IOException("Resource not found: "+initScript);
		}
		initDB(new InputStreamReader(in), null);
	}
	
	private static String createTmpDir() throws IOException {
		String tmpDir=System.getProperty("java.io.tmpdir");
		SimpleDateFormat sdf=new SimpleDateFormat("yyyyMMddHHmmss");
		String prefix = "HypersonicDB_"+sdf.format(new Date())+"_";
		String suffix = "_TMP";
		final File tmpDbDir = tmpDir==null ? File.createTempFile(prefix, suffix) : File.createTempFile(prefix, suffix, new File(tmpDir));
		if (tmpDbDir.delete()) {
			if (tmpDbDir.mkdirs()) {
				Runtime.getRuntime().addShutdownHook( new Thread() {
					public void run() {						
						
						File[] dbFiles=tmpDbDir.listFiles();
						for (int i=0; i<dbFiles.length; i++) {
							dbFiles[i].deleteOnExit();
						}
						
						tmpDbDir.deleteOnExit();
					}
				});
				return tmpDbDir.getAbsolutePath()+File.separator+"HypersonicTmpDB";
			}
			throw new IOException("Cannot create directory "+tmpDbDir.getAbsolutePath());
		}
		throw new IOException("Cannot delete file "+tmpDbDir.getAbsolutePath());
	}	
}
