/*
  * hgcommons 7
 * Hammurapi Group Common Library 
 * Copyright (C) 2003  Hammurapi Group
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * URL: http://www.hammurapi.biz/hammurapi-biz/ef/xmenu/hammurapi-group/products/products/hgcommons/index.html
 * e-Mail: support@hammurapi.biz
  */
package biz.hammurapi.swing;

import java.io.IOException;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Stack;

import javax.swing.table.DefaultTableModel;
import javax.swing.table.TableModel;
import javax.swing.tree.DefaultMutableTreeNode;
import javax.swing.tree.MutableTreeNode;

import biz.hammurapi.RuntimeException;
import biz.hammurapi.config.ConfigurationException;
import biz.hammurapi.config.DomConfigFactory;
import biz.hammurapi.util.ClassHierarchyVisitable;
import biz.hammurapi.util.Visitor;



/**
 * @author Pavel Vlasov
 *
 * @version $Revision: 1.2 $
 */
public class CompositeVisualizer {
	
	public interface Member {
		/**
		 * Callback method
		 * @param owner
		 */
		void setOwner(CompositeVisualizer owner);
	}
	
	private Collection visualizers=new ArrayList();
	
	/**
	 * Visualizer which to delegate visualization to 
	 * if this visualizer doesn't containt appropriate 
	 * means to visualize given object.
	 */ 	
	protected CompositeVisualizer upInStack;
	
	/**
	 * Visualizer to delegate requests for visualization from members.
	 */
	protected CompositeVisualizer downInStack;
	
	/**
	 * For use by members.
	 * @return CompositeVisualizer at stack's head to delegate requests 
	 * from members to.
	 */
	public CompositeVisualizer getStackHead() {
		return (downInStack==null || downInStack==this) ? this : downInStack.getStackHead();
	}

	/**
	 * Creates visualizer prepopulated with standard visualizers
	 * loaded from com/pavelvlasov/swing/CompositeVisualizer.xml resource.
	 */
	public CompositeVisualizer() {
		Iterator it=standardVisualizers.iterator();
		while (it.hasNext()) {
			addVisualizer(it.next());
		}
	}

	/**
	 * Creates visualizer populated with given collection of visualizers
	 * @param visualizers
	 */
	public CompositeVisualizer(Collection visualizers) {
		Iterator it=visualizers.iterator();
		while (it.hasNext()) {
			addVisualizer(it.next());
		}
	}
	
	private interface Visualizer {
		Class getSourceClass();
		Visualizable toVisualizable(Object o);
	}
	
	public void addVisualizer(final Object v) {
		if (v instanceof Member) {
			((Member) v).setOwner(this);
		}
		
		final Class clazz=v.getClass();
		Method[] ma=clazz.getMethods();
		
		for (int i=0; i<ma.length; i++) {
			if ("toVisualizable".equals(ma[i].getName()) 
					&& Visualizable.class.isAssignableFrom(ma[i].getReturnType()) 
					&& ma[i].getParameterTypes().length==1) {
				
				final Method m=ma[i];
				visualizers.add(new Visualizer() {

					public Class getSourceClass() {
						return m.getParameterTypes()[0];
					}

					public Visualizable toVisualizable(Object o) {						
						try {
							return (Visualizable) m.invoke(v, new Object[] {o});
						} catch (IllegalAccessException e) {
							throw new RuntimeException(e);
						} catch (InvocationTargetException e) {
							throw new RuntimeException(e);
						}
					}					
				});
				
			}
		}
	}	

	private Map visualizerMap=new HashMap();

	private static final Visualizable nullVisualizible=new Visualizable() {

		public MutableTreeNode toTree(final String title) {
			DefaultMutableTreeNode ret=new DefaultMutableTreeNode() {
				public String toString() {
					return title + " - null";
				}
			};
			return ret;
		}

		public TableModel toTable() {
			DefaultTableModel ret=new DefaultTableModel(1,2);
			ret.setValueAt("Value", 0, 0);
			ret.setValueAt("(null)", 0, 1);
			ret.setColumnIdentifiers(new String[] {"Property", "Value"});
			return null;
		}
		
	}; 
	
	public Visualizable toVisualizable(Object object) {
		if (object==null) {
			return nullVisualizible;
		}
		
		if (object instanceof Visualizable) {
			return (Visualizable) object;
		}
		
		Visualizer visualizer;
		synchronized (visualizerMap) {
			visualizer=(Visualizer) visualizerMap.get(object.getClass().getName());
			if (visualizer==null) {
				int affinity=Integer.MAX_VALUE;
				Iterator it=visualizers.iterator();
				while (it.hasNext()) {
					final Visualizer v=(Visualizer) it.next();
					if (v.getSourceClass().isInstance(object) && v.getSourceClass().isArray()==object.getClass().isArray()) {
						final int[] caffinity={0};
						new ClassHierarchyVisitable(object.getClass()).accept(new Visitor() {
	
							public boolean visit(Object target) {
								if (target.equals(v.getSourceClass())) {
									return false;
								} 
								
								caffinity[0]++;
								return true;
							}
							
						});
												
//						System.out.println(object.getClass()+" - "+ds.getSourceClass()+" : "+caffinity[0]);
						
						if (visualizer==null || caffinity[0]<affinity) {
							visualizer=v;
							affinity=caffinity[0];
						}
					}
				}
				
				if (visualizer!=null) {
					visualizerMap.put(object.getClass().getName(), visualizer);
				}
			}			
		}
		
		if (visualizer==null) {
			return upInStack==null ? null : upInStack.toVisualizable(object);
		}
		
		return visualizer.toVisualizable(object);
	}
	
	private static CompositeVisualizer defaultInstance;
	
	private static Collection standardVisualizers;
	
	static {
		try {
			DomConfigFactory factory=new DomConfigFactory();
			standardVisualizers=(Collection) factory.create(CompositeVisualizer.class.getResourceAsStream("CompositeVisualizer.xml"), null);
		} catch (ConfigurationException e) {
			throw new ExceptionInInitializerError(e);
		} catch (IOException e) {
			throw new ExceptionInInitializerError(e);
		}
		defaultInstance=new CompositeVisualizer();
	}
		
	private static ThreadLocal threadVisualizerTL=new ThreadLocal() {
		protected Object initialValue() {
			return new Stack();
		}
	};
	
	public static CompositeVisualizer getThreadInstance() {
		Stack stack=(Stack) threadVisualizerTL.get();
		
		if (stack.isEmpty()) {
			return defaultInstance;
		}
		
		return (CompositeVisualizer) stack.peek();
	}
	
	public static void pushThreadVisualizer(CompositeVisualizer threadVisualizer) {
		if (threadVisualizer!=null) {
			CompositeVisualizer cti = getThreadInstance();
			if (cti!=threadVisualizer) {
				threadVisualizer.upInStack=cti;
				cti.downInStack=threadVisualizer;
				((Stack) threadVisualizerTL.get()).push(threadVisualizer);
			}
		}
	}
	
	public static CompositeVisualizer popThreadVisualizer() {
		Stack stack=(Stack) threadVisualizerTL.get();
		if (stack.isEmpty()) {
			return null;
		}
		
		CompositeVisualizer ret = (CompositeVisualizer) stack.pop();
		if (ret.upInStack!=null) {
			ret.upInStack.downInStack=null;
			ret.upInStack=null;
		}
		return ret;
	}
	
}
