/*
 * hgcommons 7
 * Hammurapi Group Common Library 
 * Copyright (C) 2003  Hammurapi Group
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * URL: http://www.hammurapi.biz/hammurapi-biz/ef/xmenu/hammurapi-group/products/products/hgcommons/index.html
 * e-Mail: support@hammurapi.biz
 */
package biz.hammurapi.util;

import java.lang.reflect.InvocationHandler;
import java.lang.reflect.Method;
import java.lang.reflect.Proxy;

/**
 * Creates a proxy which combines several classes and sequentially searches for 
 * matching method to invoke. This allows to implement lazy instantiation for
 * situations where you have part of object data in cheaply accessible storage (cache or DB)
 * and the other part in expensively accessible storage (e.g. XML file shall be
 * parsed to access all object data).
 * @author Pavel Vlasov
 * @version $Revision: 1.1 $
 */
public class CompositeProxyFactory {
	
	/**
	 * @author Pavel Vlasov
	 * @version $Revision: 1.1 $
	 */
	public interface TargetFactory {
		
		/**
		 * @return target class for the proxy to know which methods it implements.
		 * You might want to return only one of class interfaces or base classes
		 * to limit number of exposed methods.
		 */
		Class getTargetClass();
		
		/**
		 * @return target itself to perform invocation.
		 */
		Object getTarget();
	}
	
	private ClassLoader classLoader;

	/**
	 * Creates factory which uses default classloader. 
	 */
	public CompositeProxyFactory() {
		super();
		classLoader=getClass().getClassLoader();
	}
	
	/**
	 * Creates factory whith specified classloader which will be
	 * passed to createProxy(ClassLoader, Class[], TargetFactory[]) 
	 */
	public CompositeProxyFactory(ClassLoader classLoader) {
		super();
		this.classLoader=classLoader;
	}
	
	public static Object createProxy(ClassLoader classLoader, Class[] interfaces, final TargetFactory[] targetFactories) {
		return Proxy.newProxyInstance(
				classLoader, 
				interfaces, 
				new InvocationHandler() {
					public Object invoke(Object proxy, Method method, Object[] args) throws Throwable {
						for (int i=0; i<targetFactories.length; i++) {
							try {
								return targetFactories[i]
									.getTargetClass()
									.getMethod(method.getName(), method.getParameterTypes())
									.invoke(targetFactories[i].getTarget(), args);
							} catch (NoSuchMethodException e) {
								// Ignore it.
							}
						}
						
						throw new NoSuchMethodException("Method not found in targets: "+method);
					}					
				});
		
	}
	
	public Object createProxy(Class[] interfaces, TargetFactory[] targetFactories) {
		return createProxy(classLoader, interfaces, targetFactories);		
	}	
}
