/*
 * hgcommons 7
 * Hammurapi Group Common Library 
 * Copyright (C) 2003  Hammurapi Group
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * URL: http://www.hammurapi.biz/hammurapi-biz/ef/xmenu/hammurapi-group/products/products/hgcommons/index.html
 * e-Mail: support@hammurapi.biz 
 */

package biz.hammurapi.util;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.ArrayList;
import java.util.List;

/**
 * 
 * Copies all data from an input stream to an output stream.
 * @author Pavel Vlasov
 * @version $Revision: 1.3 $
 */
public class StreamPumper implements Runnable {

    private final static int SIZE = 1024;
    private InputStream is;
    private OutputStream os;
    
	private ExceptionSink sink;
    
    private boolean closeStreams=false;
    
    /**
     * Create a new stream pumper.
     *
     * @param is input stream to read data from
     * @param os output stream to write data to.
     */
    public StreamPumper(InputStream is, OutputStream os, ExceptionSink sink, boolean closeStreams) {
        this.is = is;
        this.os = os;
        this.closeStreams=closeStreams;
        this.sink=sink;
    }

    private List listeners=new ArrayList();
        
    public void addListener(StreamPumpListener listener, int tickSize) {
        synchronized (listeners) {
            listeners.add(new StreamPumpListenerEntry(listener, tickSize));
        }
    }
    
    public void removeListener(StreamPumpListener listener) {
        synchronized (listeners) {
            listeners.remove(listener);
        }
    }

    /**
     * Copies data from the input stream to the output stream.
     * Creates a copy of listeners collection before pumping. 
     * addListener() and removeListener() have no effect once pumping has started. 
     * Terminates as soon as the input stream is closed or an error occurs.
     */
    public void run() {
        StreamPumpListenerEntry[] listenersArray;
        synchronized (listeners) {
            listenersArray= (StreamPumpListenerEntry[]) listeners.toArray(new StreamPumpListenerEntry[listeners.size()]);
        }
        
        for (int i=0; i<listenersArray.length; i++) {
        	listenersArray[i].listener.pumpStarted(this);
        }
        
        long counter=0;
        try {
            final byte[] buf = new byte[SIZE];
            int length;
            while ((length = is.read(buf)) != -1) {
                os.write(buf, 0, length);
                counter+=length;
                for (int i=0; i<listenersArray.length; i++) {
                    listenersArray[i].counter+=length;
                    if (listenersArray[i].counter>=listenersArray[i].tickSize) {
                        listenersArray[i].counter=0;
                        listenersArray[i].listener.tick(this, counter);
                    }
                }                
            }
        } catch(IOException e) {
            handleException(e, listenersArray);
        } finally {
            for (int i=0; i<listenersArray.length; i++) {
            	listenersArray[i].listener.pumpFinished(this);
            }
            if (closeStreams) {
                try {
                    is.close();
                } catch (IOException ie) {
                	handleException(ie, listenersArray);
                }
                
                try {
                    os.close();
                } catch (IOException ie) {
                	handleException(ie, listenersArray);
                }
            }
        }            
    }

	/**
	 * @param e
	 * @param listenersArray
	 */
	private void handleException(Exception e, StreamPumpListenerEntry[] listenersArray) {
        for (int i=0; i<listenersArray.length; i++) {
        	listenersArray[i].listener.pumpingError(this, e);
        }
        
		if (sink==null) {
			e.printStackTrace();
		} else {
			sink.consume(this, e);
		}
	}
}
