/*
  * hgcommons 7
 * Hammurapi Group Common Library 
 * Copyright (C) 2003  Hammurapi Group
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * URL: http://www.hammurapi.biz/hammurapi-biz/ef/xmenu/hammurapi-group/products/products/hgcommons/index.html
 * e-Mail: support@hammurapi.biz
  */
package biz.hammurapi.xml.dom;

import java.io.IOException;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Stack;

import biz.hammurapi.RuntimeException;
import biz.hammurapi.config.ConfigurationException;
import biz.hammurapi.config.DomConfigFactory;
import biz.hammurapi.util.ClassHierarchyVisitable;
import biz.hammurapi.util.Visitor;



/**
 * @author Pavel Vlasov
 *
 * @version $Revision: 1.8 $
 */
public class CompositeDomSerializer {
	public interface Member {
		/**
		 * Callback method
		 * @param owner
		 */
		void setOwner(CompositeDomSerializer owner);
	}
	
	private Collection serializers=new ArrayList();
	
	/**
	 * Serializer which to delegate serialization to 
	 * if this serializer doesn't containt appropriate 
	 * means to serialize given object.
	 */ 	
	protected CompositeDomSerializer upInStack;
	
	/**
	 * Serializer to delegate requests for serialization from members.
	 */
	protected CompositeDomSerializer downInStack;
	
	/**
	 * For use by members.
	 * @return CompositeDomSerializer at stack's head to delegate requests 
	 * from members to.
	 */
	public CompositeDomSerializer getStackHead() {
		return (downInStack==null || downInStack==this) ? this : downInStack.getStackHead();
	}

	/**
	 * Creates serializer prepopulated with standard serializers
	 * loaded from com/pavelvlasov/xml/dom/CompositeDomSerializer.xml resource.
	 */
	public CompositeDomSerializer() {
		synchronized (standardSerializers) {
			Iterator it=standardSerializers.iterator();
			while (it.hasNext()) {
				addDomSerializer(it.next());
			}
		}
	}

	/**
	 * Creates serializer populated with given collection of serializers
	 * @param serializers
	 */
	public CompositeDomSerializer(Collection serializers) {
		Iterator it=serializers.iterator();
		while (it.hasNext()) {
			addDomSerializer(it.next());
		}
	}
	
	private interface DomSerializer {
		Class getSourceClass();
		DomSerializable toDomSerializable(Object o);
	}
	
	public void addDomSerializer(final Object ds) {
		if (ds instanceof Member) {
			((Member) ds).setOwner(this);
		}
		
		final Class clazz=ds.getClass();
		Method[] ma=clazz.getMethods();
		
		for (int i=0; i<ma.length; i++) {
			if ("toDomSerializable".equals(ma[i].getName()) 
					&& DomSerializable.class.isAssignableFrom(ma[i].getReturnType()) 
					&& ma[i].getParameterTypes().length==1) {
				
				final Method m=ma[i];
				serializers.add(new DomSerializer() {

					public Class getSourceClass() {
						return m.getParameterTypes()[0];
					}

					public DomSerializable toDomSerializable(Object o) {						
						try {
							return (DomSerializable) m.invoke(ds, new Object[] {o});
						} catch (IllegalAccessException e) {
							throw new RuntimeException(e);
						} catch (InvocationTargetException e) {
							throw new RuntimeException(e);
						}
					}					
				});
				
			}
		}
	}	

	private Map serializerMap=new HashMap();
	
	public DomSerializable toDomSerializable(Object object) {
		if (object==null) {
			return null;
		}
		
		if (object instanceof DomSerializable) {
			return (DomSerializable) object;
		}
		
		DomSerializer serializer;
		synchronized (serializerMap) {
			serializer=(DomSerializer) serializerMap.get(object.getClass().getName());
			if (serializer==null) {
				int affinity=Integer.MAX_VALUE;
				Iterator it=serializers.iterator();
				while (it.hasNext()) {
					final DomSerializer ds=(DomSerializer) it.next();
					if (ds.getSourceClass().isInstance(object) && ds.getSourceClass().isArray()==object.getClass().isArray()) {
						final int[] caffinity={0};
						new ClassHierarchyVisitable(object.getClass()).accept(new Visitor() {
	
							public boolean visit(Object target) {
								if (target.equals(ds.getSourceClass())) {
									return false;
								} 
								
								caffinity[0]++;
								return true;
							}
							
						});
												
//						System.out.println(object.getClass()+" - "+ds.getSourceClass()+" : "+caffinity[0]);
						
						if (serializer==null || caffinity[0]<affinity) {
							serializer=ds;
							affinity=caffinity[0];
						}
					}
				}
				
				if (serializer!=null) {
					serializerMap.put(object.getClass().getName(), serializer);
				}
			}			
		}
		
		if (serializer==null) {
			return upInStack==null ? null : upInStack.toDomSerializable(object);
		}
		
		return serializer.toDomSerializable(object);
	}
	
	private static CompositeDomSerializer defaultInstance;
	
	private static Collection standardSerializers;
	
	static {
		try {
			DomConfigFactory factory=new DomConfigFactory();
			standardSerializers=(Collection) factory.create(CompositeDomSerializer.class.getResourceAsStream("CompositeDomSerializer.xml"), null);
		} catch (ConfigurationException e) {
			throw new ExceptionInInitializerError(e);
		} catch (IOException e) {
			throw new ExceptionInInitializerError(e);
		}
		defaultInstance=new CompositeDomSerializer();
		if (defaultInstance.serializers.isEmpty()) {
			throw new IllegalStateException("Default composite serializer doesn't contain any serializers.");
		}
	}
		
	private static ThreadLocal threadSerializerTL=new ThreadLocal() {
		protected Object initialValue() {
			return new Stack();
		}
	};
	
	public static CompositeDomSerializer getThreadInstance() {
		Stack stack=(Stack) threadSerializerTL.get();
		
		if (stack.isEmpty()) {
			return defaultInstance;
		}
		
		return (CompositeDomSerializer) stack.peek();
	}
	
	public static void pushThreadSerializer(CompositeDomSerializer threadSerializer) {
		if (threadSerializer!=null) {
			CompositeDomSerializer cti = getThreadInstance();
			if (cti!=threadSerializer) {
				threadSerializer.upInStack=cti;
				cti.downInStack=threadSerializer;
				((Stack) threadSerializerTL.get()).push(threadSerializer);
			}
		}
	}
	
	public static CompositeDomSerializer popThreadSerializer() {
		Stack stack=(Stack) threadSerializerTL.get();
		if (stack.isEmpty()) {
			return null;
		}
		
		CompositeDomSerializer ret = (CompositeDomSerializer) stack.pop();
		if (ret.upInStack!=null) {
			ret.upInStack.downInStack=null;
			ret.upInStack=null;
		}
		return ret;
	}
	
}
